Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude idynamic pcm unionmap heap codes feaps state.
Require Import zigzag worlds rely inject actions schedule retract process.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Section Always.
Variable W : mod. 
Notation coherent := (Mod.coh W).
Notation guar := (Mod.internal W).
Notation tree A := (tree W A).
Notation menv_steps := (menv_steps W).

(* Definition of the modal operator for adherence to a prototol *)

Fixpoint always_path A (s1 : state) p scs (P : state -> tree A -> Prop) :=
  if scs is sc :: scs' then 
    s1 \In coherent /\ 
    forall s2, menv_steps s1 s2 -> 
      [/\ s2 \In tsafe p sc, P s2 p &
          forall s3 q, tstep s2 p sc s3 q -> always_path s3 q scs' P]
  else s1 \In coherent /\ forall s2, menv_steps s1 s2 -> P s2 p.

(* unfolding of the fixed point in always_path *)
(* to pull out the common parts of the branches *)

Lemma alwpE A s1 (p : tree A) scs P :
        always_path s1 p scs P <->
        s1 \In coherent /\
        forall s2, menv_steps s1 s2 -> P s2 p /\
          if scs is sc :: scs' then 
             s2 \In tsafe p sc /\
             forall s3 q, tstep s2 p sc s3 q -> always_path s3 q scs' P
          else True.
Proof. 
split; case: scs=>[|sc scs] /=. 
- by case=>C H; split=>// s2; move/H.
- by case=>C H; split=>// s2; case/H. 
- by case=>C H; split=>// s2; case/H. 
by case=>C H; split=>// s2; case/H=>? []. 
Qed.

Definition always A s (p : tree A) P := forall scs, always_path s p scs P.

(* some structural properties of always_path and always *)

Lemma alwp_coh A s (p : tree A) scs P : 
        always_path s p scs P -> s \In coherent.
Proof. by case/alwpE. Qed.

Lemma alw_coh A s (p : tree A) P : 
        always s p P -> s \In coherent.
Proof. by move/(_ [::]); move/alwp_coh. Qed.

Lemma alwp_safe A s (p : tree A) sc scs P : 
        always_path s p (sc :: scs) P -> s \In tsafe p sc.
Proof. by case/alwpE=>C T; case/T: (menvs_refl C)=>? []. Qed.

Lemma alw_safe A s (p : tree A) P :
        always s p P -> forall sc, s \In tsafe p sc.
Proof. by move=>H sc; apply: alwp_safe (H [:: sc]). Qed.

Lemma alwp_refl A s (p : tree A) sc P : always_path s p sc P -> P s p.
Proof. by case/alwpE=>C T; case/T: (menvs_refl C). Qed.

Lemma alw_refl A s (p : tree A) P : always s p P -> P s p.
Proof. by move/(_ [::])/alwp_refl. Qed.

Lemma alwp_envs A s1 (p : tree A) scs s2 P : 
        always_path s1 p scs P -> menv_steps s1 s2 -> always_path s2 p scs P.
Proof.
case/alwpE=>C1 T E; case/menvs_coh: (E)=>_ C2; apply/alwpE. 
by split=>// s3; move/(menvs_trans E)/T. 
Qed.
 
Lemma alw_envs A s1 (p : tree A) s2 P :
        always s1 p P -> menv_steps s1 s2 -> always s2 p P.
Proof. by move=>S E scs; apply: alwp_envs (S scs) E. Qed.

Lemma alw_tstep A s1 (p : tree A) sc s2 q P :
        always s1 p P -> tstep s1 p sc s2 q -> always s2 q P.
Proof.
move=>Ls Ts; move: (Ls [:: sc])=>/= [C].
case/(_ _ (menvs_refl C))=>_ _; move/(_ _ _ Ts)=>_. 
move=>scs; move: (Ls (sc :: scs))=>/= [_].
by case/(_ _ (menvs_refl C))=>_ _; move/(_ _ _ Ts). 
Qed.

Lemma alwp_unfin A s1 scs (P : state -> tree A -> Prop) :
        s1 \In coherent -> 
        (forall s2, menv_steps s1 s2 -> P s2 Unfinished) -> 
        always_path s1 Unfinished scs P.
Proof. 
case: scs=>[|sc scs] C H; split=>// s2 E.
split; last by [move=>s3 q; move/tstUnfin].
- by split; [case/menvs_coh: E | case: sc].
by apply: H E.
Qed.

Lemma alw_unfin A s1 (P : state -> tree A -> Prop) :
        s1 \In coherent -> 
        (forall s2, menv_steps s1 s2 -> P s2 Unfinished) ->
        always s1 Unfinished P.
Proof. by move=>C H scs; apply: alwp_unfin. Qed.

Lemma alwp_ret A s1 (v : A) scs (P : state -> tree A -> Prop) : 
        s1 \In coherent -> 
        (forall s2, menv_steps s1 s2 -> P s2 (Ret v)) -> 
        always_path s1 (Ret v) scs P.
Proof. 
case: scs=>[|sc scs] C H; split=>// s2 E.
split; last by [move=>s3 q; move/tstRet].
- by split; [case/menvs_coh: E | case: sc].
by apply: H E.
Qed.

Lemma alw_ret A s1 (v : A) (P : state -> tree A -> Prop) : 
        s1 \In coherent -> 
        (forall s2, menv_steps s1 s2 -> P s2 (Ret v)) -> 
        always s1 (Ret v) P.
Proof. by move=>C H ps; apply: alwp_ret. Qed.

Lemma alw_act A s1 (a : int_action guar A) (P : state -> tree A -> Prop) :
        s1 \In coherent ->
        (forall s2, menv_steps s1 s2 -> 
          [/\ s2 \In asafe a, P s2 (Act a) &
              forall s3 v s4, astep s2 a s3 v -> 
                              menv_steps s3 s4 -> P s4 (Ret v)]) ->
        always s1 (Act a) P. 
Proof.
move=>C H [|sc scs]; split=>// s2; case/H=>// H1 H2 H3.
split=>//; first by split; [move/asf_coh: H1 | case: sc]. 
move=>s3 q /tstAct [v][_ -> St].
apply: alwp_ret; first by case/ast_coh: St.
by move=>s4; apply: H3 St.
Qed.

(* special case of always_path and always, when P = True *)

Notation alwsafe_path s p scs := (always_path s p scs (fun _ _ => True)).
Notation alwsafe s p := (always s p (fun _ _ => True)).

(* the frame lemma *)

Lemma alw_zagzig A s p (t : tree A) Q : 
        always (s \> p) t Q ->
        always (s \< p) t (fun s t => 
          exists s', s = s' \< p /\ Q (s' \> p) t). 
Proof.
move=>Ls scs.
have: s \< p \In coherent by apply/coh_zigzag; apply: alw_coh Ls.
elim: scs s t Ls=>[|sc scs IH] s t Ls C /=;
split=>// s' E; case/menvs_zigeq: E (E)=>s2 [->{s'}] E;
move/(menvs_zigzag E)=>{E} E {C}.
- by eexists s2; split=>//; apply: alw_refl; apply: alw_envs Ls E.
move/(alw_envs Ls): E=>{Ls} Ls; split.
- by apply: tsf_zagzig; apply: alw_safe.
- by eexists s2; split=>//; apply: alw_refl. 
move=>s3 q; case/(tst_frame (alw_safe Ls sc))=>s1 [->] E.
move/(alw_tstep Ls)=>{Ls} Ls; apply: IH=>//.
by apply/coh_zigzag; apply: alw_coh.
Qed.

(* always is a natural operator *)
(* what's the official name for this in modal logic? K? *)

Lemma alwp_imp A s (p : tree A) scs (P1 P2 : state -> tree A -> Prop) : 
         (forall s p, s \In coherent -> P1 s p -> P2 s p) -> 
         always_path s p scs P1 -> always_path s p scs P2.
Proof.
elim: scs s p=>[|sc scs IH] s p /= I; case=>C L; split=>// s2 E.
- by apply: I (L _ E); case/menvs_coh: E.
case/L: (E)=>S H T; split=>//; first by apply: I H; case/menvs_coh: E. 
by move=>s3 q; move/T; apply: IH I. 
Qed.

Lemma alw_imp A s (p : tree A) (P1 P2 : state -> tree A -> Prop) : 
        (forall s p, s \In coherent -> P1 s p -> P2 s p) -> 
        always s p P1 -> always s p P2.
Proof. by move=>I H ps; apply: alwp_imp I _. Qed.

(* always commutes with universals *)

Lemma alwpA A B s (p : tree A) scs (P : B -> state -> tree A -> Prop) : 
        alwsafe_path s p scs ->
        (always_path s p scs (fun s' p' => forall x, P x s' p') <->
         forall x, always_path s p scs (fun s' p' => P x s' p')).
Proof.
move=>Ls; split=>[{Ls}|]. 
- elim: scs s p=>[|sc scs IH] s p /= [C Et x]; split=>// s2; move/Et=>//.
  by case=>S Ha L; split=>// s3 q; move/L/IH.
elim: scs s p Ls=>[|sc scs IH] s p /= [C Et Ha]; split=>// s2 E.
- by move=>x; case: (Ha x)=>_; apply. 
case/Et: (E)=>/= S _ L; split=>//.
- by move=>x; case: (Ha x)=>_; case/(_ _ E). 
move=>s3 q T; apply: IH; first by apply: L T.
by move=>x; case: (Ha x)=>_; case/(_ _ E)=>_ _; apply. 
Qed.

Lemma alwA A B s (p : tree A) (P : B -> state -> tree A -> Prop) : 
        alwsafe s p ->
        (always s p (fun s' p' => forall x, P x s' p') <->
         forall x, always s p (fun s' p' => P x s' p')).
Proof.
move=>Ls; split.
- by move=>H x ps; move: x; apply/(alwpA _ (Ls ps)).
by move=>H ps; apply/(alwpA _ (Ls ps))=>x; apply: H.
Qed.

Implicit Arguments alwA [A B s p P].

(* always commutes with implication *)

Lemma alwpI A s (p : tree A) scs (P : Prop) (Q : state -> tree A -> Prop) : 
        alwsafe s p ->
        (always_path s p scs (fun s' p' => P -> Q s' p') <->
         (P -> always_path s p scs (fun s' p' => Q s' p'))).
Proof.
move=>Ls; split.
- elim: scs s p Ls=>[|sc scs IH] s p Ls /= [C Et H]; split=>// s2.
  - by move/Et; apply.
  move=>M; move: (alw_envs Ls M)=>{Ls} Ls.
  case/Et: M=>H1 /(_ H) H2 H3; split=>// s3 q T.
  by apply: IH H; [apply: alw_tstep T | apply: H3].
elim: scs s p Ls=>[|sc scs IH] s p /= Ls H; 
move: (alw_coh Ls)=>C; split=>// s2 M; first by case/H=>_; apply. 
move: (alw_envs Ls M)=>{Ls} Ls.
split; first by move/alw_safe: Ls.  
- by case/H=>_; case/(_ _  M). 
move=>s3 q T; move: (alw_tstep Ls T)=>{Ls} Ls.
by apply: IH Ls _; case/H=>_; case/(_ _ M)=>_ _; apply. 
Qed.

Lemma alwI A s (p : tree A) (P : Prop) (Q : state -> tree A -> Prop) : 
        alwsafe s p ->
        (always s p (fun s' p' => P -> Q s' p') <->
         P -> always s p (fun s' p' => Q s' p')). 
Proof.
move=>Ls; split; first by move=>H Hp scs; apply/alwpI: Hp.
by move=>H scs; apply/alwpI=>//; move/H; move/(_ scs).
Qed.

Implicit Arguments alwI [A s p P Q].

(**************************************)
(* Always with sequential composition *)
(**************************************)

Lemma alw_bnd A B (p1 : tree A) (p12 : tree B) pp2 s1 P :
         p12 \In tcat p1 pp2 -> 
         always s1 p1 (fun s2 p2 => 
           forall p12, p12 \In tcat p2 pp2 -> 
                       always s2 p12 P) ->
         always s1 p12 P.
Proof.
move=>Tc Ls scs. 
elim: scs p1 Tc Ls=>[|sc scs IH] p1 Tc Ls /=.
- split=>[|s2]; first by apply: alw_coh Ls.
  by move/(alw_envs Ls)/alw_refl/(_ _ Tc)/alw_refl. 
split=>[|s2 E]; first by apply: alw_coh Ls.
split; move/(alw_envs Ls)/alw_refl/(_ _ Tc): E.
- by move/alw_safe.
- by move/alw_refl. 
by move=>{Ls} Ls s3 q; move/(alw_tstep Ls). 
Qed.

(* A slightly different formulation of alw_bnd *)
(* which is more appropriate for Hoare logic stuff *)
(* I could not, however, prove it out of alw_bnd *)

Lemma alw_bnd' A B (p1 : tree A) (p12 : tree B) pp2 s1 
                 (P : state -> B -> Prop) :
        p12 \In tcat p1 pp2 -> 
        always s1 p1 (fun s2 p2 => 
          forall p v, p2 = Ret v -> p \In pp2 v -> 
            always s2 p (fun s p => forall v, p = Ret v -> P s v)) ->
        always s1 p12 (fun s p => forall v, p = Ret v -> P s v).
Proof.
move=>Tc Ls scs.
elim: scs s1 p1 p12 Tc Ls=>[|sc scs IH] s1 p1 p12 Tc Ls /=.
- by split=>[|s2]; [apply: alw_coh Ls | case: Tc=>k [->]].
split=>[|s2 E]; first by apply: alw_coh Ls.
case: Tc IH=>k [->{p12}] H IH.
split=>//; last first.
- move=>s3 q T; case/tstSeq: T Ls.
  - case=>v [_ ->->-> C]; move/alw_refl/(_ _ _ (erefl _)). 
    by move/(_ _ (H v))/alw_envs; apply. 
  case=>sc' [p'][_ ->{q}] T; move/alw_envs/(_ E)/alw_tstep.
  by move/(_ _ _ _ T)/IH; apply; exists k.
move/(alw_envs Ls): E=>{Ls} Ls. 
have Ls': always s2 p1 (fun s2 p2 =>
  forall sc p v, p2 = Ret v -> p \In pp2 v -> s2 \In tsafe p sc).
- by apply: alw_imp Ls=>s p _ I sc' q v E; move/(I _ _ E)/alw_safe. 
split; first by apply: alw_coh Ls'.
case: sc=>//=; first by case: p1 {Ls} Ls'.
by move=>sc; case: (Ls [:: sc])=>C; case/(_ _ (menvs_refl C)); case.
Qed.

Lemma alwsafe_bnd' A B (p1 : tree A) (p12 : tree B) s1 pp2 :
        p12 \In tcat p1 pp2 ->
        always s1 p1 (fun s2 p2 =>
          forall p v, p2 = Ret v -> p \In pp2 v -> alwsafe s2 p) ->
        alwsafe s1 p12.
Proof.
move=>T Ls.
suff H: always s1 p12 (fun s p => forall v, p = Ret v -> True).
- by apply: alw_imp H.
apply: alw_bnd' T _; apply: alw_imp Ls=>s p _ I q v H.
by move/(I _ _ H); apply: alw_imp.
Qed.

(* this suggest capturing the pattern by means of a new definition *)

Definition after A s (p : tree A) (P : A -> state -> Prop) := 
  always s p (fun s2 p2 => forall v, p2 = Ret v -> P v s2).

Lemma aft_zagzig A s p (t : tree A) Q : 
        after (s \> p) t Q ->
        after (s \< p) t (fun v s => 
          exists s', s = s' \< p /\ Q v (s' \> p)).
Proof.
move/alw_zagzig; apply: alw_imp=>s1 q _ [s2][E] T v F.
by exists s2; split=>//; apply: T F.
Qed.

Lemma aft_bnd A B (p1 : tree A) (p12 : tree B) pp2 s1 P :
        p12 \In tcat p1 pp2 -> 
        after s1 p1 (fun v s => 
          forall p, p \In pp2 v -> after s p P) ->
        after s1 p12 P.
Proof. 
move=>T H; apply: alw_bnd' T _.
by apply: alw_imp H=>s p _ I q v E; move/(I v E).
Qed.

Lemma aft_alwsf A s (p : tree A) :
        alwsafe s p <-> after s p (fun v s => True).
Proof. by split; apply: alw_imp. Qed.

Lemma aft_imp A s (p : tree A) (P1 P2 : A -> state -> Prop) : 
        (forall v s, s \In coherent -> 
                     P1 v s -> P2 v s) -> 
        after s p P1 -> after s p P2.
Proof. by move=>I; apply: alw_imp=>s1 p1 C H v; move/H; apply: I. Qed.

Lemma aftA A B s (p : tree A) (P : B -> A -> state -> Prop) : 
        alwsafe s p ->
        (after s p (fun v s' => forall x, P x v s') <->
         forall x, after s p (fun v s' => P x v s')).
Proof.
move=>Ls; rewrite -(alwA Ls).
split; apply: alw_imp=>t q _ I.
- by move=>x v; move/I.
by move=>v; move/I.
Qed.

Implicit Arguments aftA [A B s p P].

Lemma aftI A s (p : tree A) (P : Prop) (Q : A -> state -> Prop) : 
        alwsafe s p ->
        (after s p (fun v s' => P -> Q v s') <->
         P -> after s p (fun v s' => Q v s')).
Proof.
move=>Ls; rewrite -(alwI Ls).
split; apply: alw_imp=>t q _ I.
- by move=>Hp v; move/I; apply. 
by move=>v; move/I. 
Qed.

Implicit Arguments aftI [A s p P Q].

Lemma alwsf_bnd A B (p1 : tree A) (p12 : tree B) s1 pp2 :
        p12 \In tcat p1 pp2 ->
        after s1 p1 (fun v s => 
          forall p, p \In pp2 v -> alwsafe s p) ->
        alwsafe s1 p12.
Proof.
move=>T Ha; apply/aft_alwsf; apply: aft_bnd T _.
by apply: aft_imp Ha=>s v C I p K; apply/aft_alwsf; apply: I K.
Qed.

(***********************************)
(* always and parallel composition *)
(***********************************)

Section AlwaysPar.
Variables (A1 A2 : Type).

Lemma alw_safe_par lp1 lp2 sh ep (t1 : tree A1) (t2 : tree A2) P1 P2 sc :
        always ([lp1, sh, ep] \> lp2) t1 P1 ->
        always ([lp2, sh, ep] \> lp1) t2 P2 -> 
        [lp1 \* lp2, sh, ep] \In tsafe (Par t1 t2 Ret) sc.
Proof.
move/alw_zagzig=>Ls1; move/alw_zagzig=>Ls2.
rewrite !zigE (zipC lp2) in Ls1 Ls2.
split; first by move/alw_coh: Ls1.
case: sc=>//=; first by case: t1 Ls1 Ls2=>// v1; case: t2.
- move=>sc.
  move/alw_coh: (Ls1)=>C1; case/(_ [:: sc]): Ls1=>_. 
  by case/(_ _ (menvs_refl C1)); case. 
move=>sc. 
move/alw_coh: (Ls2)=>C2; case/(_ [:: sc]): Ls2=>_.
by case/(_ _ (menvs_refl C2)); case.  
Qed.

Lemma alw_envs_zip lp1 lp2 sh ep s (t1 : tree A1) (t2 : tree A2) P1 P2 :
        always ([lp1, sh, ep] \> lp2) t1 P1 ->
        always ([lp2, sh, ep] \> lp1) t2 P2 -> 
        menv_steps [lp1 \* lp2, sh, ep] s ->
        exists sh' ep', 
          [/\ s = [lp1 \* lp2, sh', ep'], 
              always ([lp1, sh', ep'] \> lp2) t1 P1 & 
              always ([lp2, sh', ep'] \> lp1) t2 P2].
Proof.
rewrite -zigE=>Ls1 Ls2 E; case/menvs_zigeq: E (E).
case=>_ sh' ep' [->{s}] /= <- E.
exists sh', ep'; split=>//; last first. 
- by apply: alw_envs Ls2 (menvs_zigzag _ _).
apply: alw_envs Ls1 (menvs_zigzag _ _)=>//. 
by rewrite !zigE !(zipC lp2) -!zigE. 
Qed.

Lemma alw_par_main lp1 lp2 sh ep (t1 : tree A1) (t2 : tree A2) P1 P2 :
        always ([lp1, sh, ep] \> lp2) t1 P1 ->
        always ([lp2, sh, ep] \> lp1) t2 P2 -> 
        always [lp1 \* lp2, sh, ep] (Par t1 t2 Ret)
          (fun s' t12' => 
             exists lp1' lp2' sh' ep',
               [/\ s' = [lp1' \* lp2', sh', ep'],
                   forall t1' t2', t12' = Par t1' t2' Ret ->
                     P1 ([lp1', sh', ep'] \> lp2') t1' /\
                     P2 ([lp2', sh', ep'] \> lp1') t2' &
                   forall v1' v2', t12' = Ret (v1', v2') ->
                     P1 ([lp1', sh', ep'] \> lp2') (Ret v1') /\
                     P2 ([lp2', sh', ep'] \> lp1') (Ret v2')]).
Proof.
move=>Ls1 Ls2 scs.
have C: [lp1 \* lp2, sh, ep] \In coherent.
- by move/alw_coh/coh_zigzag: Ls1; rewrite zigE zipC.
elim: scs lp1 lp2 sh ep t1 t2 C Ls1 Ls2 => [|sc scs IH]
          lp1 lp2 sh ep t1 t2 C Ls1 Ls2; split=>//= s'; 
case/(alw_envs_zip Ls1 Ls2)=>sh' [ep'][->] {C Ls1 Ls2 s'} Ls1 Ls2.
- exists lp1, lp2, sh', ep'; split=>// t1' t2'.
  by case/parEX=><-<- _; split; apply: alw_refl.
split; first by apply: alw_safe_par Ls1 Ls2.
- exists lp1, lp2, sh', ep'; split=>// t1' t2'.
  by case/parEX=><-<- _; split; apply: alw_refl.
move=>s3 q T; case/tstPar: T Ls1 Ls2.
- case=>_ [v1][v2][->->->->{t1 t2 q s3}] C Ls1 Ls2; apply: alw_ret=>// s''.
  case/(alw_envs_zip Ls1 Ls2)=>sh'' [ep''][->] {C Ls1 Ls2 s''} Ls1 Ls2.
  by exists lp1, lp2, sh'', ep''; split=>// _ _ [<-<-]; split; apply: alw_refl.
case=>t1' [t2'][sc'][->]; case; case=>_ -> T Ls1 Ls2. 
- rewrite zipC -zigE in T; case/(tst_frame (alw_safe Ls1 sc')): T.
  case=>lp1' sh1' _ [->{s3}] /= -> T; move/(alw_tstep Ls1): (T)=>Ls1'.
  move/coh_zigzag: (alw_coh Ls1'); rewrite zigE zipC => C'.
  apply: IH (C') (Ls1') _; apply: alw_envs (Ls2) _.
  exists 1, ([lp2, sh1', ep'] \> lp1'); do !split=>//. 
  - by left; move/tst_guar/guar_rely: T. 
  by rewrite -coh_zigzag zigE.
rewrite -zigE in T; case/(tst_frame (alw_safe Ls2 sc')): T. 
case=>lp2' sh2' _ [->{s3}] /= -> T; move/(alw_tstep Ls2): (T)=>Ls2'.
move/coh_zigzag: (alw_coh Ls2'); rewrite zigE=>C'.
apply: IH (C') _ (Ls2'); apply: alw_envs (Ls1) _.
exists 1, ([lp1, sh2', ep'] \> lp2'); do !split=>//. 
- by left; move/tst_guar/guar_rely: T. 
by rewrite -coh_zigzag zigE zipC. 
Qed.

(* splitting the two cases of alw_par_main: *)
(* when p1 || p2 did not terminate *)

Lemma alw_par lp1 lp2 sh ep (t1 : tree A1) (t2 : tree A2) P1 P2 :
        always ([lp1, sh, ep] \> lp2) t1 P1 ->
        always ([lp2, sh, ep] \> lp1) t2 P2 -> 
        always [lp1 \* lp2, sh, ep] (Par t1 t2 Ret)
          (fun s' t12' => 
             exists lp1' lp2' sh' ep',
               s' = [lp1' \* lp2', sh', ep'] /\
               forall t1' t2', t12' = Par t1' t2' Ret ->
                 P1 ([lp1', sh', ep'] \> lp2') t1' /\
                 P2 ([lp2', sh', ep'] \> lp1') t2').
Proof.
move=>Ls1 Ls2; move: (alw_par_main Ls1 Ls2); apply: alw_imp.
by move=>s p _ [lp1'][lp2'][sh'][ep'][->] L1 L2; exists lp1', lp2', sh', ep'. 
Qed.

(* and when p1 || p2 did terminate *)

Lemma aft_par lp1 lp2 sh ep (t1 : tree A1) (t2 : tree A2) P1 P2 :
        after ([lp1, sh, ep] \> lp2) t1 P1 ->
        after ([lp2, sh, ep] \> lp1) t2 P2 -> 
        after [lp1 \* lp2, sh, ep] (Par t1 t2 Ret)
          (fun v' s' => 
             exists lp1' lp2' sh' ep',
               [/\ s' = [lp1' \* lp2', sh', ep'], 
                   P1 v'.1 ([lp1', sh', ep'] \> lp2') &
                   P2 v'.2 ([lp2', sh', ep'] \> lp1')]).
Proof.
move=>Ls1 Ls2; move: (alw_par_main Ls1 Ls2); apply: alw_imp.
move=>s p _ [lp1'][lp2'][sh'][ep'][->] L1 L2 [v1 v2]. 
case/L2=>H1 H2; exists lp1', lp2', sh', ep'.
by split=>//; [apply: H1 | apply: H2].
Qed.

Lemma alwsf_par lp1 lp2 sh ep (t1 : tree A1) (t2 : tree A2) :
        alwsafe ([lp1, sh, ep] \> lp2) t1 ->
        alwsafe ([lp2, sh, ep] \> lp1) t2 ->
        alwsafe [lp1 \* lp2, sh, ep] (Par t1 t2 Ret).
Proof. by move=>Ls1 Ls2; move: (alw_par Ls1 Ls2); apply: alw_imp. Qed.

End AlwaysPar.

End Always.

Notation alwsafe_path s p scs := (always_path s p scs (fun _ _ => True)).
Notation alwsafe s p := (always s p (fun _ _ => True)).

(**************************)
(* always and hide/refine *)
(**************************)

Section AlwaysHide.
Variables (W V : mod) (A : Type) (U : Type) (f : refinement U W V).

(* Steps by environment in the coarse world, don't change the refining
ghost gE. This is so because gE is a part of the refined world, so
it's out of reach to the stepping in the coarse world. *)

Lemma rf_envs i m (gL : U) i' : 
        f gL i i' -> menv_steps W i m -> 
        exists m', f gL m m' /\ menv_steps V i' m'.
Proof.
move=>Rfi [n E].
elim: n i m gL i' E Rfi=>[|n IH] i m gL i' /= E Rfi.
- case: E=><-{m} C; exists i'; split=>//.
  by apply: menvs_refl; case/rf_coh: Rfi. 
case: E=>h [Rih] E.
case: (rf_mrely Rfi Rih)=>h' [Rfh] Rih'. 
case: (IH _ _ _ _ E Rfh)=>m' [Rfm] E'. 
exists m'; split=>//; apply: menvs_trans E'.
by exists 1, h'; do !split=>//; case/rf_coh: Rfh. 
Qed.

(* let's try forward reasonign *)

(* I need to say that whatever P states about i' p *)
(* when run in Hide, we can assume that gE remains fixed *)
Lemma alw_hide (p : tree V A) (P : state -> tree V A -> Prop) i i' gL :
        f gL i i' -> always i' p P ->
        always i (Hide (Retracts gL f) p) (fun m q =>
          (exists m' gL' q', 
             [/\ q = Hide (Retracts gL' f) q', 
                 f gL' m m' & P m' q']) \/
           exists m' gL' v', 
             [/\ q = Ret v', f gL' m m' & 
                 P m' (Ret v')]).
Proof.
move=>Rfi Lsi scs.
elim: scs p i i' gL Rfi Lsi=>[|sc scs IH] p i i' gL Rfi Lsi /=;
(split=>[|m]; first by [case/rf_coh: Rfi];
 case/(rf_envs Rfi)=>m' [Rfm];
 move/(alw_envs Lsi)=>Lsm).
- by left; move/alw_refl: Lsm; exists m', gL, p. 
split. 
- split; first by case/rf_coh: Rfm. 
  case: sc=>//. 
  - by move=>s; case: (alw_safe Lsm s); exists m'. 
  by case: p Lsm Lsi=>// v Lsm Lsi; exists m'. 
- by left; exists m', gL, p; do !split=>//; move/alw_refl: Lsm. 
move=>h q; rewrite tstHideMorph; case=>H; last first. 
- case: H=>_ [h'][gL'][t'][sc'][_ ->] /(rf_func Rfm) <- Rfh.
  by move/(alw_tstep Lsm); apply: IH Rfh. 
case: H Lsi Lsm=>s1' [v][_ ->->->] _ Lsi Lsm.
apply: alwp_ret=>[|k]; first by case/rf_coh: Rfm. 
case/(rf_envs Rfm)=>k' [Rfk].
by move/(alw_envs Lsm)/alw_refl; right; exists k', gL, v.  
Qed.

(*
(* a backward version is derivable *)
Lemma alw_hide' (p : tree V A) (P : state -> tree W A -> Prop) i i' gL gE : 
        refine f i (gL, gE) i' -> 
        always i' p (fun m' p' => 
          exists m gL', refine f m (gL', gE) m' /\ 
            (P m (Hide (Morph f gL' gE) p') /\
            forall v, p' = Ret v -> P m (Ret v))) ->
        always i (Hide (Morph f gL gE) p) P.
Proof.
move=>Rfi /(alw_hide Rfi); apply: alw_imp=>m q C.
case; case=>m' [gL'][q'][-> Rfm]; case=>_ [_][] /(rf_inj Rfm) [<-<-][H1 H2] //.
by apply: H2.
Qed.
*)

(* specializing to after *)

Lemma aft_hide (p : tree V A) (P : A -> state -> Prop) i i' gL : 
        f gL i i' -> 
        after i' p P ->
        after i (Hide (Retracts gL f) p) (fun v m =>
           exists m' gL', f gL' m m' /\ P v m').
Proof.
move=>R /(alw_hide R); apply: alw_imp=>s q C.  
case; first by case=>m' [gL'][q'][->].  
case=>m' [gL'][v'][-> Rm H] v [<-]. 
by exists m', gL'; split=>//; apply: H. 
Qed.

(*
Lemma aft_hide' (p : tree V A) (P : A -> state -> Prop) i i' gL gE : 
        refine f i (gL, gE) i' -> 
        after i' p (fun v' m' =>
           exists m gL', refine f m (gL', gE) m' /\ P v' m) ->
        after i (Hide (Morph f gL gE) p) P. 
Proof.
move=>R /(alw_hide R); apply: alw_imp=>s q C.
case; first by case=>m' [gL'][q'][->].  
case=>m' [gL'][v'][-> Rm H] v [<-]. 
by case: (H _ (erefl _))=>_ [_][] /(rf_inj Rm) [<- _]. 
Qed.
*)

End AlwaysHide.

(*********************)
(* always and extend *)
(*********************)

Section AlwaysExtend.
Variables (V W : mod) (A : Type) (w : injects V W).
Notation W2 := (inj_ext w).

(* a helper, to avoid casing over w constantly *)

Lemma menvs_ext i j s : 
        i \In Mod.coh V -> 
        menv_steps W (i \+ j) s ->  
        exists i' j', s = i' \+ j' /\ i' \In Mod.coh V.
Proof.
move=>C M; case: (menvs_coh M)=>_; rewrite (cohE w).
by case=>_ [s1][s2][->] Cs1 _; exists s1, s2. 
Qed.

Lemma alw_extend (p : tree V A) (P : state -> tree V A -> Prop) i j :
        i \+ j \In Mod.coh W ->
        always i p P ->
        always (i \+ j) (Extend w p)
          (fun m q => exists i' j', 
             [/\ m = i' \+ j', i' \In Mod.coh V, menv_steps W2 j j' &
                 (exists q', q = Extend w q' /\ P i' q') \/
                 (exists v', q = Ret v' /\ P i' (Ret v'))]).
Proof.
move=>C Ls scs; elim: scs i j p C Ls=>[|sc scs IH] i j p C Ls /=; 
split=>// {C} s M; move: (alw_coh Ls) (proj2 (menvs_coh M))=>Ci C;
case/(menvs_ext Ci): M C (M)=>i1 [j1][->{s}] Ci1 C;
case/(menvs_split w Ci Ci1)=> /(alw_envs Ls) {Ls} Ls S1.
- by exists i1, j1; split=>//; left; exists p; move/alw_refl: Ls. 
split.
- split=>//; case: sc=>//; last by case: p Ls=>// v; exists i1, j1. 
  by move=>sc; case: (alw_safe Ls sc)=>_ Sf; exists i1; split=>//; exists j1. 
- by exists i1, j1; split=>//; left; exists p; move/alw_refl: Ls.
move=>s q; rewrite tstExtend => H.
case: H Ls.
- case=>_ [v][_ ->->->{p q s} _] Ls;  apply: alwp_ret=>// s M. 
  case/(menvs_ext Ci1): M (M)=>i2 [j2][->{s}] Ci2.
  case/(menvs_split w Ci1 Ci2)=> /(alw_envs Ls) {Ls} Ls S2.
  exists i2, j2; split=>//; first by apply: menvs_trans S1 S2.
  by right;  exists v; move/alw_refl: Ls.
case=>sc' [q'][x1][i2][y1][_ -> E -> {sc q s}] _ T Ls.  
have [E1 E2] : x1 = i1 /\ y1 = j1.
- case/tst_coh: T=>Cx1 _.
  move: (coh_prec (cohV C) E Ci1 Cx1) (E)=><-{E Cx1 x1}.
  by move/(joinsK (cohV C)).
rewrite {E x1}E1 {y1}E2 in T *.
have C' : i2 \+ j1 \In Mod.coh W.
- by rewrite (cohE w) in C *; move/tst_guar/int_guar_extend: T=><-. 
move/(alw_tstep Ls): T=>{Ls} Ls.
apply: alwp_imp (IH _ _ _ C' Ls)=>{IH Ls C' C Ci Ci1 i i1 i2 p q' sc' scs}.
move=>s p _ [i2][j2][->{s}] Ci2 S2 H; exists i2, j2; split=>//.
by apply: menvs_trans S1 S2.
Qed.

(* specializing to after *)

Lemma aft_extend (p : tree V A) (P : A -> state -> Prop) i j :
        i \+ j \In Mod.coh W ->
        after i p P ->
        after (i \+ j) (Extend w p)
          (fun v m => exists i' j', 
             [/\ m = i' \+ j', i' \In Mod.coh V, 
                 menv_steps W2 j j' & P v i']).
Proof.
move=>C /(alw_extend C); apply: alw_imp=>{p i C} s q _.
case=>i1 [j1][->{s} Ci1 S1] H v E.
move: E H=>-> [[q'][//]|[_][[<-]] H].
by exists i1, j1; split=>//; apply: H. 
Qed.

End AlwaysExtend.




