Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path rels ordtype.
Require Import finmap Eqdep prelude idynamic pcm unionmap heap codes.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Section Feaps.
Notation K := [ordType of nat].
Variable V : Type.

Structure feap := FUndef | FDef of {finMap K -> V}. 

Definition fupd h x d := nosimpl
  (if h is FDef f then FDef (ins x d f) else FUndef).

Definition fdom h : pred K := nosimpl
  (if h is FDef f then [fun x => x \in supp f] else pred0).

Definition ffree x h : feap := 
  if h is FDef f then FDef (rem x f) else FUndef.

Definition flook (x : K) h := 
  if h is FDef f then fnd x f else None.

Definition fempb h :=
  if h is FDef f then supp f == [::] else false. 

Definition ffresh h := 
  (if h is FDef f then last 0 (supp f) else 0).+1.

Definition ffilter (p : pred nat) h := nosimpl
  (if h is FDef f then FDef (kfilter p f) else FUndef).

End Feaps.

Implicit Arguments FUndef [V].

Lemma feapE V f g : f = g <-> @FDef V f = FDef g.
Proof. by split=>[->|[]]. Qed.

(***********************)
(* monoidal properties *)
(***********************)

Module FeapPCM.
Section FeapPCM.
Notation K := [ordType of nat].
Variable V : Type.
Notation feap := (feap V).

Definition fdef (h : feap) := if h is FDef _ then true else false.

Definition fempty : feap := FDef (nil K V).

Definition funion2 h1 h2 : feap := 
  if (h1, h2) is (FDef f1, FDef f2) then  
    if disj f1 f2 then FDef (fcat f1 f2) else FUndef
  else FUndef.

Local Notation "h1 \+ h2" := (funion2 h1 h2) 
  (at level 43, left associativity).

Lemma funC h1 h2 : h1 \+ h2 = h2 \+ h1.
Proof.
case: h1=>[|h1]; case: h2=>[|h2] //; rewrite /funion2. 
by case: ifP=>E; rewrite disjC E // -feapE fcatC.
Qed.

Lemma funCA h1 h2 h3 : h1 \+ (h2 \+ h3) = h2 \+ (h1 \+ h3).
Proof.
case: h1 h2 h3=>[|h1][|h2][|h3] //.
rewrite /funion2; case E1: (disj h2 h3); last first.
- by case E2: (disj h1 h3)=>//; rewrite disj_fcat E1 andbF.
rewrite disj_fcat andbC.
case E2: (disj h1 h3)=>//.
rewrite disj_fcat (disjC h2) E1 /= andbT.
by case E3: (disj h1 h2)=>//; rewrite -feapE fcatAC // E1 E2 E3.
Qed.

Lemma funA h1 h2 h3 : h1 \+ (h2 \+ h3) = h1 \+ h2 \+ h3.
Proof. by rewrite (funC h2) funCA funC. Qed.

Lemma fun0h h : fempty \+ h = h.
Proof. by case: h=>[|h] //; apply/feapE; rewrite fcat0s. Qed.

Lemma defUnl h1 h2 : fdef (h1 \+ h2) -> fdef h1. 
Proof. by case: h1. Qed.

Lemma def0 : fdef fempty.
Proof. by []. Qed.

Definition pts x v := fupd fempty x v.

End FeapPCM.
End FeapPCM.

Definition feapPCMMixin V := 
  PCMMixin (@FeapPCM.funC V) (@FeapPCM.funA V) 
           (@FeapPCM.fun0h V) (@FeapPCM.defUnl V) (@FeapPCM.def0 V).
Canonical feapPCM V := Eval hnf in PCM (feap V) (feapPCMMixin V).

Notation "x \\\-> v" := (FeapPCM.pts x v) (at level 30).

(* if V is an equality type, then feaps are an equality type too *)

Module FeapEq. 
Section FeapEq. 
Notation K := [ordType of nat].
Variable V : eqType.

Definition feap_eq (h1 h2 : feap V) := 
  match h1, h2 with 
  | FUndef, FUndef => true
  | FDef m1, FDef m2 => m1 == m2
  | _, _ => false
  end. 

Lemma feap_eqP : Equality.axiom feap_eq.
Proof.
case=>[|m1]; case=>[|m2] /=; try by constructor. 
case: eqP=>[->|E]; constructor=>//.
by case.
Qed.

End FeapEq.
End FeapEq.

Definition feap_eqMixin V := EqMixin (@FeapEq.feap_eqP V).
Canonical feap_eqType V := EqType (feap _) (feap_eqMixin V).

(***********)
(* fdom_eq *)
(***********)

(* creating opaque definition *)

Module Type fdomeq_Sig.
Parameter fdom_eq : forall V, feap V -> feap V -> bool. 
Axiom fdomeqP : forall V (f1 f2 : feap V),
        reflect (valid f1 = valid f2 /\ fdom f1 =i fdom f2) (fdom_eq f1 f2).
End fdomeq_Sig.

Module fdomeq_Def : fdomeq_Sig.

Definition fdom_eq V (f1 f2 : feap V) := 
  ((valid f1 == valid f2) && 
   match f1, f2 with
     FDef f1', FDef f2' => supp f1' == supp f2'
   | FUndef, FUndef => true
   | _, _ => false
   end).

Lemma fdomeqP V (f1 f2 : feap V) : 
        reflect (valid f1 = valid f2 /\ fdom f1 =i fdom f2) (fdom_eq f1 f2).
Proof.
case: f1=>[|f1]; case: f2=>[|f2]; rewrite /fdom_eq /fdom /=. 
- by constructor.
- by constructor; case.
- by constructor; case.
case: eqP=>H; constructor; first by rewrite H.
by case=>_ G; apply: H; apply/suppE=>x; rewrite G.
Qed.

Prenex Implicits fdom_eq.

End fdomeq_Def.

Export fdomeq_Def.


Section Lemmas.
Notation K := [ordType of nat].
Variable V : Type.
Notation feap := (feap V).

Implicit Types (h : feap) (v : V).

(*****************)
(* cancelativity *)
(*****************)

Lemma funKhl h h1 h2 : valid (h1 \+ h) -> h1 \+ h = h2 \+ h -> h1 = h2.
Proof.
case: h h1 h2=>[|h]; case=>[|h1]; case=>[|h2] //; 
rewrite /PCM.join /= /FeapPCM.funion2;  
case: ifP=>//; case: ifP=>// D1 D2 _ [E]; apply/feapE.
by apply: fcatsK E; rewrite D1 D2.
Qed.

Lemma funhKl h h1 h2 : valid (h \+ h1) -> h \+ h1 = h \+ h2 -> h1 = h2.
Proof. by rewrite !(joinC h); apply: funKhl. Qed.

Lemma funKhr h h1 h2 : valid (h2 \+ h) -> h1 \+ h = h2 \+ h -> h1 = h2.
Proof. by move=>H1 H2; symmetry in H2; rewrite (funKhl H1 H2). Qed.

Lemma funhKr h h1 h2 : valid (h \+ h2) -> h \+ h1 = h \+ h2 -> h1 = h2.
Proof. by rewrite !(joinC h); apply: funKhr. Qed.

(********************)
(* union and filter *)
(********************)

(* to iterate over unions, define a new function vfilter p h pf *)
(* which takes a proof pf : valid h *)

Lemma ffiltUn p h1 h2 :
        valid (h1 \+ h2) -> 
        ffilter p (h1 \+ h2) = ffilter p h1 \+ ffilter p h2.
Proof.
rewrite /ffilter /PCM.join /= /FeapPCM.funion2.
case: h1=>[|h1]; case: h2=>[|h2] //; case: ifP=>D //= _.
by rewrite kfilt_fcat disj_kfilt.
Qed.

Lemma ffilt0 p : @ffilter V p Unit = Unit.
Proof. by rewrite /ffilter -feapE kfilt_nil. Qed.

Lemma ffilt_pred0 h : valid h -> ffilter pred0 h = Unit.
Proof.
case: h=>[|h] //= _; case: h=>s H /=.
by rewrite -feapE fmapE /= /kfilter' /= filter_pred0.
Qed.

Lemma ffilt_predT h : ffilter predT h = h.
Proof. by case: h=>[|h] //; rewrite /ffilter -feapE kfilter_predT. Qed.

Lemma ffilt_predI p1 p2 h : 
        ffilter (predI p1 p2) h = ffilter p1 (ffilter p2 h).
Proof. by case: h=>[|h] //=; rewrite /ffilter kfilter_predI. Qed.

Lemma ffilt_predU p1 p2 h : 
        ffilter (predU p1 p2) h = ffilter p1 h \+ ffilter (predD p2 p1) h.
Proof.
case: h=>[|h] //=; rewrite /ffilter /PCM.join /= /FeapPCM.funion2.
rewrite (in_disj_kfilt); last by move=>x _; rewrite /= negb_and orbA orbN. 
rewrite -feapE -kfilter_predU.
by apply: eq_in_kfilter=>x _; rewrite /= orb_andr orbN.
Qed.

Lemma eq_in_ffilter p1 p2 h : 
        {in fdom h, p1 =1 p2} -> ffilter p1 h = ffilter p2 h.
Proof.
case: h=>[|h] //=; rewrite /fdom /ffilter -feapE. 
by apply: eq_in_kfilter.
Qed.

(*******)
(* dom *)
(*******)

Lemma fdom0 : @fdom V Unit = pred0. 
Proof. by []. Qed.

Lemma fdom0E h : valid h -> fdom h =i pred0 -> h = Unit. 
Proof.
case: h=>[|h] //= _; rewrite /fdom /supp -feapE fmapE /=. 
case: h; case=>[|kv s] //= _.
by move/(_ kv.1); rewrite inE /= eq_refl. 
Qed.

Lemma fdomU h y d : 
        fdom (fupd h y d) =i 
        [pred x | if x == y then valid h else x \in fdom h].
Proof.
case: h=>[|h] /= x; rewrite !inE /fupd /=. 
- by case: ifP=>// E; rewrite andbF.
by rewrite /fdom supp_ins inE.
Qed.

Lemma fdomPt x v : fdom (x \\\-> v) =i [pred y | (x == y)].
Proof. by move=>y; rewrite fdomU valid_unit !inE /= eq_sym. Qed.

Lemma fdomF h x : 
        fdom (ffree x h) =i 
        [pred y | if x == y then false else y \in fdom h].
Proof.
case: h=>[|h] y /=; rewrite !inE /=; case: ifP=>// E;
by rewrite supp_rem inE /= eq_sym E.
Qed.

Lemma fdomUn h1 h2 :
        fdom (h1 \+ h2) =i 
        [pred x | valid (h1 \+ h2) && (x \in [predU fdom h1 & fdom h2])].
Proof.
case: h1 h2 =>[|h1] // [|h2] // x.
rewrite inE /= !inE /PCM.join /= /FeapPCM.funion2.   
by case: ifP=>//= E; rewrite supp_fcat. 
Qed.

Lemma fdom_valid h x : x \in fdom h -> valid h.
Proof. by case: h. Qed.

Lemma fdom_ffree h x : x \notin fdom h -> ffree x h = h.
Proof. by case: h=>[|h] // E; apply/feapE; apply: rem_supp. Qed.

CoInductive fdom_look_spec h x : bool -> Type := 
| fdom_look_none : flook x h = None -> fdom_look_spec h x false
| fdom_look_some v : flook x h = Some v -> fdom_look_spec h x true.

Lemma fdom_flook h x : fdom_look_spec h x (x \in fdom h).
Proof.
case: h=>[|h]; first by constructor.
rewrite /fdom -topredE /=. 
case: suppP=>[v|] H; first by apply: (fdom_look_some (v:=v)).    
by apply: fdom_look_none.
Qed.

Lemma flook_some h x v : flook x h = Some v -> x \in fdom h.
Proof. by case: fdom_flook=>// ->. Qed.

Lemma flook_none h x : flook x h = None -> x \notin fdom h.
Proof. by case: fdom_flook=>// v ->. Qed.

Prenex Implicits flook_some flook_none.

Lemma fdom_filt p h : fdom (ffilter p h) =i [predI p & fdom h].
Proof.
case: h=>[|h] x; first by rewrite !inE /= andbF. 
by rewrite inE /fdom /= -!topredE /= supp_kfilt mem_filter. 
Qed.

Lemma fdomE (g1 g2 f1 f2 : feap) : 
        valid (f1 \+ g1) -> 
        f1 \+ g1 = f2 \+ g2 ->
        fdom f1 =i fdom f2 -> f1 = f2.
Proof.
move=>D E; case: f1 f2 g1 g2 E D=>[|f1][|f2][|g1][|g2] // E; rewrite E //. 
move: E; rewrite /PCM.join /= /FeapPCM.funion2 /fdom.
case: ifP=>D1 //; case: ifP=>D2 //=.
move/feapE=>E _ E1; apply/feapE/fmapP=>k.
move/(_ k): E1; rewrite -!topredE /= => E1.
case E2: (k \in supp f2) in E1; last first.
- by move/negbT/fnd_supp: E1=>->; move/negbT/fnd_supp: E2=>->.
suff L: forall f s, k \in supp f -> disj f s -> 
  fnd k f = fnd k (fcat f s) :> option V.
- by rewrite (L _ _ E1 D1) (L _ _ E2 D2) E.
move=>f s S; case: disjP=>//; move/(_ _ S)/negbTE.
by rewrite fnd_fcat=>->. 
Qed.

Lemma fdomK (f1 f2 g1 g2 : feap) : 
        valid (f1 \+ g1) -> valid (f2 \+ g2) ->
        fdom (f1 \+ g1) =i fdom (f2 \+ g2) ->
        fdom f1 =i fdom f2 -> fdom g1 =i fdom g2.
Proof.
case: f1 f2 g1 g2=>[|f1][|f2][|g1][|g2] //=.
rewrite /PCM.join /= /FeapPCM.funion2 /fdom /valid /=.
case: disjP=>// D1 _; case: disjP=>// D2 _ E1 E2 x. 
move: {E1 E2} (E2 x) (E1 x).
rewrite -!topredE /= !supp_fcat !inE /= => E.
move: {D1 D2} E (D1 x) (D2 x)=><- /implyP D1 /implyP D2.
case _ : (x \in supp f1) => //= in D1 D2 *.
by move/negbTE: D1=>->; move/negbTE: D2=>->.
Qed.

Lemma ffilt_fdom (f1 f2 : feap) : 
        valid (f1 \+ f2) -> ffilter (fdom f1) (f1 \+ f2) = f1.
Proof.
case: f1=>// f1; case: f2=>// f2; rewrite /PCM.join /= /FeapPCM.funion2 /=. 
case: ifP=>// D _; apply/feapE; rewrite /fdom kfilt_fcat /=.  
have E1: {in supp f1, supp f1 =i predT} by [].
have E2: {in supp f2, supp f1 =i pred0}.
- by move=>x; rewrite disjC in D; case: disjP D=>// D _ /D /negbTE ->. 
rewrite (eq_in_kfilter E1) (eq_in_kfilter E2).
by rewrite kfilter_predT kfilter_pred0 fcats0.
Qed.

(*********)
(* valid *)
(*********)

Lemma fvalidU h x v : valid (fupd h x v) = valid h.
Proof. by case: h. Qed.

Lemma fvalidPt x v : valid (x \\\-> v).
Proof. by []. Qed.

Lemma fvalidF h x : valid (ffree x h) = valid h.
Proof. by case: h. Qed. 

CoInductive fvalidUn_spec h1 h2 : bool -> Type :=
| def_false1 of ~~ valid h1 : fvalidUn_spec h1 h2 false
| def_false2 of ~~ valid h2 : fvalidUn_spec h1 h2 false
| def_false3 x of x \in fdom h1 & x \in fdom h2 : fvalidUn_spec h1 h2 false
| def_true of valid h1 & valid h2 & 
    (forall x, x \in fdom h1 -> x \notin fdom h2) : fvalidUn_spec h1 h2 true.

Lemma fvalidUn : forall h1 h2, fvalidUn_spec h1 h2 (valid (h1 \+ h2)).
Proof.
case=>[|h1][|h2] /=; try by [apply: def_false1 | apply: def_false2].
rewrite /PCM.join /= /FeapPCM.funion2 /=; case: ifP=>E.
- by apply: def_true=>// x H; case: disjP E=>//; move/( _ _ H). 
by case: disjP E=>// x T1 T2 _; apply: (def_false3 (x:=x)).
Qed.

Lemma fvalidFUn h1 h2 x : valid (h1 \+ h2) -> valid (ffree x h1 \+ h2).
Proof.
case: fvalidUn=>// H1 H2 H _.
case: fvalidUn=>//; [by rewrite fvalidF H1|by rewrite H2|].
move=>k; rewrite fdomF inE /=.
by case: ifP=>_ //; move/H; move/negbTE=>->.
Qed.

Lemma fvalidUnF h1 h2 x : valid (h1 \+ h2) -> valid (h1 \+ ffree x h2).
Proof. by rewrite joinC; move/(fvalidFUn x); rewrite joinC. Qed. 

Lemma finvalidE x : ~~ valid x <-> x = @FUndef V.
Proof. by case: x; split. Qed.

Lemma fvalidUnUr h1 h2 x v : 
        x \in fdom h2 -> valid (h1 \+ fupd h2 x v) = valid (h1 \+ h2).
Proof.
move=>D; apply: esym; move: D.
case: fvalidUn; first by [move/finvalidE=>->]; first by [move/finvalidE=>->].
- move=>z H1 H2 _.
  case: fvalidUn=>//; rewrite fvalidU=>D1 D2. 
  by move/(_ z H1); rewrite fdomU !inE /= D2 H2; case: ifP.
move=>D1 D2 H1 H2; case: fvalidUn=>//.
- by rewrite D1.
- by rewrite fvalidU D2.
move=>z; rewrite fdomU !inE /= D2; move/H1=>H3; rewrite (negbTE H3).
by case: ifP H2 H3=>//; move/eqP=>->->.
Qed.

Lemma fvalidUnUl h1 h2 x v : 
        x \in fdom h1 -> valid (fupd h1 x v \+ h2) = valid (h1 \+ h2).
Proof. by move=>D; rewrite -!(joinC h2); apply: fvalidUnUr D. Qed.

Lemma fvalid_filt p h : valid (ffilter p h) = valid h.
Proof. by case: h=>[|h]. Qed.

(***********)
(* fdom_eq *)
(***********)

Lemma fdomeq0E h : fdom_eq h Unit -> h = Unit.
Proof. by case/fdomeqP; apply: fdom0E. Qed.

Lemma fdomeq_refl h : fdom_eq h h.
Proof. by case: fdomeqP=>//; case. Qed.

Hint Resolve fdomeq_refl.

Lemma fdomeq_sym h1 h2 : fdom_eq h1 h2 = fdom_eq h2 h1.
Proof. 
suff L h h' : fdom_eq h h' -> fdom_eq h' h by apply/idP/idP; apply: L.
by case/fdomeqP=>H1 H2; apply/fdomeqP; split.
Qed.

Lemma fdomeq_trans h1 h2 h3 : 
        fdom_eq h1 h2 -> fdom_eq h2 h3 -> fdom_eq h1 h3.
Proof.
case/fdomeqP=>E1 H1 /fdomeqP [E2 H2]; apply/fdomeqP=>//.
by split=>//; [rewrite E1 E2 | move=>x; rewrite H1 H2]. 
Qed.

Lemma fdomeq_validUn h1 h2 h1' h2' : 
        fdom_eq h1 h2 -> fdom_eq h1' h2' -> 
        valid (h1 \+ h1') = valid (h2 \+ h2').
Proof.
have L h h' f : fdom_eq h h' -> valid (h \+ f) -> valid (h' \+ f).
- case/fdomeqP=>E F; case: fvalidUn=>// Vg1 Vf H _; case: fvalidUn=>//.
  - by rewrite -E Vg1. 
  - by rewrite Vf.
  by move=>x; rewrite -F; move/H/negbTE=>->. 
move=>H H'; case X: (valid (h1 \+ h1')); apply/esym. 
- by move/(L _ _ _ H): X; rewrite !(joinC h2); move/(L _ _ _ H').
rewrite fdomeq_sym in H; rewrite fdomeq_sym in H'.
apply/negbTE; apply: contra (negbT X); move/(L _ _ _ H). 
by rewrite !(joinC h1); move/(L _ _ _ H').
Qed.

Lemma fdomeqUn h1 h2 h1' h2' : 
        fdom_eq h1 h2 -> fdom_eq h1' h2' -> 
        fdom_eq (h1 \+ h1') (h2 \+ h2').
Proof.
suff L h h' f : fdom_eq h h' -> fdom_eq (h \+ f) (h' \+ f).
- move=>H H'; apply: fdomeq_trans (L _ _ _ H).
  by rewrite !(joinC h1); apply: fdomeq_trans (L _ _ _ H').
move=>F; case/fdomeqP: (F)=>E H.
apply/fdomeqP; split; first by apply: fdomeq_validUn F _.
move=>x; rewrite !fdomUn !inE /= !inE /=.
by rewrite (fdomeq_validUn F (fdomeq_refl f)) H.
Qed.

Lemma fdomeqK h1 h2 h1' h2' : 
        valid (h1 \+ h1') ->
        fdom_eq (h1 \+ h1') (h2 \+ h2') ->
        fdom_eq h1 h2 -> fdom_eq h1' h2'.
Proof.
move=>V1 /fdomeqP [E1 H1] /fdomeqP [E2 H2]; move: (V1); rewrite E1=>V2.
apply/fdomeqP; split; first by rewrite (validR V1) (validR V2). 
by apply: fdomK V1 V2 H1 H2.
Qed.

Lemma fdomeqPt v1 v2 x : fdom_eq (x \\\-> v1) (x \\\-> v2).
Proof. by apply/fdomeqP. Qed.

(**********)
(* update *)
(**********)

Lemma fupd_inj h x d1 d2 : 
        valid h -> fupd h x d1 = fupd h x d2 -> d1 = d2.
Proof.
case: h=>[|h] // _ [E]. 
have: fnd x (ins x d1 h) = fnd x (ins x d2 h) by rewrite E. 
by rewrite !fnd_ins eq_refl; case.
Qed. 

Lemma feap_etaV h x : 
        x \in fdom h -> 
        exists v, flook x h = Some v /\ h = fupd (ffree x h) x v.
Proof.
case: h=>[|h] //; case: fdom_flook=>// v H _; exists v; split=>//.
apply/feapE; apply/fmapP=>k; rewrite fnd_ins.
case: ifP=>[|E]; last by rewrite fnd_rem E.
by move/eqP=>->.
Qed.

Lemma fupdU h x y d1 d2 : 
        fupd (fupd h x d1) y d2 = 
        if x == y then fupd h x d2 else fupd (fupd h y d2) x d1.
Proof.
case: h =>[|h]; rewrite /fupd; case: ifP=>// H; apply/feapE.
- by rewrite ins_ins (eqP H) eq_refl. 
by rewrite ins_ins eq_sym H.
Qed.

Lemma fupdF h x y d : 
        fupd (ffree x h) y d = 
        if x == y then fupd h x d else ffree x (fupd h y d).
Proof.
case: h=>[|h] /=; case: ifP=>E //; rewrite /fupd; last first.
- by apply/feapE; rewrite ins_rem eq_sym E. 
by apply/feapE; rewrite ins_rem (eqP E) eq_refl.
Qed.

Lemma fupdUnl h1 h2 x d : 
        fupd (h1 \+ h2) x d = 
        if x \in fdom h1 then fupd h1 x d \+ h2 else h1 \+ fupd h2 x d.
Proof.
case: h1 h2=>[|h1][|h2] //; case: ifP=>H //; 
rewrite /fupd /PCM.join /= /FeapPCM.funion2; case: ifP=>D //.
- rewrite disjC disj_ins disjC D.
  case: disjP D=>//; move/(_ _ H)=>H4 _; rewrite H4.
  by apply/feapE; rewrite fcat_inss.
- by rewrite disjC disj_ins disjC D andbF.
- rewrite disj_ins D H /=; apply/feapE.
  rewrite (@fcatC _ _ h1) // (@fcatC _ _ h1).
  - by rewrite fcat_inss // H.
  by rewrite disj_ins H D.
by rewrite disj_ins D andbF.
Qed.

Lemma fupdUnr h1 h2 x d : 
        fupd (h1 \+ h2) x d = 
        if x \in fdom h2 then h1 \+ fupd h2 x d else fupd h1 x d \+ h2.
Proof. by rewrite joinC fupdUnl (joinC h1) (joinC h2). Qed.

Lemma fpts_injV x v1 v2 : x \\\-> v1 = x \\\-> v2 -> v1 = v2.
Proof. by case. Qed.

Lemma filt_fupd p x d h : 
         ffilter p (fupd h x d) = 
         if p x then fupd (ffilter p h) x d else ffilter p h.
Proof. 
by case: h=>[|h]; case: ifP=>//; rewrite /ffilter /= kfilt_ins => ->. 
Qed.

(********)
(* free *)
(********)

Lemma ffree0 x : @ffree V x Unit = Unit.
Proof. by apply/feapE; rewrite rem_empty. Qed.

Lemma ffreeU h x y d : 
        ffree x (fupd h y d) = 
        if x == y then ffree x h
        else fupd (ffree x h) y d.
Proof.
case: h=>[|h] /=; first by case: ifP.
by rewrite /fupd; case: ifP=>E1; rewrite rem_ins E1.
Qed.

Lemma ffreeF h x y : 
        ffree x (ffree y h) = if x == y then ffree x h else ffree y (ffree x h).
Proof. 
by case: h=>*; case: ifP=>E //; apply/feapE; rewrite rem_rem E. 
Qed.

Lemma ffreeUn h1 h2 x : 
        ffree x (h1 \+ h2) = 
        if x \in fdom (h1 \+ h2) then ffree x h1 \+ ffree x h2 
        else h1 \+ h2.
Proof.
case: h1 h2=>[|h1] [|h2] //; rewrite /PCM.join /= /FeapPCM.funion2.
rewrite /ffree /fdom /=.
case: ifP=>E1 //; rewrite supp_fcat inE /=.
case: ifP=>E2; last first.
- by apply/feapE; rewrite rem_supp // supp_fcat inE /= E2.
rewrite disj_rem; last first.
- by rewrite disjC disj_rem // disjC.
apply/feapE; case/orP: E2=>E2.
- suff E3: x \notin supp h2 
    by rewrite -fcat_rems // (rem_supp E3).
  by case: disjP E1 E2=>// H _; move/H.
suff E3: x \notin supp h1 
  by rewrite -fcat_srem // (rem_supp E3).
by case: disjP E1 E2=>// H _; move/contra: (H x); rewrite negbK.
Qed.

Lemma ffreeUnD h1 h2 x :
        valid (h1 \+ h2) -> ffree x (h1 \+ h2) = ffree x h1 \+ ffree x h2.
Proof.
move=>D; rewrite ffreeUn fdomUn D !inE /=; case: ifP=>//.
by move/negbT; rewrite negb_or; case/andP=>H1 H2; rewrite !fdom_ffree.
Qed.

Lemma ffreeUnl h1 h2 x : 
        x \notin fdom h1 -> ffree x (h1 \+ h2) = h1 \+ ffree x h2.
Proof.
move=>D1; rewrite ffreeUn fdomUn !inE (negbTE D1) /=.
case: ifP; first by case/andP; rewrite fdom_ffree.
move/negbT; rewrite negb_and; case/orP=>D2; last by rewrite fdom_ffree.
suff: ~~ valid (h1 \+ ffree x h2).
- by case: (h1 \+ ffree x h2)=>// _; case: (h1 \+ h2) D2.
apply: contra D2; case: fvalidUn=>// H1 H2 H _.
case: fvalidUn=>//; first by [rewrite H1]; 
  first by move: H2; rewrite fvalidF=>->.
move=>k H3; move: (H _ H3); rewrite fdomF inE /=.
by case: ifP H3 D1=>[|_ _ _]; [move/eqP=><- -> | move/negbTE=>->].
Qed.

Lemma ffreeUnr h1 h2 x : 
        x \notin fdom h2 -> ffree x (h1 \+ h2) = ffree x h1 \+ h2.
Proof. by move=>H; rewrite joinC ffreeUnl // joinC. Qed.

Lemma ffree_filt p x h : 
        ffree x (ffilter p h) = 
        if p x then ffilter p (ffree x h) else ffilter p h.
Proof.
rewrite /ffree /ffilter; case: h=>[|h]; case: ifP=>//= E.
- by rewrite kfilt_rem E. 
by move: (rem_kfilt p x h); rewrite E=><-. 
Qed.

Lemma filt_ffree p x h : 
        ffilter p (ffree x h) = 
        if p x then ffree x (ffilter p h) else ffilter p h.
Proof. by case: h=>[|h]; case: ifP=>// E; rewrite /ffilter /= kfilt_rem E. Qed.

(**********)
(* lookup *)
(**********)

Lemma flookU h x y d : 
        flook x (fupd h y d) = if x == y then
          if valid h then Some d else None
        else flook x h.
Proof. by case: h=>[|h] /=; case: ifP=>E //; rewrite fnd_ins E. Qed.

Lemma flookF h x y : 
        flook x (ffree y h) = if x == y then None else flook x h.
Proof. 
by case: h=>[|h]; case: ifP=>E //; rewrite /flook /ffree fnd_rem E. 
Qed.

Lemma flookUnl h1 h2 x : 
        valid (h1 \+ h2) -> 
        flook x (h1 \+ h2) = if x \in fdom h1 then flook x h1 else flook x h2.
Proof.
case: h1 h2=>[|h1] // [|h2] //.
rewrite /flook /fdom /PCM.join /= /FeapPCM.funion2.
case: ifP=>D //= _; case: ifP=>E1; last first.
- rewrite fnd_fcat; case: ifP=>// E2.
  by rewrite fnd_supp ?E1 // fnd_supp ?E2.
suff E2: x \notin supp h2 by rewrite fnd_fcat (negbTE E2).
by case: disjP D E1=>// H _; apply: H.
Qed.

Lemma flookUnr h1 h2 x : 
        valid (h1 \+ h2) ->
        flook x (h1 \+ h2) = if x \in fdom h2 then flook x h2 else flook x h1.
Proof. by rewrite joinC=>D; rewrite flookUnl. Qed.

Lemma flook_eta h1 h2 :  
        valid h1 -> valid h2 ->
        (forall x, flook x h1 = flook x h2) -> h1 = h2.
Proof. by case: h1 h2=>[|h1] [|h2] //= _ _ H; apply/feapE; apply/fmapP. Qed.

Lemma flook_filt p h x : 
        flook x (ffilter p h) = 
        if p x then flook x h else None.
Proof. by case: h=>[|h]; case: ifP=>E //=; rewrite fnd_kfilt E. Qed.

(********)
(* empb *)
(********)

Lemma fempP h : reflect (h = Unit) (fempb h).
Proof.
case: h=>[|h] /=; first by right.
case: eqP=>E; first by apply: ReflectT; apply/feapE; apply/supp_nilE.
by apply: ReflectF; move/feapE=>S; rewrite S supp_nil in E.
Qed.

Lemma fempU h x d : fempb (fupd h x d) = false. 
Proof. 
case: h=>[|h] //; rewrite /fupd /fempb.
suff: x \in supp (ins x d h) by case: (supp _).
by rewrite supp_ins inE /= eq_refl.
Qed. 

Lemma fempPt x v : fempb (x \\\-> v) = false.
Proof. by rewrite fempU. Qed.

Lemma fempUn h1 h2 : fempb (h1 \+ h2) = fempb h1 && fempb h2.
Proof.
case: h1 h2=>[|h1] // [|h2] /=; first by rewrite andbC.
rewrite /fempb /PCM.join /= /FeapPCM.funion2; case: ifP=>E;
case: eqP=>E1; case: eqP=>E2 //=; last 2 first.
- by move: E2 E1; move/supp_nilE=>->; rewrite fcat0s; case: eqP.
- by move: E1 E2 E; do 2![move/supp_nilE=>->]; case: disjP.
- by move/supp_nilE: E2 E1=>-> <-; rewrite fcat0s eq_refl.
have [k H3]: exists k, k \in supp h1.
- case: (supp h1) {E1 E} E2=>[|x s _] //.
  by exists x; rewrite inE eq_refl.
suff: k \in supp (fcat h1 h2) by rewrite E1.
by rewrite supp_fcat inE /= H3.
Qed.

(* some transformation lemmas *)

Lemma fempbE h : h = Unit <-> fempb h.
Proof. by split=>[-> //|]; case: fempP. Qed.

Lemma fun0E h1 h2 : h1 \+ h2 = Unit <-> h1 = Unit /\ h2 = Unit.
Proof. by rewrite !fempbE fempUn; case: andP. Qed.

Lemma fvalidE h : reflect (valid h /\ forall x, x \notin fdom h) (fempb h).
Proof.
case: fempP=>W; constructor; first by rewrite W.
case=>D E; case: W; case: h D E=>// f H; rewrite /fdom => E.
apply/feapE; apply/supp_nilE.
by case: (supp f) E=>// x s; move/(_ x); rewrite inE eq_refl.
Qed.

Lemma fvalidUnhh h : valid (h \+ h) = fempb h. 
Proof.
case E: (fempb h); first by move/fempbE: E=>->.
case: fvalidUn=>// D _ L.
case: fvalidE E=>//; case; split=>// x.
case E: (x \in fdom h)=>//.
by move: (L x E); rewrite E.
Qed.

Lemma femp_filt p h : fempb h -> fempb (ffilter p h).
Proof. by case: h=>[|h] //= /eqP/supp_nilE ->; rewrite kfilt_nil. Qed.

(*********)
(* fresh *)
(*********)

Lemma fpath_last n s x : path ord x s -> ord x (last x s + (n+1)).
Proof.
move: n s x.
suff L: forall s x, path ord x s -> ord x (last x s + 1).
- elim=>[|n IH] // s x; move/IH=>E; apply: trans E _. 
  by rewrite /ord /= addSn [_ + (n + 1).+1]addnS.
elim=>[|y s IH x] /=; first by case=>x; rewrite /ord /= addn1.
by case/andP=>H1; move/IH; apply: trans H1.
Qed.

Lemma fpath_filter (A : ordType) (s : seq A) (p : pred A) x : 
        path ord x s -> path ord x (filter p s).
Proof.
elim: s x=>[|y s IH] x //=.
case/andP=>H1 H2.
case: ifP=>E; first by rewrite /= H1 IH.
apply: IH; elim: s H2=>[|z s IH] //=.
by case/andP=>H2 H3; rewrite (@trans _ y).
Qed.

Lemma fdom_fresh h n : (ffresh h) + n \notin fdom h.
Proof.
suff L2: forall h x, x \in fdom h -> ord x (ffresh h).
- by apply: (contra (L2 _ _)); rewrite -leqNgt leq_addr. 
case=>[|[s]] //; rewrite /supp => /= H2 x.
rewrite /fdom /ffresh /supp -topredE /=.  
elim: s 0 H2 x=>[|[y d] s IH] //= m.
move=>H z; rewrite inE /=.
case/orP; last by apply: IH; apply: path_sorted H.
by move/eqP=>->{z}; rewrite -2!addn1; apply: fpath_last H.
Qed.

Lemma ffresh_null h : ffresh h != 0.
Proof. by []. Qed.

Opaque ffresh.

Hint Resolve fdom_fresh ffresh_null.

(***********************)
(* Some derived lemmas *)
(***********************)

Lemma fvalidPtUn h x v :
        valid (x \\\-> v \+ h) = (valid h) && (x \notin fdom h).
Proof. 
case: fvalidUn=>//; last 1 first.
- by move=>_ -> H; rewrite (H x) // fdomPt inE.
- by move/negbTE=>->.
by move=>y; rewrite fdomPt inE /=; move/eqP=><-->; rewrite andbF.
Qed.

(* the projections from defPtUn are often useful *)

Lemma fvalidPt_def h x v : valid (x \\\-> v \+ h) -> valid h.
Proof. by rewrite fvalidPtUn; case/andP. Qed.

Lemma fvalidPt_dom h x v : valid (x \\\-> v \+ h) -> x \notin fdom h.
Proof. by rewrite fvalidPtUn; case/andP. Qed.

(* now dom *)

Lemma fdomPtUn h x v : 
        fdom (x \\\-> v \+ h) =i 
        [pred y | valid (x \\\-> v \+ h) && ((x == y) || (y \in fdom h))].
Proof. by move=>y; rewrite fdomUn !inE !fvalidPtUn /= eq_sym. Qed.

(* look and free *)

Lemma flookPtUn h x v : 
        valid (x \\\-> v \+ h) -> flook x (x \\\-> v \+ h) = Some v.
Proof. by move=>D; rewrite flookUnl // flookU fdomPt !inE eq_refl. Qed.

Lemma ffreePtUn h x v : 
        valid (x \\\-> v \+ h) -> ffree x (x \\\-> v \+ h) = h.
Proof.
move=>D; rewrite ffreeUnr; last by rewrite (fvalidPt_dom D).
by rewrite ffreeU eqxx ffree0 unitL.
Qed.

Lemma fupdPtUn x i v1 v2 : 
        fupd (x \\\-> v1 \+ i) x v2 = x \\\-> v2 \+ i.
Proof.
case E1: (valid (x \\\-> v1 \+ i)).
- by rewrite fupdUnl fdomPt inE fupdU eq_refl.
have E2: valid (x \\\-> v2 \+ i) = false by rewrite !fvalidPtUn in E1 *.
by case: (_ \+ _) E1=>// _; case: (_ \+ _) E2. 
Qed.

Lemma feap_etaPV h x : 
        x \in fdom h -> exists d, flook x h = Some d /\ 
        h = x \\\-> d \+ ffree x h.
Proof.
move=>H1; case/feap_etaV: (H1)=>d [H2 E]; exists d. 
split=>//; rewrite {1}E -{1}[ffree x h]unitL fupdUnr. 
by rewrite fdomF inE /= eq_refl.
Qed.

Lemma fcancelV h1 h2 x v1 v2 : 
        valid (x \\\-> v1 \+ h1) -> 
        x \\\-> v1 \+ h1 = x \\\-> v2 \+ h2 -> [/\ v1 = v2, valid h1 & h1 = h2].
Proof.
move=>D E.
have: flook x (x \\\-> v1 \+ h1) = flook x (x \\\-> v2 \+ h2) by rewrite E.
rewrite !flookPtUn -?E //; case=>X. 
by rewrite -{}X in E *; rewrite -(funhKl D E) (validR D).
Qed.

Lemma fdomPtUnX v x i : valid (x \\\-> v \+ i) -> x \in fdom (x \\\-> v \+ i).
Proof. by move=>D; rewrite fdomPtUn inE /= D eq_refl. Qed.

Lemma fdomPtX v x : x \in fdom (x \\\-> v).
Proof. by rewrite -(unitR (x \\\-> v)) fdomPtUnX. Qed.

Lemma fcancelPtV x v1 v2 : x \\\-> v1 = x \\\-> v2 -> v1 = v2.
Proof.
rewrite -[x \\\-> v1]unitR -[x \\\-> v2]unitR.
by move/fcancelV; rewrite unitR fvalidPt; case/(_ (erefl _)). 
Qed.

Lemma ffiltPtV p x v : 
        ffilter p (x \\\-> v) = if p x then x \\\-> v else Unit.
Proof. 
by case: ifP=>E; apply/feapE; apply/fmapE; rewrite /= /kfilter' /= E. 
Qed.

(* induction over feaps, using union *)

Lemma feap_ind' (P : feap -> Prop) : 
         P FUndef -> P Unit ->
         (forall h x v, P h -> valid (x \\\-> v \+ h) -> P (x \\\-> v \+ h)) ->
         forall h, P h.
Proof.
move=>H1 H2 H3; apply: feap_ind=>//; apply: fmap_ind'=>// x v s H4 H5.
set h := FDef s in H5.
have N : x \notin supp s by apply: notin_path H4.
have: valid (x \\\-> v \+ h) by rewrite fvalidPtUn.
move/(H3 h x v H5).
rewrite /PCM.join /= /FeapPCM.funion2 /disj /= N /=. 
rewrite (_ : FinMap (sorted_ins' x v (sorted_nil K V)) = ins x v (@nil K V)) //.
by rewrite fcat_inss // fcat0s. 
Qed.

End Lemmas.

Prenex Implicits flook_some flook_none.

Hint Resolve fdomeq_refl fdomeqPt.

(**********************************)
(* Mapping and folding over feaps *)
(**********************************)

Section Map.
Notation K := [ordType of nat].
Variables (U V : Type) (f : U -> V). 

Definition fmap (fp : feap U) : feap V := 
  if fp is FDef m then FDef (mapf f m) else FUndef.

End Map.

(* flattening is folding whose range is a pcm *)
(* we will use it to flatten into heaps, or collect *)
(* validity into bool *)

Section Folding.
Notation K := [ordType of nat].
Variables (V : Type) (U : pcm).
Variable to_U : V -> U.
Notation finmap := ({finMap K -> V}).

Fixpoint seq_flat (s : seq (K * V)) : U := 
  if s is (k, v)::s' then to_U v \+ seq_flat s' else Unit.

Definition map_flat (m : finmap) := seq_flat (seq_of m).

Lemma map_flat_ins f k v : 
        k \notin supp f -> map_flat (ins k v f) = to_U v \+ map_flat f.
Proof.
case: f=>s S; rewrite /supp /map_flat /= {S}.
elim: s=>[|[k1 v1] s IH] //=; rewrite inE /= negb_or eq_sym.
by case: totalP=>//= O N; rewrite IH // joinCA.
Qed.

Lemma map_flat_fcat f1 f2 : 
       disj f1 f2 -> map_flat (fcat f1 f2) = map_flat f1 \+ map_flat f2.
Proof.
elim/fmap_ind': f1=>[|k v f1 S IH]; first by rewrite fcat0s unitL. 
rewrite disjC disj_ins disjC; case/andP=>S2 D.
have S1: k \notin supp f1 by rewrite notin_path.
have S3 : k \notin supp (fcat f1 f2) by rewrite supp_fcat inE /= negb_or S1. 
by rewrite fcat_inss // !map_flat_ins // IH // joinA. 
Qed.

End Folding.

(* in the particular case of flattening, we fold into a heap *)

Section Flattening.
Notation K := [ordType of nat].
Variable V : Type.
Variable to_heap : V -> heap.
Notation finmap := ({finMap K -> V}).


Definition fflat f := if f is FDef m then map_flat to_heap m else um_undef.

Lemma fflat_join f1 f2 : 
        fflat (f1 \+ f2) = 
        if valid (f1 \+ f2) then fflat f1 \+ fflat f2 else um_undef.
Proof.
case: f1 f2=>//; rewrite /fflat.   
elim/fmap_ind'=>[|k v s S _] f2; first by rewrite !unitL; case: f2.
case: ifP; last by move/negbT/finvalidE=>->. 
case: f2=>// f2; rewrite /PCM.join /= /FeapPCM.funion2. 
by case: ifP=>// D _; rewrite map_flat_fcat.
Qed.

End Flattening.

(* if we want to collect the validity of all pcm leafs *)
(* we fold into bool *)

Definition to_bool (v : (idynamic (to_type \o pcm_sort))) := valid (idyn_val v).

Module Type allvalid_Sig.
Parameter allvalid : feap (idynamic (to_type \o pcm_sort)) -> bool. 

Axiom allvalidE : forall h, allvalid h = 
         if h is FDef m then map_flat to_bool m else false.
End allvalid_Sig.

Module allvalid_fDef : allvalid_Sig.
Definition allvalid f := if f is FDef m then map_flat to_bool m else false.

Definition allvalidE f := erefl (allvalid f). 
End allvalid_fDef.

Export allvalid_fDef.

Lemma allvalid_join f1 f2 : 
        allvalid (f1 \+ f2) = 
        if valid (f1 \+ f2) then allvalid f1 \+ allvalid f2 
        else false.
Proof.
rewrite !allvalidE; case: f1 f2=>//; rewrite /fflat.   
elim/fmap_ind'=>[|k v s S _] f2; first by rewrite !unitL; case: f2.
case: ifP; last by move/negbT/finvalidE=>->. 
case: f2=>// f2; rewrite /PCM.join /= /FeapPCM.funion2. 
by case: ifP=>// D _; apply: map_flat_fcat.
Qed.

(*********************)
(* Now dynamic feaps *)
(*********************)

(* so far, the whole development was generic in the range of the feap *)
(* it is only now that we make the range idynamic *)

Section DynFeapDefs.
Variable C : code.
Implicit Type U : encoded C. 

Definition pts_encode U x (v : to_type U) := FeapPCM.pts x (code_dyn v).

Definition weq (f1 f2 : feap (idynamic (to_type \o @encode C))) := 
  fmap idyn_tp f1 == fmap idyn_tp f2.

End DynFeapDefs.

Implicit Arguments weq [].

Notation "f1 =w f2" := (@weq _ f1 f2) (at level 50).

Lemma weqUn (C : code) (f1 f2 f1' f2' : feap (idynamic (to_type \o @encode C))) : 
        f1 =w f1' -> f2 =w f2' -> f1 \+ f2 =w f1' \+ f2'.
Proof.
case: f1 f1' f2 f2'=>[|f1][|f1'][|f2][|f2'] //= /eqP/feapE E1 /eqP/feapE E2.
rewrite /weq /= /PCM.join /= /FeapPCM.funion2 /fmap; apply/eqP.

(*
rewrite (mapf_disj E1 E2); case: ifP=>// H; apply/feapE. 
by rewrite mapf_fcat E1 E2 -mapf_fcat.
*)
admit.
Qed.

Lemma weqPt (C : code) (U : encoded C) x (v1 v2 : to_type U) : 
        pts_encode x v1 =w pts_encode x v2. 
Proof. by apply/eqP/feapE/fmapE. Qed.

Hint Resolve weqPt.

(* A structure adding a flattening function to a code *)

Module FlatCode.

Structure mixin_of (C : code) := Mixin {
  to_heap_op : idynamic (to_type \o @encode C) -> heap}.

Section ClassDef.

Notation class_of := mixin_of (only parsing).

Structure type : Type := Pack {sort : code; _ : class_of sort; _ : code}.
Local Coercion sort : type >-> code. 

Variables (v : code) (cT : type).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.

Definition pack c := @Pack v c v.

Definition clone := fun c & cT -> v & phant_id (pack c) cT => pack c. 

Definition to_heap := to_heap_op class.

End ClassDef.

Module Exports.
Coercion sort : type >-> code.
Notation flatcode := type.
Notation flatcodeMixin C := (@Mixin C).
Notation FlatCode C m := (@pack C m).
Notation to_heap := to_heap.

Notation "[ 'flatcodeMixin' 'of' T ]" := (class _ : mixin_of T)
  (at level 0, format "[ 'flatcodeMixin'  'of'  T ]") : form_scope.
Notation "[ 'flatcode' 'of' T 'for' C ]" := (@clone T C _ idfun id)
  (at level 0, format "[ 'flatcode'  'of'  T  'for'  C ]") : form_scope.
Notation "[ 'flatcode' 'of' T ]" := (@clone T _ _ id id)
  (at level 0, format "[ 'flatcode'  'of'  T ]") : form_scope.

End Exports.

End FlatCode.

Export FlatCode.Exports.

(* specific functions for pcms and types *)

Lemma flatten_pcmP (v : idynamic (to_type \o pcm_sort)) :
        idyn_tp v == pcm_heap -> pcm_sort (idyn_tp v) = heap :> Type.
Proof. by move/eqP=>->. Qed.

Definition flatten_pcm (v : idynamic (to_type \o pcm_sort)) := 
  if @decP _ (idyn_tp v == pcm_heap) idP is
    left pf then icoerce id (idyn_val v) (flatten_pcmP pf)
  else Unit.

Definition pcmFlatCodeMixin := flatcodeMixin pcmCode flatten_pcm.
Canonical pcmFlatCode := FlatCode pcmCode pcmFlatCodeMixin.

Lemma flatten_tpP (v : idynamic (to_type \o tp_sort)) : 
        idyn_tp v == tp_heap -> tp_sort (idyn_tp v) = heap :> Type.
Proof. by move/eqP=>->. Qed.

Definition flatten_tp (v : idynamic (to_type \o tp_sort)) : heap := 
  if @decP _ (idyn_tp v == tp_heap) idP is
    left pf then coerce (T:=id) (idyn_val v) (flatten_tpP pf)
  else Unit.

Definition typeFlatCodeMixin := flatcodeMixin typeCode flatten_tp.
Canonical typeFlatCode := FlatCode typeCode typeFlatCodeMixin.

(* flattening, made opaque *)
Module Type flat_Sig.
Parameter flat : forall C : flatcode, 
  feap (idynamic (to_type \o (@encode C))) -> heap. 
Axiom flatE : forall C, @flat C = fflat (@to_heap C).
End flat_Sig.

Module flat_Def : flat_Sig.
Definition flat C := fflat (@to_heap C). 
Definition flatE C := erefl (fflat (@to_heap C)). 
Prenex Implicits flat.
End flat_Def.

Export flat_Def.

(* that's it with definitions; now the lemmas *)

Notation "x /-> v" := (@pts_encode _ _ x v) (at level 30).
Notation "x \-> v" := (@pts_encode pcmCode _ x v) (at level 30).
Notation "x ::-> v" := (@pts_encode typeCode _ x v) (at level 30).

Section DynFeapLemmas.
Notation K := [ordType of nat]. 
Variable C : code.
Implicit Type U : encoded C. 

Lemma fpts_injT U1 U2 x1 x2 (v1 : to_type U1) (v2 : to_type U2) : 
        x1 /-> v1 = x2 /-> v2 -> decode U1 = decode U2. 
Proof. by case. Qed.

Lemma fpts_inj U x (v1 v2 : to_type U) : x /-> v1 = x /-> v2 -> v1 = v2.
Proof. 
case=>H; move: (esym _) (inj_pairT2 _ _ _ _ _ H) => /= {H}.   
by rewrite encdecE=>pf; rewrite !ieqc.
Qed.

Lemma feap_etaPt U h x (v : to_type U) : 
        flook x h = Some (code_dyn v) -> h = x /-> v \+ ffree x h.
Proof.
move=>H; move/flook_some: (H); case/feap_etaPV=>d []. 
by rewrite H; case=><-. 
Qed.

Lemma fcancelT U1 U2 h1 h2 x (v1 : to_type U1) (v2 : to_type U2) : 
        valid (x /-> v1 \+ h1) -> 
        x /-> v1 \+ h1 = x /-> v2 \+ h2 -> decode U1 = decode U2.
Proof.
move=>D E.
have: flook x (x /-> v1 \+ h1) = flook x (x /-> v2 \+ h2) by rewrite E.
by rewrite !flookPtUn -?E //; case.
Qed.

Lemma fcancel U h1 h2 x (v1 v2 : to_type U) : 
        valid (x /-> v1 \+ h1) -> 
        x /-> v1 \+ h1 = x /-> v2 \+ h2 -> [/\ v1 = v2, valid h1 & h1 = h2].
Proof. 
move=>D; case/(fcancelV D); move/idyn_inj; move: (esym _)=>/=.
by rewrite encdecE=>pf; rewrite !ieqc.
Qed.

Lemma fcancelTPt U1 U2 x (v1 : to_type U1) (v2 : to_type U2) : 
        x /-> v1 = x /-> v2 -> decode U1 = decode U2.
Proof.
rewrite -[x /-> v1]unitR -[x /-> v2]unitR.
by move/fcancelT; rewrite unitR fvalidPt; apply.  
Qed.

Lemma fcancelPt U x (v1 v2 : to_type U) : 
        x /-> v1 = x /-> v2 -> v1 = v2. 
Proof.
rewrite -[x /-> v1]unitR -[x /-> v2]unitR; move/fcancel.
by rewrite unitR fvalidPt; case/(_ (erefl _)).
Qed.

Lemma ffiltPt U p x (v : to_type U) : 
        ffilter p (x /-> v) = 
        if p x then x /-> v else Unit.
Proof. by rewrite ffiltPtV. Qed.

(* induction principle *)

Lemma dynfeap_ind (P : feap (idynamic (to_type \o encode)) -> Prop) : 
        P FUndef -> P Unit -> 
        (forall U h x (u : to_type U), 
           P h -> valid (x /-> u \+ h) -> P (x /-> u \+ h)) ->
        forall h, P h.
Proof.
move=>H1 H2 H3; apply: feap_ind'=>// h x v.
move/(H3 _ _ x (code_val v)). 
by rewrite /pts_encode -code_eta.
Qed.

End DynFeapLemmas.

(***************)
(* And Zipping *)
(***************)

(* with dynamics, have to check if the types of zipped components match *)

Module ZipLemma.
Section Lemmas.
Notation K := [ordType of nat].
Notation sort := (to_type \o pcm_sort).
Notation V := (@idynamic pcm_code sort). 
Notation idyn_tp := (@idyn_tp _ sort). 
Notation finmap := ({finMap K -> V}).

Definition zip_f v1 v2 : option V :=
  if @decP _ (idyn_tp v1 == idyn_tp v2) idP is left pf then
    Some (idyn (A := idyn_tp v1) 
         (idyn_val v1 \+ 
          icoerce sort (idyn_val v2) (esym (eqP pf))))
  else None.

Definition unit_f v : V := idyn (A := idyn_tp v) Unit.

Lemma zip_comm_f : commutative zip_f.
Proof. 
case=>tpx x; case=>tpy y; rewrite /zip_f /=. 
case: decP=>E1; case: decP=>E2 //; last first. 
- by rewrite eq_sym E2 in E1.
- by rewrite eq_sym E1 in E2.
congr Some; move: (eqP E1) (eqP E2).
rewrite -{E1 E2}(eqP E1) in y * => pf1 pf2.
by rewrite !ieqc joinC.
Qed.

Lemma zip_assoc_f x y z :
        obind (zip_f x) (zip_f y z) = obind (zip_f^~ z) (zip_f x y).
Proof.
case: x y z=>tpx x [tpy y] [tpz z]; rewrite /zip_f /=. 
case: decP=>E1 //=; case: decP=>E2 //=; case: decP=>E3 //=; last first. 
- by rewrite -(eqP E2) E3 in E1.
- by rewrite (eqP E2) E1 in E3.
congr Some; move: (eqP E1) (eqP E2) (eqP E3). 
rewrite -(eqP E2) -{E1 E2 E3}(eqP E3) in y z * => pf1 pf2 pf3.
by rewrite !ieqc joinA.
Qed.

Lemma zip_unitL_f x : zip_f (unit_f x) x = Some x.
Proof.
case: x=>tpx x; rewrite /zip_f /=.
case: decP=>// E; congr Some; move: (eqP E)=>pf.
by rewrite ieqc unitL.
Qed.

Lemma zip_unitE_f x y : 
        (exists z, zip_f x y = Some z) <-> unit_f x = unit_f y.
Proof.
case: x y=>tpx x [tpy y]; rewrite /zip_f /unit_f /=. 
split.
- case; case=>tpz z; case: decP=>// E. 
  by move: (eqP E); rewrite -{E}(eqP E) in y *.
move/idyn_injT=>/= E; rewrite -{tpy}E in y *.
exists (@idyn _ sort _ (x \+ y)).
by case: decP=>// E; rewrite ieqc.
Qed.

End Lemmas.
End ZipLemma.

(* zip and zipunit implemented opaquely *)
Module Type zip_Sig.
Local Notation K := [ordType of nat].
Local Notation sort := (to_type \o pcm_sort).
Local Notation V := (@idynamic pcm_code sort). 

Parameter zip : feap V -> feap V -> feap V. 
Parameter zip_unit : feap V -> feap V.

Axiom zip_E : forall f1 f2, zip f1 f2 = 
  if (f1, f2) is (FDef s1, FDef s2) then 
     if finmap.zip ZipLemma.zip_f s1 s2 is Some s then FDef s
     else FUndef
   else FUndef.

Axiom zip_unit_E : forall f, zip_unit f = 
  if f is FDef s then FDef (finmap.zip_unit ZipLemma.zip_unitL_f s)
   else FUndef.

End zip_Sig.

Module zip_Def : zip_Sig.
Local Notation K := [ordType of nat].
Local Notation sort := (to_type \o pcm_sort).
Local Notation V := (@idynamic pcm_code sort). 
Local Notation idyn_tp := (@idyn_tp _ sort). 
Local Notation finmap := ({finMap K -> V}).

Definition zip (f1 f2 : feap V) := 
  if (f1, f2) is (FDef s1, FDef s2) then 
     if finmap.zip ZipLemma.zip_f s1 s2 is Some s then FDef s
     else FUndef
  else FUndef.

Definition zip_unit f := 
  if f is FDef s then FDef (finmap.zip_unit ZipLemma.zip_unitL_f s)
  else FUndef.

Definition zip_E f1 f2 := erefl (zip f1 f2).
Definition zip_unit_E f := erefl (zip_unit f). 
End zip_Def.

Export zip_Def.

Notation "f1 \* f2" := (zip f1 f2) 
  (at level 41, left associativity).

Notation "[zunit  f ]" := (zip_unit f) (at level 0).

(* lemmas for zip *)

Section ZipMonoidLemmas.
Notation K := [ordType of nat].
Notation sort := (to_type \o pcm_sort).
Notation V := (@idynamic pcm_code sort). 
Notation feap := (feap V).
Implicit Type f : feap.

Lemma zipC f1 f2 : f1 \* f2 = f2 \* f1.
Proof.
case: f1 f2=>[|s1] [|s2]; rewrite !zip_E //. 
by rewrite finmap.zipC //; apply: ZipLemma.zip_comm_f.
Qed.

Lemma zipA f1 f2 f3 : f1 \* (f2 \* f3) = f1 \* f2 \* f3.
Proof.
case: f1 f2 f3=>[|s1] [|s2] [|s3]; rewrite !zip_E //. 
- by case: (finmap.zip _ _ _).
case E1: (finmap.zip _ s2 s3)=>[v1|]; last first.
- case E3: (finmap.zip _ s1 s2)=>[v3|] //=.
  move: (finmap.zipA ZipLemma.zip_assoc_f s1 s2 s3).
  by rewrite E1 E3 /= => <-. 
case E3: (finmap.zip _ s1 s2)=>[v3|]; last first.
- move: (finmap.zipA ZipLemma.zip_assoc_f s1 s2 s3).
  by rewrite E1 E3 /= => ->.
move: (finmap.zipA ZipLemma.zip_assoc_f s1 s2 s3).
by rewrite E1 E3 /= => ->.
Qed.

Lemma zipCA f1 f2 f3 : f1 \* (f2 \* f3) = f2 \* (f1 \* f3).
Proof. by rewrite zipA (zipC f1) -zipA. Qed.

Lemma zipAC f1 f2 f3 : f1 \* f2 \* f3 = f1 \* f3 \* f2.
Proof. by rewrite -zipA (zipC f2) zipA. Qed.

Lemma zunitL f : [zunit f] \* f = f.
Proof.
case: f=>[|s]; rewrite zip_E zip_unit_E //. 
by rewrite finmap.zip_unitL.
Qed.

Lemma zunitR f : f \* [zunit f] = f.
Proof. by rewrite zipC zunitL. Qed.

Lemma zvalidL f1 f2 : valid (f1 \* f2) -> valid f1.
Proof. by rewrite zip_E; case: f1. Qed.

Lemma zvalidR f1 f2 : valid (f1 \* f2) -> valid f2.
Proof. by rewrite zipC; move/zvalidL. Qed.

Lemma valid_zunit f : valid [zunit f] = valid f.
Proof. by rewrite zip_unit_E; case: f. Qed.

Lemma zunitP f1 f2 : 
        reflect ([zunit f1] = [zunit f2]) (weq pcmCode f1 f2).
Proof.
case: f1 f2=>[|s1] [|s2] /=; rewrite !zip_unit_E /weq /fmap /=;
case: eqP=>E; constructor=>//; case: s1 s2 E=>s1 H1 [s2 H2]; 
move/feapE/fmapE=>/= E; apply/feapE; apply/fmapE=>/= {H1 H2}; last first.
- move=>T; apply: E; elim: s1 s2 T=>[|[k1 v1] s1 IH] /= [|[k2 v2] s2] //=.
  by case=>-> E1 _ E2; congr ((_,_)::_)=>//; apply: IH.
elim: s1 s2 E=>[|[k1 v1] s1 IH] /= [|[k2 v2] s2] //=.
case=>-> E1 E2; congr ((_,_)::_); last by apply: IH. 
by apply: idynE=>// pf; rewrite E1 in pf *; rewrite ieqc. 
Qed.

Lemma validZP f1 f2 : 
        reflect ([/\ valid f1, valid f2 & [zunit f1] = [zunit f2]]) 
                (valid (f1 \* f2)).
Proof.
case: f1 f2=>[|s1] [|s2]; rewrite /= /PCM.valid /= zip_E !zip_unit_E;
try by constructor; case.
set j := (@finmap.zip_unitE K V ZipLemma.zip_f ZipLemma.unit_f
         ZipLemma.zip_unitL_f ZipLemma.zip_unitE_f). 
case E: (finmap.zip _ s1 s2)=>[s|] /=; constructor.
- by split=>//; apply/feapE; apply/j; exists s. 
by case=>_ _ []; case/j=>s; rewrite E. 
Qed.

Lemma validZE (f1 f2 : feap) : 
        valid (f1 \* f2) = [&& valid f1, valid f2 & weq pcmCode f1 f2].
Proof.
rewrite andbA andbC; case: zunitP=>E; case: validZP=>//=; last by case.
- by case=>->. 
case: (valid f1); case: (valid f2)=>//.
by move=>H; suff: False by []; apply: H.
Qed.

Lemma zip_undefL f : FUndef \* f = FUndef.
Proof. by rewrite zip_E. Qed.

Lemma zip_undefR f : f \* FUndef = FUndef.
Proof. by rewrite zipC zip_undefL. Qed. 

End ZipMonoidLemmas.

(* Interaction of Zip with the other primitives *)

Section OtherLemmas.
Notation K := [ordType of nat].
Notation V := (idynamic (to_type \o pcm_sort)).
Notation feap := (feap V).
Implicit Type f g : feap.
Implicit Type U V : encoded_pcm.

(*******************************)
(* zunit and other connectives *)
(*******************************)

Lemma zunit_undef : [zunit FUndef] = FUndef.
Proof. by rewrite zip_unit_E. Qed.

Lemma zunit_unitf f : [zunit f] = Unit <-> f = Unit.
Proof. 
rewrite zip_unit_E; case: f=>//; case. 
by case=>[|[k v] s] //= H; split=>_; apply/feapE; apply/fmapE=>/=. 
Qed.

Lemma zunit_unit : [zunit Unit] = Unit.
Proof. by apply/zunit_unitf. Qed.

Lemma zunit_zunit f : [zunit [zunit f]] = [zunit f].
Proof.
rewrite !zip_unit_E; case: f=>//.
case=>s /= H; apply/feapE; apply/fmapE=>/=. 
rewrite /zip_unit' /mapf' /ZipLemma.unit_f.  
elim: s H=>[|[k v] s IH] //=.
by move/path_sorted/IH=>->.
Qed.

Lemma fdomZU f : fdom [zunit f] =i fdom f. 
Proof.
move=>x; rewrite zip_unit_E /=.
case: f=>[//|[s pf]]; rewrite -!topredE /fdom /supp /=.
elim: s pf=>[//|[k v] s IH] /= pf; rewrite !inE /=.
by case: eqP=>//= H; apply: IH; apply: path_sorted pf.
Qed.

Lemma flookZU U f x (u : U) : 
        flook x [zunit f] = Some (pcm_dyn u) -> u = Unit. 
Proof.
case: f=>[|s]; rewrite zip_unit_E /= /fnd //=.
case: s=>s H; elim: s H=>[|[k v] s /= IH] //= H.
case: eqP=>[_|N].
- case=>->; move/(inj_pairT2 _ _ _ _ _). 
  by case: U u {IH}=>sort [decode] pf t u; subst sort; rewrite ieqc.
case E: (filter _ _ ) IH=>[//|[k2 v2] s2]; apply.
by apply: path_sorted H.
Qed.

Lemma ffreeZU f x : ffree x [zunit f] = [zunit ffree x f].
Proof.
case: f=>[|s]; rewrite !zip_unit_E //=.
apply/feapE; case: s=>s H; rewrite /zip_unit; apply/fmapE=>/=. 
elim: s H=>[|[k v] s IH] //= H.
by case: eqP=>_ /=; rewrite IH //; apply: path_sorted H.
Qed.

Lemma zunitUn f1 f2 : [zunit f1 \+ f2] = [zunit f1] \+ [zunit f2].
Proof.
case: f1 f2=>[|h1][|h2];
rewrite !zip_unit_E /PCM.join //= /FeapPCM.funion2.
by case: ifP=>D /=; rewrite -zunit_disj D // zunit_fcat. 
Qed.

Lemma zunitPt U x (u : U) : [zunit x \-> u] = x \-> @Unit U.
Proof. 
rewrite zip_unit_E; apply/feapE; apply/fmapE.
congr [:: (_, _)]; apply: idynE=>//.
by case: U u=>sort [decode] E s u pf; subst sort; rewrite !ieqc. 
Qed.

Lemma zunitPtUn U x (u : U) f : 
        [zunit x \-> u \+ f] = x \-> @Unit U \+ [zunit f].
Proof. by rewrite zunitUn zunitPt. Qed.

Lemma zunitZL f1 f2 : 
        [zunit f1 \* f2] = if valid (f1 \* f2) then [zunit f1] else FUndef.
Proof.
case: f1 f2=>[|s1][|s2]; rewrite !zip_unit_E !zip_E;
rewrite //= /valid /= /FeapPCM.fdef.
case E: (finmap.zip _ s1 s2)=>[a|//]; apply/feapE.
rewrite /zip /zip_unit /=. 
case: s1 s2 a E =>s1 H1 [s2 H2] [s H] T; apply/fmapE. 
move: T; rewrite /zip /=; move: (zip_sorted' _).
case E: (zip' _ _ _ )=>[a|//] _ [F]; rewrite {a H1 H2 H}F in E.
elim: s1 s2 s E=>[|[k1 v1] s1 IH]; case=>[|[k2 v2] s2] //= s; first by case=><-.
case: eqP=>// ->{k1}.
case E1: (ZipLemma.zip_f v1 v2)=>[b|//].
case E2: (zip' _ s1 s2)=>[c|//]; case=><-{s} /=.
move/IH: E2=>->; congr ((_, _)::_).
rewrite /ZipLemma.unit_f /ZipLemma.zip_f in E1 *.
by case: decP E1=>// E1; case=><-{b}; apply: idynE=>// pf.
Qed.

Lemma zunitZR f1 f2 : 
        [zunit f1 \* f2] = if valid (f1 \* f2) then [zunit f2] else FUndef.
Proof. by rewrite zipC zunitZL. Qed.

Lemma flookZUf U x (u : U) f : 
        flook x [zunit f] = Some (pcm_dyn u) -> 
        exists u1 : U, flook x f = Some (pcm_dyn u1).
Proof.
case: f=>[|s]; rewrite zip_unit_E //=. 
case: s=>s H; rewrite /fnd /zip_unit {H} /=. 
elim: s u=>[|[k v] s IH] //= u.
case: eqP=>_; last by apply: IH. 
case=>E {u x k s IH} _; case: v E=>tp u /= E; subst tp.
case: U u=>sort [decode] E t u; subst sort.
exists u; congr Some; apply: idynE=>//; rewrite !ieqc =>pf.  
by rewrite ieqc. 
Qed.

Lemma flookfZU U x f (u : U) : 
        flook x f = Some (pcm_dyn u) ->
        flook x [zunit f] = Some (pcm_dyn (@Unit U)).
Proof.
case: f=>[|s]; rewrite zip_unit_E //=.
case: s=>s H; rewrite /fnd /zip_unit {H} /=.
elim: s u=>[|[k v] s IH] //= u.
case: eqP=>_; last by apply: IH.
case=>->{v}; congr Some.
apply: idynE=>// pf; rewrite {pf IH s x k}ieqc.
case: U u=>sort [decode] E t i; subst sort.
by rewrite ieqc.
Qed.

Lemma zunitPtUnf U x f k (u : U) :
        valid (x \-> u \+ k) ->
        [zunit f] = x \-> u \+ k -> 
        exists v : U, exists g, f = x \-> v \+ g /\ [zunit g] = k.
Proof.
move=>Vu H; move: (Vu); rewrite -H=>Vz.
have E: flook x [zunit f] = Some (pcm_dyn u).
- by rewrite H flookPtUn.
case/flookZUf: E Vz H=>v.
move/(@feap_etaPt pcmCode)=>-> Vz H.
exists v, (ffree x f); split=>//.
rewrite zunitUn zunitPt in Vz H.
by case/(fcancel Vz): H.
Qed.

Lemma zunitPtf U f x (u : U): 
        [zunit f] = x \-> u -> exists v : U, f = x \-> v.
Proof.
rewrite -[x \-> u]unitR; move/zunitPtUnf.
rewrite unitR fvalidPt; move/(_ (erefl _)).
by case=>v [g][->]; move/zunit_unitf=>->; exists v; rewrite unitR.
Qed.

(*********************************)
(* now zip and other connectives *)
(*********************************)

(* zip and unit *)

Lemma zip_unitE f1 f2 : f1 \* f2 = Unit -> f1 = Unit /\ f2 = Unit.
Proof.
case: f1 f2=>[|s1][|s2]; rewrite zip_E //=.
case: s1 s2=>s1 H1 [s2 H2] /= => L.
suff: s1 = [::] /\ s2 = [::] by case; split; apply/feapE; apply/fmapE. 
move: {H1 H2} (zip_sorted' _) L.  
case: s1 s2=>[|[k1 v1] s1]; case=>[|[k2 v2] s2] //=. 
case: eqP=>// <-{k2}. 
by case: (zip' _ s1 s2)=>//; case: (ZipLemma.zip_f v1 v2).
Qed.

Lemma unit_zipE : Unit \* Unit = Unit.
Proof. by rewrite /Unit zip_E; apply/feapE; apply/fmapE. Qed.

(* zip and update don't have any nice equations; just this one *)
(* which is further specialized for the same-type case in zipPt *)

Lemma zipPtd x y d1 d2 : x \\\-> d1 \* y \\\-> d2 = 
        if x == y then 
          if ZipLemma.zip_f d1 d2 is Some d then x \\\-> d
          else FUndef
        else FUndef.
Proof.
rewrite zip_E; case E: (ZipLemma.zip_f d1 d2)=>[a|] /=; 
move: (zip_sorted' _)=>/=; case: eqP=>// ->; rewrite E => pf //.
by apply/feapE; apply/fmapE. 
Qed.

(* zip and valid: a few are already listed in the monoidal section  *)

Lemma validUnZL f1 f2 g1 g2 : 
        valid (f1 \* f2) -> valid (g1 \* g2) -> 
        valid (f1 \* f2 \+ g1 \* g2) = valid (f1 \+ g1).
Proof.
case: f1 f2 g1 g2=>[|s1][|s2][|t1][|t2]; rewrite !zip_E //=.
case E1: (finmap.zip _ s1 s2)=>[s|//] _; 
case E2: (finmap.zip _ t1 t2)=>[t|//] _.
case: fvalidUn=>//; case: fvalidUn=>//=.
- move=>_ _ L x H1 H2; move: (zip_supp E1 x) (zip_supp E2 x).
  by rewrite H1 H2 sym; move/L/negbTE=>->. 
move=>x H1 H2 _ _ L; move: (zip_supp E1 x) (zip_supp E2 x).
by rewrite H1 H2; move/L/negbTE=>->. 
Qed.

Lemma validUnZR f1 f2 g1 g2 : 
        valid (f1 \* f2) -> valid (g1 \* g2) -> 
        valid (f1 \* f2 \+ g1 \* g2) = valid (f2 \+ g2).
Proof.
by move=>V1 V2; rewrite (zipC f1) (zipC g1) in V1 V2 *; apply: validUnZL.
Qed.

Lemma validZUn f1 f2 g1 g2 : 
        valid ((f1 \+ g1) \* (f2 \+ g2)) -> 
        valid (f1 \* f2) = valid (g1 \* g2).
Proof.
suff L : forall f1 f2 g1 g2, valid ((f1 \+ g1) \* (f2 \+ g2)) ->
  valid (f1 \* f2) -> valid (g1 \* g2).
- move=>Vz; apply/idP/idP; apply: L=>//.
  by rewrite (joinC g1) (joinC g2). 
move=>{f1 f2 g1 g2} f1 f2 g1 g2.
case: validZP=>//; case=>V1 V2 Efg _.
case: validZP=>//; case=>_ _ Ef _.
case: validZP=>// V; elim: V.
split.
- by move/validR: V1.
- by move/validR: V2.
rewrite -valid_zunit !zunitUn in V1 Efg.
by rewrite -Ef in Efg; apply: funhKl V1 Efg.
Qed.

(* zip and dom *)

Lemma fdomZL f1 f2 : 
        fdom (f1 \* f2) =i if valid (f1 \* f2) then fdom f1 else pred0.
Proof.
move=>x. 
case: f1 f2=>[|s1][|s2]; rewrite zip_E //=. 
by case E: (finmap.zip _ s1 s2)=>[s|//] /=; move/zip_supp: E; apply.
Qed.

Lemma fdomZR f1 f2 : 
        fdom (f1 \* f2) =i if valid (f1 \* f2) then fdom f2 else pred0.
Proof. by rewrite zipC; apply: fdomZL. Qed.

(* zip and free *)

Lemma ffreeZ f1 f2 x : 
        valid (f1 \* f2) ->
        ffree x (f1 \* f2) = (ffree x f1) \* (ffree x f2).
Proof.
case: f1 f2=>[|s1][|s2]; rewrite !zip_E //=.
by case E: (finmap.zip _ s1 s2)=>[s|] //=; rewrite (zip_rem _ E).
Qed.

Lemma flookZf U x f1 f2 (u : U) : 
        flook x (f1 \* f2) = Some (pcm_dyn u) -> 
        exists u1, exists u2, 
          [/\ u = u1 \+ u2, flook x f1 = Some (pcm_dyn u1) & 
              flook x f2 = Some (pcm_dyn u2)].
Proof.
case: f1 f2=>[|s1][|s2]; rewrite !zip_E //=.
rewrite /flook; case E: (finmap.zip _ s1 s2)=>[s|//].
case/(zip_fnd E)=>[[tp1 /= v1]][[tp2 /= v2]][]. 
rewrite /ZipLemma.zip_f.
case: decP=>//= pf; move: (esym _); rewrite -(eqP pf) in v2 *.
move=>qf; rewrite ieqc {E qf pf}; case=>E.
rewrite {tp1}E in v1 v2 *; move/(inj_pairT2 _ _ _ _ _).
(* time to break up U *)
case: U u v1 v2=>sort [decode] E t /= v v1 v2; subst sort.
rewrite ieqc=><- ->->; exists v1, v2.
by split=>//; congr Some; apply: idynE=>// pf; rewrite !ieqc. 
Qed.

Lemma flookfZT U1 U2 x f1 f2 (u1 : U1) (u2 : U2) v : 
        flook x f1 = Some (pcm_dyn u1) ->
        flook x f2 = Some (pcm_dyn u2) ->
        flook x (f1 \* f2) = Some v -> U1 = U2 :> pcm.
Proof.
case: f1 f2=>[|s1][|s2]; rewrite !zip_E //=. 
rewrite /flook /=; case E: (finmap.zip _ s1 s2)=>[s|//] H1 H2.
case/(fnd_zip H1 H2): E=>-> {H1 H2}.
rewrite /ZipLemma.zip_f /=. 
case: decP=>//= pf _; move/eqP: pf.
case: U1 u1 U2 u2=>sort1 [decode1] E1 t1 /= v1.
case=>sort2 [decode2] E2 t2 /= v2. 
by subst sort1 sort2; rewrite /decode /= => ->. 
Qed.

Lemma flookfZ U x f1 f2 (u1 u2 : U) v : 
        flook x f1 = Some (pcm_dyn u1) ->
        flook x f2 = Some (pcm_dyn u2) ->
        flook x (f1 \* f2) = Some v -> v = pcm_dyn (u1 \+ u2).
Proof.
case: f1 f2=>[|s1][|s2]; rewrite !zip_E //=. 
rewrite /flook /=; case E: (finmap.zip _ s1 s2)=>[s|//] H1 H2.
case/(fnd_zip H1 H2): E=>-> {H1 H2}.
rewrite /ZipLemma.zip_f /=. 
case: decP=>//= pf [<-]; rewrite {s1 s2 s x pf}ieqc.
(* now brute force; case over U when nothing else works *)
case: U u1 u2=>sort [decode] m s /= v1 v2. 
subst sort; move: (esym _)=>pf.
have {pf}-> /=: pf = (erefl _) by move: pf; apply: Streicher_K. 
apply: idynE=>// pf; rewrite {pf}ieqc.
move: (esym _)=>pf. 
by have {pf}->/=: pf = (erefl _) by move: pf; apply: Streicher_K. 
Qed.

(* zip and union *)

Lemma zipUn f1 f2 g1 g2 : 
        valid (f1 \* f2) -> valid (g1 \* g2) ->
        f1 \* f2 \+ g1 \* g2 = (f1 \+ g1) \* (f2 \+ g2).
Proof.
move=>Vf Vg.
case Vu: (valid (f1 \* f2 \+ g1 \* g2)); last first.
- move/finvalidE: (negbT Vu)=>->; rewrite validUnZL // in Vu.
  by move/finvalidE: (negbT Vu)=>->; rewrite zip_undefL. 
have [V1 V2] : valid (f1 \+ g1) /\ valid (f2 \+ g2).
- by rewrite -(validUnZL Vf Vg) -(validUnZR Vf Vg) Vu.
have Vz: valid ((f1 \+ g1) \* (f2 \+ g2)).
- case: validZP=>// H; suff: False by []; apply: H; rewrite V1 V2 !zunitUn.
  case: validZP Vf=>//; case=>_ _ ->.
  by case: validZP Vg=>//; case=>_ _ ->.
apply: flook_eta=>// x.
case E1: (flook x _)=>[a|]; last first.
- case E2: (flook x _)=>[b|//].
  rewrite (pcm_eta b) in E2; case/flookZf: E2=>v1 [v2][_]. 
  move/flook_some. 
  have: x \notin fdom (f1 \* f2 \+ g1 \* g2).  
  - by case: fdom_flook E1=>// v ->. 
  rewrite !fdomUn !inE /= negb_and negb_or Vu /= !fdomZL Vf Vg.
  by case/andP; do 2![move/negbTE=>->]; rewrite andbF.
case E2: (flook x _)=>[b|]; last first.
- rewrite (pcm_eta a) in E1.
  move/flook_some: E1; rewrite fdomUn !inE Vu /=.
  move/flook_none: E2=>E2; move: E2 (E2).
  rewrite {1}fdomZL {1}fdomZR Vz. 
  rewrite !fdomUn !inE /= V1 V2 /= !negb_or.
  by case/andP=>H1 H2 _; rewrite !fdomZL Vf Vg (negbTE H1) (negbTE H2). 
rewrite (pcm_eta b) (pcm_eta a) in E1 E2 *.
case/flookZf: E2=>v1 [v2][E].
move/flook_some: (E1); rewrite fdomUn !inE Vu /=.
case E2: (x \in fdom (f1 \* f2))=>/=.
- move=>_; move: E1; rewrite flookUnl // E2.
  case/flookZf=>w1 [w2][F] F1 F2.
  rewrite !flookUnl // (flook_some F1) (flook_some F2) {}F1 {}F2 {}E {}F.
  rewrite /pcm_enc/code_enc /= in v1 v2 w1 w2 *.
  case=>/= T; rewrite /decode /= in T *.
  move: (esym _) (esym _); rewrite T in w1 w2 * =>pf1 pf2. 
  move/(inj_pairT2 _ _ _ _ _); rewrite !ieqc=>->; case=>/=.  
  by move/(inj_pairT2 _ _ _ _ _); rewrite !ieqc=>->. 
move=>{E2} E2; move: E1.
rewrite flookUnr // E2; case/flookZf=>w1 [w2][F] F1 F2. 
rewrite !flookUnr // (flook_some F1) (flook_some F2) {}F1 {}F2 {}E {}F.
rewrite /pcm_enc/code_enc /= in v1 v2 w1 w2 *.
case=>/= T; rewrite /decode /= in T *.
move: (esym _) (esym _); rewrite T in w1 w2 * =>pf1 pf2.
move/inj_pair2; rewrite !ieqc=>->; case=>/=.
by move/inj_pair2; rewrite !ieqc=>->. 
Qed.

(* specializing the lemmas to the \-> notation *)

Lemma zipPtUnf U f1 f2 g x (u : U) : 
        valid (x \-> u \+ g) ->
        f1 \* f2 = x \-> u \+ g -> 
        exists u1 : U, exists u2 : U, exists g1 : feap, exists g2 : feap,
         [/\ u = u1 \+ u2, g = g1 \* g2, 
             f1 = x \-> u1 \+ g1 & f2 = x \-> u2 \+ g2].
Proof.
move=>Vu H; move: (Vu); rewrite -H=>Vz.
have E: flook x (f1 \* f2) = Some (pcm_dyn u).
- by rewrite H flookPtUn. 
case/flookZf: E Vu H=>v1 [v2][->].
move/(@feap_etaPt pcmCode)=>E1.
move/(@feap_etaPt pcmCode)=>E2 Vu H.
exists v1, v2, (ffree x f1), (ffree x f2); split=>//.
by rewrite -ffreeZ // H ffreePtUn. 
Qed.

Lemma zipPtf U f1 f2 x (u : U) : 
        f1 \* f2 = x \-> u -> 
        exists u1 : U, exists u2 : U,
         [/\ u = u1 \+ u2, f1 = x \-> u1 & f2 = x \-> u2].
Proof.
rewrite -[x \-> u]unitR; move/zipPtUnf. 
rewrite unitR fvalidPt; move/(_ (erefl _)).
case=>u1 [u2][g1][g2][->]; move/sym; case/zip_unitE=>->->->->.  
by exists u1, u2; rewrite !unitR.
Qed.

Lemma validPt U1 U2 x y (u1 : U1) (u2 : U2) : 
       valid (x \-> u1 \* y \-> u2) = (x == y) && (decode U1 == decode U2).
Proof.
rewrite zipPtd; case: eqP=>//= ->; rewrite /ZipLemma.zip_f.
by case: decP=>//; case: eqP.
Qed.

Lemma zipPt U x (u1 u2 : U) : x \-> u1 \* x \-> u2 = x \-> (u1 \+ u2).
Proof.
rewrite zipPtd eq_refl /ZipLemma.zip_f; case: decP=>//=.
case: U u1 u2=>sort [decode] E s /= u1 u2; subst sort=>/= pf. 
rewrite /pts_encode ieqc; congr (x \\\-> _); apply: idynE=>// qf.
by rewrite !ieqc.  
Qed.

Lemma zipUnPt U x (u1 u2 : U) g1 g2 : 
        (x \-> u1 \+ g1) \* (x \-> u2 \+ g2) = x \-> (u1 \+ u2) \+ g1 \* g2.
Proof.
case Vg : (valid (g1 \* g2)).
- rewrite -zipPt -zipUn //; case: validZP=>// H.
  suff: False by []; apply: H.
  by rewrite !fvalidPt !zunitPt.
move/negbT: Vg=>Vg; move/finvalidE: (Vg)=>->.
suff: ~~ valid ((x \-> u1 \+ g1) \* (x \-> u2 \+ g2)) by move/finvalidE=>->.
apply: contra Vg.
rewrite !validZE; case/and3P=>V; move: (V).
rewrite -valid_zunit zunitPtUn in V. 
rewrite !fvalidPtUn; case/andP=>-> _; case/andP=>-> _ /=.
case: zunitP=>//; rewrite !zunitPtUn.
by case/(fcancel (C:=pcmCode) V)=>_ _ E _; case: zunitP.
Qed.

(* the inverted zipUn lemma, proved by feap induction on s1 *)

Lemma zipUnf f g s1 s2 : 
        valid (s1 \+ s2) ->
        f \* g = s1 \+ s2 -> 
        exists f1, exists g1, exists f2, exists g2,  
         [/\ s1 = f1 \* g1, s2 = f2 \* g2, 
             f = f1 \+ f2 & g = g1 \+ g2].
Proof.
move: s1 f g s2.
apply: (@dynfeap_ind pcmCode)=>[//||U h k v IH V] f g s2 Vu H.
- by exists Unit, Unit, f, g; rewrite !unitL -{2}zunit_unit zunitL in H *.
rewrite -joinA in Vu H.
case/(zipPtUnf Vu): H=>u1 [v1][f'][g'][->]; move/sym=>E ->->{f g}.
case/(IH _ _ _ (validR Vu)): E=>f1 [g1][f2][g2][->->->->] {IH V Vu}.
exists (k \-> u1 \+ f1), (k \-> v1 \+ g1), f2, g2.
by rewrite zipUnPt !joinA.
Qed.

End OtherLemmas.

(* Now the lemmas about flattening *)

Section FlatteningLemmas.
Variable C : flatcode.

Lemma flat_unit : @flat C Unit = Unit.
Proof. by rewrite flatE. Qed.

Lemma pcm_flatPtP (U : encoded_pcm) (u : U) : 
        decode U == pcm_heap -> U = heap :> Type.
Proof. by case: U u=>sort []; case=>// s; subst sort. Qed.

Lemma pcm_flatPt (U : encoded_pcm) x (u : U) :
         @flat pcmFlatCode (x \-> u) = 
         if @decP _ (decode U == pcm_heap) idP is 
            left pf then icoerce id u (pcm_flatPtP u pf)
         else Unit.
Proof.
rewrite flatE /= /map_flat/= /to_heap/= /flatten_pcm /=. 
case: decP=>pf; last by rewrite unitL.
case: U u pf=>sort [decode] E /= s pf qf; subst sort.
by rewrite !ieqc unitR; congr icoerce; apply: proof_irrelevance.
Qed.

Lemma pcm_flatPtE x (u : heap) : @flat pcmFlatCode (x \-> u) = u.
Proof. by rewrite pcm_flatPt /= ieqc. Qed.

Lemma type_flatPtP (U : encoded_type) (u : U) : 
        decode U == tp_heap -> U = heap :> Type.
Proof. by case: U u=>sort []; case=>// s; subst sort. Qed.

Lemma type_flatPt (U : encoded_type) x (u : U) :
         @flat typeFlatCode (x ::-> u) = 
         if @decP _ (decode U == tp_heap) idP is 
            left pf then icoerce id u (type_flatPtP u pf)
         else Unit.
Proof.
rewrite flatE /= /map_flat/= /to_heap/= /flatten_tp /=. 
case: decP=>pf; last by rewrite unitL.
case: U u pf=>sort [decode] E /= s pf qf; subst sort.
by rewrite !ieqc unitR; congr icoerce; apply: proof_irrelevance.
Qed.

Lemma tp_flatPtE x (u : heap) : @flat typeFlatCode (x ::-> u) = u.
Proof. by rewrite type_flatPt /= ieqc. Qed.

Lemma flatUn f1 f2 : @flat C (f1 \+ f2) = 
                     if valid (f1 \+ f2) then flat f1 \+ flat f2 else um_undef.
Proof. by rewrite flatE; apply: fflat_join. Qed.

(* the last lemma shows that reasoning about flattening is conditional *)
(* and will hence not be easy to iterate (that is, we need some automation *)
(* hopefully this won't be something we often do, so maybe the current set *)
(* of lemmas suffices *)

End FlatteningLemmas.


(********************)
(* And for allvalid *)
(********************)

Section AllValidLemmas.
Notation K := [ordType of nat].
Notation V := (idynamic (to_type \o pcm_sort)).
Notation feap := (feap V).
Implicit Type f g : feap.
Implicit Type U V : encoded_pcm.

Lemma allvalidV f : allvalid f -> valid f.
Proof. by rewrite allvalidE; case: f. Qed.

Lemma allvalid_unit : allvalid Unit.
Proof. by rewrite allvalidE. Qed.

Lemma allvalidPt U x (u : U) : allvalid (x \-> u) = valid u. 
Proof.
rewrite allvalidE /map_flat /to_bool /=. 
case: U u=>sort [decode] E /= s u; subst sort. 
by rewrite !ieqc unitR.
Qed.

Lemma allvalidUn f1 f2 : 
        allvalid (f1 \+ f2) = valid (f1 \+ f2) && allvalid f1 && allvalid f2.
Proof.
move: (allvalid_join f1 f2); rewrite allvalidE.  
by case: (valid _).
Qed.

Lemma allvalidPtUn U x (u : U) f : 
         allvalid (x \-> u \+ f) = (valid u) && (x \notin fdom f) && allvalid f.
Proof. rewrite allvalidUn fvalidPtUn allvalidPt.
case V: (allvalid f); last by rewrite !andbF. 
by rewrite (allvalidV V) !andbT /= andbC. 
Qed.

End AllValidLemmas.

Hint Resolve allvalid_unit.

(* what can be said for flat and zip, and allvalid and zip? *)

(* flat and zip *)

Lemma flatZ f1 f2 : 
        valid (f1 \* f2) ->
        @flat pcmFlatCode (f1 \* f2) = 
        @flat pcmFlatCode f1 \+ @flat pcmFlatCode f2.
Proof.
case: f1 f2=>[|s1][|s2]; rewrite flatE zip_E; 
rewrite //= /valid /= /FeapPCM.fdef.
case E: (finmap.zip _ s1 s2)=>[a|//]; move=>_ /=.
case: s1 s2 a E=>s1 H1 [s2 H2][s H]; rewrite /map_flat /=. 
move: (zip_sorted' _).
case E: (zip' _ _ _)=>[a|//] _ [F].  
rewrite {a H1 H2 H}F in E *.
elim: s1 s2 s E=>[|[k1 v1] s1 IH]; case=>[|[k2 v2] s2] //= s.
- by case=><-; rewrite unitR.
case: eqP=>// ->{k1}.
case E1: (ZipLemma.zip_f v1 v2)=>[b|//].
case E2: (zip' _ s1 s2)=>[c|//].
case=><-{s} /=; move/IH: E2=>-> {IH k2 c}.
suff {s1 s2}: @to_heap pcmFlatCode b = 
              @to_heap pcmFlatCode v1 \+ @to_heap pcmFlatCode v2.
- by move=>->; rewrite -joinA -(joinCA (seq_flat _ _)) joinA. 
rewrite /ZipLemma.zip_f /to_heap /= /flatten_pcm in E1 *. 
case: decP E1=>// E1; case=><-{b} /=. 
case: decP=>E2; last first.
- case: decP=>E3; last by rewrite unitR.
  by rewrite (eqP E1) E3 in E2.
case: v1 v2 E1 E2=>U /= v1 [V /= v2] E1 E2.
move: (eqP E1) (flatten_pcmP _) (flatten_pcmP _)=>/= X1 X2 X3.
rewrite -{E1 V}(eqP E1) in v2 X1 *.
rewrite (eqP E2) in v1 v2 X1 X2 X3 *.
rewrite !ieqc {X1 X2 X3}. 
by case: decP=>// H; rewrite ieqc.
Qed.

(* allvalid and zip *)

(* hmm, there's not much one can say there *)
(* as you actually have to look at the leaf values *)
(* not just the leaf types *)

(* Lemma allvalidZ f1 f2 : 
        allvalid (f1 \* f2) =  *)

(* filter and union *)
(* should be subsumed by union_maps *)

Lemma ffiltUnK V p (h1 h2 : feap V) : 
        valid (h1 \+ h2) ->    
        ffilter p (h1 \+ h2) = h1 -> 
        ffilter p h1 = h1 /\ ffilter p h2 = Unit. 
Proof.
move=>Vl; rewrite (ffiltUn _ Vl)=>E.
have {Vl} Vl : valid (ffilter p h1 \+ ffilter p h2). 
- by rewrite E; move/validL: Vl.
have F : fdom (ffilter p h1) =i fdom h1.
- move=>x; rewrite fdom_filt inE /=.
  apply/andP/idP=>[[//]| H1]; split=>//; move: H1.
  rewrite -{1}E fdomUn /= !inE /= !fdom_filt inE /= inE /=.
  by case: (x \in p)=>//=; rewrite andbF.
rewrite -{2}[h1]unitR in E F; move/(fdomE Vl E): F=>X.
by rewrite X in E Vl; move/(funhKl Vl): E. 
Qed.
