Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path rels ordtype.
Require Import Eqdep prelude idynamic pcm unionmap heap codes feaps state zigzag.
Require Import finmap worlds inject rely privstate.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.


Section GeneralStability.
Implicit Types (U : encoded_pcm) (V : encoded_type).
Variables (U U1 U2 : encoded_pcm) (V V1 V2 : encoded_type).

Lemma stb_ex (W : mod) lp sh ep s : 
        menv_steps W [lp, sh, ep] s -> 
        exists sh' ep', s = [lp, sh', ep']. 
Proof. by move/menvs_loc=>/= ->; exists (shr s), (env s); rewrite st_eta. Qed.

Lemma stb_ex1 (W : mod) x (lp : U) (sh : V) (ep : U) s : 
        menv_steps W (x ->> [lp, sh, ep]) s -> 
        exists (sh' : V) (ep' : U), s = x ->> [lp, sh', ep'].
Proof.
move=>M; case: (menvs_coh M)=>C C'.
have [Vl2 Vl3] : valid (shr s) /\ valid (env s).
- by case/cohV/and3P: C'.
have [F2 F3] : fdom (shr s) =i pred1 x /\ fdom (env s) =i pred1 x.
- case/cohfV: C; rewrite !fsE /= => _ F2 F3; case/cohfV: C'=>_ E2 E3. 
  by split=>y; rewrite ?E2 ?E3 -?F2 -?F3 inE.
have X1 : loc s = x \-> lp by move/menvs_loc: M; rewrite fsE. 
have [V' [v' X2]] : exists (V' : encoded_type) (v' : V'), shr s = x ::-> v'.
- have: x \in fdom (shr s) by rewrite F2 inE.
  case/feap_etaPV=>d [_ ->].
  rewrite (_ : ffree x (shr s) = Unit); last first.
  - apply: fdom0E; first by rewrite fvalidF.
    by move=>y; rewrite fdomF inE F2 inE eq_sym; case: eqP.
  exists (@code_enc typeCode d), (@code_val typeCode d).
  by rewrite unitR /pts_encode -code_eta.
have [U' [u' X3]] : exists (U' : encoded_pcm) (u' : U'), env s = x \-> u'.
- have: x \in fdom (env s) by rewrite F3 inE.
  case/feap_etaPV=>d [_ ->].
  rewrite (_ : ffree x (env s) = Unit); last first.
  - apply: fdom0E; first by rewrite fvalidF.
    by move=>y; rewrite fdomF inE F3 inE eq_sym; case: eqP.
  exists (@code_enc pcmCode d), (@code_val pcmCode d).
  by rewrite unitR /pts_encode -code_eta.
have Ev : decode V = decode V'.
- move: (coh_weq C C'); rewrite -[s]st_eta fsE /= X2. 
  by case/eqP/feapE/fmapE. 
have [v2 {X2} X2] : exists v2 : V, shr s = x ::-> v2. 
- rewrite X2 /pts_encode /code_dyn.
  case: V' v' Ev {X2}=>sort [decode] /= xx ss v1; subst sort; rewrite ieqc.
  case: V=>sort2 [decode2] /= xx2 ss2; subst sort2.
  rewrite /Encoded.Exports.decode /= => ->.
  by exists v1; rewrite ieqc.
have Eu : decode U = decode U'.
- move/cohSO/allvalidV: C'. 
  by rewrite X1 X3 validPt eq_refl /=; move/eqP. 
have [u2 {X3} X3] : exists u2 : U, env s = x \-> u2.
- rewrite X3 /pts_encode /code_dyn. 
  case: U' u' Eu {X3}=>sort [decode] /= xx ss u1; subst sort; rewrite ieqc.
  case: U=>sort2 [decode2] /= xx2 ss2; subst sort2. 
  rewrite /Encoded.Exports.decode /= => ->.
  by exists u1; rewrite ieqc. 
by exists v2, u2; rewrite -[s]st_eta X1 X2 X3 fsE.
Qed.

(*
Lemma menvs_stable2 (W : mod) x1 x2 (lp1 : U1) (sh1 : V1) (ep1 : U1)
                                    (lp2 : U2) (sh2 : V2) (ep2 : U2) s : 
        menv_steps W (x1 ->> [lp1, sh1, ep1] \+ x2 ->> [lp2, sh2, ep2]) s -> 
        exists (sh1' : V1) (ep1' : U1) (sh2' : V2) (ep2' : U2), 
          s = x1 ->> [lp1, sh1', ep1'] \+ x2 ->> [lp2, sh2', ep2'].
*)

End GeneralStability.


(**************************)
(* Revise the stuff below *)
(**************************)


(* let's try some concrete scsl actions. but first a few lemmas for *)
(* switching between long and short notation for states.            *)

Section NotationSwitch. 
Implicit Types (U : encoded_pcm) (V : encoded_type).

Lemma stE1 U V x (lp ep : U) (sh : V) : 
        x ->> [lp, sh, ep] = x@sf-> lp \+ x@jn-> sh \+ x@ot-> ep. 
Proof. by rewrite fsE sfE jnE otE. Qed.

(* it's perhaps strange here that we don't pull the shared part too *)
(* but, in the present conditions, it can't really be pulled *)
(* there's no requirement on coherence that the shared part has the same *)
(* V footprint as the rest; in the case of loc and env, it's imposed by *)
(* validSO, but not so for the shared. Maybe that should change, but let's *)
(* wait and see. In particular, I'm not really sure that I need to pull *)
(* the shared part here, as that part should really not be reasoned about *)
(* explicitly (methinks), but only through invariants *)

(* Yeah, this lemma is less useful than it could be *)
(* because in the case W is something basic, say Priv *)
(* it doesn't let me know that loc s' and env s' are empty *)
(* So I have to reprove the lemma for Priv *)

(* This points to changing the definition of worlds, so that *)
(* each of them carries it's domain of labels, and coherent worlds *)
(* have the same labels in all three feaps *)
Lemma stE2 (W : mod) U x (lp : U) s : 
        x@sf-> lp \+ s \In Mod.coh W ->
        exists (ep : U) s', s = x@ot-> ep \+ s'.
Proof.
move/cohSO/allvalidV; rewrite sfE /= unitL.
case: validZP=>//; case=>V1 V2 /esym V3 _.
rewrite -valid_zunit in V1; rewrite zunitPtUn in V1 V3.
case/(zunitPtUnf V1): V3=>ep [g][E] _.
exists ep, [loc s, shr s, g].
by rewrite otE /PCM.join /= !unitL -E st_eta.
Qed.

Lemma menvs_loc_eq1 (W : mod) U x (lp : U) i m : 
        menv_steps W (x@sf-> lp \+ i) m -> 
        exists m', m = x@sf-> lp \+ m'.
Proof.
move=>M; case: (menvs_coh M)=>Ci Cm.
have Vi : valid (x \-> lp \+ loc i).
- by case/cohV/and3P: Ci; rewrite sfE. 
have Vm : valid (loc m) by case/cohV/and3P: Cm.
have: fdom (loc m) =i fdom (x \-> lp \+ loc i).
- by case/and3P: (coh_footV Cm Ci); case/fdomeqP; rewrite sfE.
move/(_ x); rewrite fdomPtUn inE /= eq_refl andbT Vi.
case/feap_etaPV=>v [_ Em].
have Ev : v = pcm_dyn lp.
- move/menvs_loc: M; rewrite /= Em sfE.
  by case/(fcancelV Vi)=><-. 
rewrite {v}Ev in Em.
exists [ffree x (loc m), shr m, env m].
by rewrite sfE /PCM.join /= -Em !unitL st_eta. 
Qed.

(*
Lemma menvs_joint_eq1 V (W : MOD) U x (sh: V) i m : 
        menv_steps W (x@jn->sh \+ i) m -> 
        exists m', m = x@jn->sh \+ m'.
Proof.
move=>M; case: (menvs_coh M)=>Ci Cm.
have Vi : valid (x ::-> sh \+ shr i). by case/cohV/and3P: Ci.
have Vm : valid (shr m) by case/cohV/and3P: Cm.
have: fdom (shr m) =i fdom (x ::-> sh \+ shr i).
- by case/and3P: (coh_footV Cm Ci)=> _; case/fdomeqP. 
move/(_ x); rewrite fdomPtUn inE /= eq_refl andbT Vi.
case/feap_etaPV=>v [_ Em].
(* have Ev : v = pcm_dyn sh. 
- by move/menvs_loc: M; rewrite /= Em; case/(fcancelV Vi)=><-. 
rewrite {v}Ev in Em. *)
exists [ffree x (loc m), shr m, env m].
rewrite /PCM.join /= !unitL -[m]st_eta /=. 
Qed.
*)


(* the above action is for non-communicating actions *)
(* we need one where two local parts exist *)
(* probably we'd need more too *)

Lemma menvs_loc_eq2 (W : mod) U U' x x' (lp : U) (lp' : U') i m : 
        menv_steps W (x@sf-> lp \+ x'@sf-> lp' \+ i) m ->
        exists m', m = x@sf-> lp \+ x'@sf-> lp' \+ m'.
Proof.
rewrite -joinA=>M; case/menvs_loc_eq1: M (M)=>m1 ->{m} M.
rewrite !(joinC (x@sf->_)) -!joinA in M. 
case/menvs_loc_eq1: M (M)=>m2; rewrite -!(joinC (x@sf->_)).
rewrite !sfE; case=>E _ _ M.
have V : valid (x \-> lp \+ loc m1).
- by case/menvs_coh/proj2/cohV/and3P: M.
have V' : valid (x' \-> lp' \+ loc m2) by rewrite -E.
have N : x != x'.
- case/menvs_coh/proj1/cohV/and3P: M. 
  rewrite /= joinA; move/validL; case: fvalidUn=>// _ _.
  by move/(_ x'); rewrite !fdomPt !inE /=; move/(_ (eq_refl _)). 
suff: exists m1', m1 = x'@sf-> lp' \+ m1'.
- by case=>m1' ->; exists m1'; rewrite joinA sfE. 
have: x' \in fdom (x \-> lp \+ loc m1).
- by rewrite E fdomPtUn inE /= eq_refl andbT. 
rewrite fdomPtUn inE /= V /= (negbTE N) /=.
case/feap_etaPV=>v [_ E2].
have Ev : v = pcm_dyn lp'.
- by rewrite E2 joinCA in E V; case/(fcancelV V): E.
rewrite {v}Ev in E2.
exists [ffree x' (loc m1), shr m1, env m1].
by rewrite sfE {1}/PCM.join /= -E2 !unitL st_eta.
Qed.

End NotationSwitch. 

(* need to repeat stE2 for a specific world *)
(* because the original stE2 lemma is not good enought *)

Section NotationPriv.

Lemma priv_stE2 hp (lp : heap) s : 
        hp@sf-> lp \+ s \In Mod.coh (Priv hp) ->
        exists (ep : heap), s = hp@jn-> tt \+ hp@ot-> ep.
Proof.
case=>/validHV V [hL][hE]; rewrite stE1 -!joinA.
by case/(sfcancel V)=>_ _ ->; exists hE. 
Qed.

Lemma mrely_priv hp (hL hE : heap) s : 
        mrely (Priv hp) (hp ->> [hL, tt, hE]) s ->
        exists hE' : heap, s = hp ->> [hL, tt, hE'].
Proof.
case.
- case; case=>hL1 [hL2][hE1][].
  rewrite -[s]st_eta {1 2}/transp !fsE /= -2!fsE.
  by case/scancelPt=>->-> [->->->]; exists hL2; rewrite -fsE. 
case=>ch [h][/Mem_map_inv][ct][->{ch}]; rewrite InE /= => ->.
case; case; case=>hL1 [hE1][];
rewrite -[s]st_eta {1 2}/transp !fsE /= -2!fsE; 
case/scancelPt=>->-> [->->->]; rewrite -!fsE.
- by exists (hL1 \+ h). 
by exists hL1.
Qed.

End NotationPriv.

(* for CAS and Unlock, I need notation lemma for priv+lock *)

(*
Section NotationPrivLock.
Variables (hp : nat) (lkT : lockT).
Notation W := (entangle (Priv hp) (Lock lkT)).
Notation l := (@label_of lkT).
Notation lk := (@lock_of lkT).
Notation U := (@pcm_of lkT).

Lemma privlock_stE2 (hL : heap) (mgL : lift mutexUnlifted * U) s : 
        hp@sf-> hL \+ l@sf-> mgL \+ s \In Mod.coh W ->
        exists (hE h : heap) (b : bool) (mgE : lift mutexUnlifted * U), 
          s = hp@jn-> tt \+ hp@ot-> hE \+ 
              l@jn-> (lk :-> b \+ h) \+ l@ot-> mgE.
Proof.
case=>Vh [s1][s2][E] C1 C2; case: C1 C2 E=>_ [hL1][hE] ->{s1}. 
case=>_ _ [mL][mE][gL][gE][b][h][->{s2}] _.
rewrite !stE1 -!joinA -!(joinCA (l@sf-> _)) in Vh *.
case/(sfcancel (validHV Vh))=>_ {Vh} V; case/(sfcancel V)=>_ _ ->{V s}. 
by exists hE, h, b, (mE, gE); rewrite -!joinA. 
Qed.

Lemma privlock_stE3 (hL : heap) (gL gE : U) s : 
        hp@sf->hL \+ l@sf->(up own, gL) \+ 
                     l@ot->(up nown, gE) \+ s \In Mod.coh W ->
        exists hE : heap, 
          s = hp@jn->tt \+ hp@ot->hE \+ l@jn->(lk :-> true).
Proof.
case=>Vh [s1][s2][E] C1 C2; case: C1 C2 E=>_ [hL1][hE] ->{s1}. 
case=>_ _ [mL][mE][gL1][gE1][b][h][->{s2}] L E. 
rewrite !stE1 -!joinA -!(joinCA (l@sf-> _)) in Vh E.
rewrite -(joinC (l@ot->_)) -2!(joinCA (l@ot->(mE, gE1))) in E.
case/(sfcancel (validHV Vh)): E L; case=>{Vh} <-<-{mL gL1} V. 
case/(sfcancel V)=>_ {V} V; case/(otcancel V).
case=>{V} <-<-{mE gE1} _ -> L. 
exists hE; case: L; last by case.
by case=>_ ->->; rewrite unitR joinA. 
Qed.

(* we also need that environment can't change the gE *)
(* of a lock which is owned by the thread *)

Lemma mrely_lock_gE (gL gE : U) (h : heap) s : 
        mrely (Lock lkT) (l ->> [(up own, gL), h, (up nown, gE)]) s -> 
        s = l ->> [(up own, gL), h, (up nown, gE)]. 
Proof.
case; first by case; rewrite /= transp_inj. 
case=>_ [k][/Mem_map_inv][ct][->]; rewrite InE => ->; case.
- case=>gL1 [gL2][gE1][]; case; rewrite {1}/transp fsE /= -fsE.
  by move/scancelPt. 
case=>gL1 [gE1][]; rewrite {1}/transp fsE /= -fsE.
by move/scancelPt. 
Qed.

Lemma mrely_privlock_gE (hL : heap) (gL gE : U) i s : 
        mrely W (hp@sf-> hL \+ l@sf-> (up own, gL) \+ 
                               l@ot-> (up nown, gE) \+ i) s -> 
        exists i', s = hp@sf-> hL \+ l@sf-> (up own, gL) \+ 
                                     l@ot-> (up nown, gE) \+ i'.
Proof.
(* first painfully turn everything into module-normal form *)
move=>M; case/mrely_coh: (M)=>C _; case/privlock_stE3: C M=>i' ->{i}. 
rewrite -!joinA -2!(joinCA (hp@jn->_)) -2!(joinCA (hp@ot->_)).
rewrite -(joinC (l@jn->_)) 2!joinA -stE1 (joinA (l@sf->_)) -stE1.
(* now the real part of the proof *)
move=>M; case/mrely_coh: M (M); case/get_lock=>C1 _ _.
case=>_ [s1][s2][->{s}] C2 _.
case/(mrely_split (entangle_extL _ _) C1 C2)=>{C1 C2}.
case/mrely_priv=>hE' ->{s1} /mrely_lock_gE ->{s2}. 
rewrite !stE1 -!joinA -2!(joinCA (l@sf->_)).
rewrite -(joinC (l@ot->_)) -2!(joinCA (l@ot->_)).
exists (hp@jn->tt \+ (hp@ot->hE' \+ l@jn-> (lk :-> true))). 
by rewrite -!joinA. 
Qed.

Lemma menvs_privlock_gE (hL : heap) (gL gE : U) i s : 
        menv_steps W (hp@sf-> hL \+ l@sf-> (up own, gL) \+ 
                                    l@ot-> (up nown, gE) \+ i) s -> 
        exists i', s = hp@sf-> hL \+ l@sf-> (up own, gL) \+ 
                                     l@ot-> (up nown, gE) \+ i'.
Proof.
case=>n; elim: n i=>[|n IH] i /=; first by case=><-; exists i. 
case=>s' [] /mrely_privlock_gE [i' ->{s'}].
by case/IH=>i'' ->{IH s}; exists i''.
Qed.

End NotationPrivLock.
*)

Section NotationPrivEntangle.
Variables (hp : nat) (V : mod). 
Notation U := (entangle (Priv hp) V). 

Lemma privent_stE2 (hL : heap) s :
        hp@sf-> hL \+ s \In Mod.coh U ->
        exists (hE : heap) s' , 
          s = hp@jn-> tt \+ hp@ot-> hE \+ s' /\ s' \In Mod.coh V.
Proof.
case=>Vh [s1][s2][E] C1 C2; case: C1 C2 E=>_ [hL1][hE] ->{s1} C2.
rewrite !stE1 -!joinA in Vh *.
case/(sfcancel (validHV Vh))=>_ _ ->{s Vh}.
by exists hE, s2; rewrite joinA. 
Qed.

End NotationPrivEntangle.

