Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path fintype ordtype.
Require Import Eqdep finmap prelude rels.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive. 

(* Finite Sets implemented as Finite Maps from (A: ordType) into unit*)

Notation finSet A := (finMap A unit).
Notation inss k s := (ins k tt s).

Section Defs.
Variables (A : ordType).
Definition emptyfs := (@nil A unit).

(* seq_of does not add any more info than support *)

Lemma seq_supp (f: finSet A): 
   seq_of f = map (fun p => (p, tt)) (supp f).
Proof. 
elim/fmap_ind':f=>// k [] f P IH.
by rewrite seqof_ins // IH /supp /= seqof_ins //.
Qed.

(* so we can reason out of support for fsets *)
Lemma fsetE (f g: finSet A): f = g  <-> supp f =i supp g .
Proof.
split=>[->|] // /suppE=> H. 
elim/fmap_ind': f H=>//[|k [] f P _].
- by rewrite supp_nil=> /esym  /supp_nilE -> //. 
elim/fmap_ind': g =>//[|k2 [] g P2 _].
- by move=> /supp_nilE ->//.
rewrite /supp !seqof_ins //=; case=>E; rewrite E in P *=>  H; apply/fmapE.
by rewrite !seqof_ins //; rewrite !seq_supp //=  (_: supp f = supp g) //. 
Qed.   


Lemma  inss_supp_in k (s : finSet A) : 
        k \in supp s -> inss k s = s. 
Proof.
move=>H; apply/fmapP=>x; rewrite fnd_ins.
by case: eqP=>// ->; case: suppP H=>// [][] ->.
Qed.

Lemma inssNE k s: 
     inss k s <> emptyfs.
Proof.
elim/fmap_ind': s k=>// k1 [] s P _ k.
case: (totalP k k1)=> [P1| /eqP ->| P1]; move/fmapE.
+ by rewrite seqof_ins // ?path_supp_ins // (ord_path P1) //. 
- rewrite ins_ins; case:eqP=>//_; rewrite seqof_ins //.
by rewrite ins_ins; case:eqP=>//[->|_]; rewrite seqof_ins // ?path_supp_ins //.
Qed.


(* finSets are ordered *)

Definition ord_fs  (x y : finSet A) := ords (seq_of x) (seq_of y).

Lemma irr_fs : irreflexive ord_fs.
Proof. by case=>x i; rewrite /ord_fs //= irr_ords. Qed.
 
Lemma trans_fs : transitive ord_fs.
Proof. by case=>x iX [y iY][z iZ]; rewrite /ord_fs //=; apply: trans_ords. Qed.

Lemma total_fs : forall x y, 
            [|| ord_fs x y, x == y | ord_fs y x].
Proof.
by case=> x iX [y iY]; rewrite /ord_fs  //=; exact: total_ords. Qed.

Definition fs_ordMixin := OrdMixin irr_fs trans_fs total_fs.
Canonical Structure fs_ordType := Eval hnf in OrdType (finSet A) fs_ordMixin.
End Defs.
Arguments emptyfs {A}.

Notation unitfs k := (inss k emptyfs).

(* finsets have less restrictive lemmas than finMaps for fcat, there
are no disjointness requirements  *) 

Section CatLemmas.  
Variable (A: ordType).

Lemma fscatss (s : finSet A) : fcat s s = s.
Proof.
elim/fmap_ind': s=>//k [] s P IH.
by rewrite fcat_sins fcat_inss ?notin_path // IH ins_ins; case:eqP=>//.
Qed.

Lemma fscat1s (s : finSet A) x : fcat (unitfs x) s = inss x s.
Proof.
elim/fmap_ind': s=>//k [] s P IH.
by rewrite fcat_sins IH ins_ins ins_ins; do 2 case:eqP=>//= ->.
Qed.


Lemma fscatC (s1 s2 : finSet A) : fcat s1 s2 = fcat s2 s1.
Proof.
elim/fmap_ind': s2 s1=>[|k2 [] s2 L IH] s1 /=; first by rewrite fcat0s.
rewrite fcat_sins IH.
elim/fmap_ind': s1 =>// k1 [] s1 P IH2.
rewrite fcat_sins ins_ins fcat_sins IH2; case:eqP=>//H.
rewrite H in IH2 L *.
by rewrite fcat_inss ?notin_path // ins_ins; case:eqP=>//.
Qed.

Lemma fscats1 (s : finSet A) x : fcat s (unitfs x)  = inss x s.
Proof. by rewrite fscatC fscat1s. Qed.

Lemma fscat_ins (s1 s2: finSet A) k:
      fcat (inss k s1) s2 = inss k (fcat s1 s2).
Proof. by rewrite fscatC fcat_sins fscatC. Qed.

Lemma fscatEmpty (s1 s2: finSet A) : 
  fcat s1 s2 = emptyfs <-> s1 = emptyfs /\ s2 = emptyfs.
Proof.
elim/fmap_ind': s2 s1=>//[| k1 [] s1 P1 IH1]s2.
- by  rewrite fcats0; split=>//; case.
elim/fmap_ind': s2=>//[| k2 [] s2 P2 _].
- by  rewrite fcat0s; split=>//; case.
rewrite fcat_sins -fscat_ins IH1; split=>//; case; first by move/inssNE.
by move/fmapE; rewrite seqof_ins //.
Qed.


Lemma fscatA (s1 s2 s3 : finSet A) : 
        fcat (fcat s1 s2) s3 = fcat s1 (fcat s2 s3).
Proof.
elim/fmap_ind': s3 s1 s2 =>[|k v s3 L IH] s1 s2 /=; first by rewrite !fcats0.
by rewrite fcat_sins ?notin_path // IH // fcat_sins ?notin_path // fcat_sins.
Qed.         

Lemma fscatAC (s1 s2 s3 : finSet A) : 
        fcat s1 (fcat s2 s3) = fcat s2 (fcat s1 s3).
Proof. by rewrite -!fscatA // (@fscatC s1 s2). Qed.

Lemma fscatCA (s1 s2 s3 : finSet A) : 
        fcat (fcat s1 s2) s3 = fcat (fcat s1 s3) s2.
Proof. by rewrite !fscatA // ?(@fscatC s2 s3). Qed.

Lemma inss_fcat (s: finSet A) k:
      inss k s = fcat (unitfs k) s.
Proof. by rewrite fscat_ins fcat0s. Qed.

End CatLemmas.



Section Foldfs.

Section Defs.
Variables (A: ordType) (B: Type).
(* Recursion on finite sets is defined using foldr and the underlying
sequenc: No conditions are required for g and e *)

Definition foldfs (g : A -> B -> B) (e: B) (s: finSet A) :=
  foldr (fun p => g (key p)) e (seq_of s).

(* Foldfs universality equations instantiated for nil/inss views *)

Lemma foldfs_nil g e : foldfs g e emptyfs = e.
Proof. by rewrite /foldfs //=. Qed.

Lemma foldfs_ins g e k f: 
   path ord k (supp f) ->
  foldfs g e (inss k f) = g k (foldfs g e f).
Proof. by move=> H; rewrite /foldfs //= seqof_ins //=. Qed.

End Defs.
Variables (A B: ordType).

Lemma foldfs_id x :
  foldfs  (fun k x => inss k x) (@emptyfs A) x  = x.
Proof. by elim/fmap_ind': x=>//= k [] s P IH; rewrite foldfs_ins // IH. Qed.

Lemma foldfs_end (X: Type) (e: X)  x :
  @foldfs A X (fun _ => @id X) e x = e.
Proof. by elim/fmap_ind': x=>//= k [] s P IH; rewrite foldfs_ins // IH. Qed.


Lemma foldfs_cat g (e: finSet B) k x:
  foldfs (fun s =>  [eta fcat (g s)]) e (inss k x) = 
   fcat (g k) (@foldfs A _ (fun s => [eta fcat (g s)]) e x).
Proof. 
elim/fmap_ind': x k=>// k1 [] x P IH k.
case:(totalP k k1)=> [H1 | /eqP H1| H1].
- rewrite  foldfs_ins //.
 by apply: (path_supp_ins _ H1); apply: (ord_path H1).
- rewrite IH H1 ins_ins; case:eqP=>// _.
 by rewrite -fscatA fscatss IH.
rewrite ins_ins; case:eqP=>//H.
- by rewrite H in H1; move:(irr k1);move/negP/(_ H1).
rewrite IH  foldfs_ins; last by apply: path_supp_ins H1 P. 
by rewrite IH fscatAC.
Qed.
End Foldfs.
Arguments foldfs {A B} g e s.


Section Flatten.
Variable (A: ordType).
Definition fsflatt (s: finSet (fs_ordType A)):=
   foldfs (fun s x => fcat s x) (@emptyfs A) s.

Lemma fsflatt_nil:
  fsflatt emptyfs = emptyfs.
Proof. by [].  Qed.

Lemma fsflatt_ins k s:
  fsflatt (inss k s) = fcat k (fsflatt s).
Proof.
elim/fmap_ind': s k=>// k2 [] s P IH k.
rewrite IH ins_ins; case:eqP=>//[->|H].
- by rewrite -fscatA fscatss /fsflatt foldfs_ins //.
rewrite -fscatA -IH fscatC IH fscatA.
case:(totalP k2 k)=> [H2| /eqP /esym /H // | H2].
- by rewrite -IH /fsflatt /= foldfs_ins ?path_supp_ins //.
rewrite ins_ins; case:eqP=>[/esym /H //|_].
rewrite /fsflatt /= foldfs_ins  ?/supp //= ?seqof_ins //= ?H2 ?P //.
by rewrite foldfs_ins // fscatAC.
Qed.

Lemma fsflatt_unit x:
  fsflatt (unitfs x) = x.
Proof. by rewrite fsflatt_ins fsflatt_nil fcats0. Qed.

(* flatten distributes over cat *)
Lemma fscat_flatt s1 s2:
    fsflatt (fcat s1 s2) = fcat (fsflatt s1) (fsflatt s2).
Proof.
elim/fmap_ind': s2 s1=>//= k [] s _ IH s1.
by rewrite fcat_sins !fsflatt_ins fscatAC IH.
Qed.

Lemma fsflattE f: 
  fsflatt f = emptyfs <-> f = emptyfs \/ f = inss emptyfs emptyfs.
Proof.
elim/fmap_ind': f=>//[| k [] f P IH].
- by rewrite fsflatt_nil; split=>//; left.
split; last first. 
- case=>[|->]; last by rewrite fsflatt_unit. 
 by move/fmapE ;rewrite seqof_ins //.
by rewrite fsflatt_ins; move/fscatEmpty=>[->]/IH [->|->]; right=>//;
  rewrite ins_ins; case:eqP=>//.
Qed.

End Flatten.
Arguments fsflatt {A} s.



(* functorial map on finSet *)

Section MapS.
Section MapDef.
Variables (A B: ordType) (V : Type).

Variable (f: A -> B).
Definition mapfs (m : finSet A ) : finSet B := 
  foldfs (fun p x => inss (f p) x) emptyfs m.

Lemma supp_mapfs k s : 
        k \in supp s -> f k \in supp (mapfs s). 
Proof.
elim/fmap_ind': s k=>// k [] s P IH k'.
rewrite supp_ins inE /=  /mapfs foldfs_ins // supp_ins inE /=.
case:eqP=>/= [-> _| _ H]; first by rewrite eq_refl.
by rewrite (IH _ H) orbT.
Qed.

Lemma mapfs_nil : mapfs emptyfs = emptyfs.
Proof. by rewrite /mapfs //=. Qed.

Lemma mapfs_ins k s :
  mapfs (inss k s) = inss (f k) (mapfs s).
Proof. 
elim/fmap_ind': s k=>// k1 [] s P IH k.
case:(totalP k k1)=> [H1 | /eqP H1| H1].
- rewrite {1}/mapfs foldfs_ins; last first.
  + by apply: (path_supp_ins _ H1); apply: (ord_path H1).
  by rewrite foldfs_ins // IH /mapfs. 
- rewrite IH H1 ins_ins; case:eqP=>// _.
  by rewrite IH ins_ins;case:eqP=>//.
rewrite ins_ins; case:eqP=>// H.
- by rewrite H in H1; move:(irr k1);move/negP/(_ H1).
rewrite IH [in X in _ = X]ins_ins;case: eqP=>//= [H2|_].
- rewrite /mapfs foldfs_ins; last by apply: path_supp_ins H1 P. 
 rewrite inss_supp_in; first by apply: IH.
 by rewrite -H2; apply: supp_mapfs; rewrite supp_ins inE eq_refl. 
rewrite -IH  /mapfs /= foldfs_ins //.
by apply: path_supp_ins H1 P.
Qed.


Lemma mapfs_kfilt p s:
  kfilter p (mapfs s) = mapfs (kfilter (p \o f) s).
Proof.
elim/fmap_ind': s=>[|/=k [] s P IH]; first by rewrite !kfilt_nil mapfs_nil.
rewrite mapfs_ins kfilt_ins IH /comp /= kfilt_ins ;case:ifP=>// _. 
by rewrite mapfs_ins.
Qed.

(* If k \in supp s, then fnd (f x) \in supp (mapf s). If not anything
 can happen: take a constant function f x = 0, then if the map is not
 empty f x \in supp *)

Lemma mapfs_fnd k s :
    k \in supp s -> fnd k s = fnd (f k) (mapfs s).
Proof. 
by move=>H;case:(fnd_supp_in H)=>[[]];case:(fnd_supp_in (supp_mapfs H))=>[[]]->.
Qed.

Lemma mapfs_fcat s1 s2:
  mapfs (fcat s1 s2) = fcat (mapfs s1) (mapfs s2).
Proof.
elim/fmap_ind': s2 s1=>//k [] s P IH s1. 
by rewrite fcat_sins mapfs_ins IH mapfs_ins fcat_sins.
Qed.

End MapDef.
Arguments mapfs {A B} f m.

Variables (A B C : ordType)(f : A -> B) (g : B -> C).

Lemma mapfs_id m : @mapfs A A id m = m.
Proof. 
by elim/fmap_ind': m=>// k [] s L IH; rewrite -{2}IH /mapfs foldfs_ins //.
Qed.

Lemma mapfs_comp m:
       mapfs g (@mapfs A B f m) = mapfs (comp g f) m.
Proof.
elim/fmap_ind': m  =>//= k [] s P IH. 
rewrite [mapfs (g \o f) _]mapfs_ins //.
by rewrite mapfs_ins mapfs_ins IH. 
Qed.

Lemma mapfs_flatt s:
  mapfs f (fsflatt s) = fsflatt (mapfs (mapfs f) s).
Proof.
elim/fmap_ind': s=>// k [] s P IH.
by rewrite mapfs_ins !fsflatt_ins mapfs_fcat IH.
Qed.

(* foldfs_fmap only holds for order preserving f/finset homomofrphisms  *)

(* Lemma foldfs_map (h: B -> C-> C) (e: C)  s: *)
(*   foldfs h e (mapfs f s) = foldfs (h \o f) e s. *)


Lemma mapfsEmpty s:
  mapfs f s = emptyfs <-> s = emptyfs.
Proof.
elim/fmap_ind':s=>//k [] s P IH.
split; last by move/fmapE; rewrite seqof_ins //.
by rewrite mapfs_ins; move/inssNE.
Qed.
End MapS.
Arguments mapfs {A B} f m.


Section SubSet.
Variable (T: ordType).

Definition fsleq (f g: finSet T) := 
     all (fun x: T => x \in (supp g)) (supp f).


Lemma flseq_refl : reflexive fsleq.
Proof. by move=> f; rewrite /fsleq; case:allP=>//. Qed.

Lemma fsleq_trans: transitive fsleq.
Proof. 
move=> f g h; rewrite /fsleq=> /allP H1 /allP=> H2.
by case:allP=>//H3; apply: False_ind; apply: H3=>x /H1/H2. 
Qed.

Lemma fsleq_antisym: antisymmetric fsleq.
Proof. 
have H x f: fnd x f = Some tt -> x \in supp f.
- by case:suppP=>// ->; discriminate.
move=> f g; rewrite /fsleq; case:allP=>// H1 /allP=> H2;apply/fsetE=>x. 
case: suppP=>//[ [] /H/H1-> //|]; case:suppP=>//[[]] /H/H2.
by case:suppP=>//[[]]-> _; discriminate.
Qed.


(* emptyfs is the least element of the poset of finSet T *)

Lemma empty_fsleq (f: finSet T) : fsleq emptyfs f.
Proof. by []. Qed.


Lemma fsleq_insW (f g: finSet T) k: 
     fsleq f g ->  fsleq f (inss k g).
Proof. 
move/allP=> H; rewrite /fsleq; case:allP=>// H1; apply: False_ind; apply:H1.
by move=> x /(H x); rewrite !supp_ins inE => ->; rewrite orbT. 
Qed.

Lemma fsleq_insWL (f g : finSet T) k:
     fsleq (inss k f) g -> fsleq f g.
Proof.
move/allP=>H; rewrite /fsleq; case:allP=>//H1; apply: False_ind; apply: H1.
by move=> x H2; apply: H; rewrite !supp_ins inE H2 orbT //.
Qed.

Lemma inss_mono (f g: finSet T) k : 
     fsleq f g ->  fsleq (inss k f) (inss k g).
Proof. 
move/allP=> H; rewrite /fsleq; case:allP=>//H1; apply: False_ind; apply: H1=>x.
rewrite !supp_ins !inE.
case:orP=>//; case=> [ /eqP -> _|]; first by rewrite eq_refl orTb.
by move/H=> -> _; rewrite orbT. 
Qed.

Lemma fsleq_catW  (f g h : finSet T):
     fsleq f g -> fsleq f (fcat g h).
Proof.
move/allP=>H; rewrite /fsleq; case:allP=>// H1; apply: False_ind; apply:H1=>x.
by move/H=>H1; rewrite supp_fcat !inE H1 orTb.
Qed.

Lemma fsleq_catWL  (f g h : finSet T):
     fsleq (fcat f h) g -> fsleq f g.
Proof.
move/allP=>H; rewrite /fsleq; case:allP=>// H1; apply: False_ind. 
by apply:H1=>x H1;apply: H ; rewrite supp_fcat !inE H1 orTb.
Qed.

Lemma fcat_mono (f g h : finSet T) :
  fsleq f g ->  fsleq (fcat f h) (fcat g h). 
Proof.
elim/fmap_ind': h f g=>//= k [] h P IH f g H.
by rewrite !fcat_sins inss_mono // ?IH //.
Qed.

End SubSet.
(* Prop is such a nice place *)

(* Section Subset. *)
(* Variable (T: ordType). *)


(* Definition fsleq (f g: finSet T) := {subset (supp f) <= (supp g)}. *)

(* (* emptyfs is the least element of the poset of finSet T *) *)

(* Lemma fsleq_weakening (f g: finSet T) k : *)
(*      fsleq f g ->  fsleq f (inss k g). *)
(* Proof. move=> H x. /H. rewrite !supp_ins inE => ->; rewrite orbT. Qed. *)

(* (* monotonicity of inss *) *)

(* Lemma inss_mono (f g: finSet T) k :  *)
(*      fsleq f g ->  fsleq (inss k f) (inss k g). *)
(* Proof.  *)
(* move=> H x. rewrite !supp_ins !inE. *)
(* case:orP=>//; case=> [ /eqP -> _|]; first by rewrite eq_refl orTb. *)
(* by move/H=> -> _; rewrite orbT.  *)
(* Qed. *)


(* End Subset. *)




