Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude idynamic pcm unionmap heap codes feaps state.
Require Import zigzag worlds.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(***************************************************)
(* Order theory of framing on coherence predicates *)
(***************************************************)

Definition leq_coh (c1 c2 : cohpred) := 
  exists c, c2 = star_coh c1 c.

Lemma leq_coh_refl c : leq_coh c c. 
Proof. by exists emp_coh; rewrite starC0. Qed.

Hint Resolve leq_coh_refl.

Lemma leq_coh_trans c2 c1 c3 : 
        leq_coh c1 c2 -> leq_coh c2 c3 -> leq_coh c1 c3.
Proof. by case=>c -> [c' ->]; exists (star_coh c c'); rewrite starA. Qed.

Lemma leq_coh_asym c1 c2 : 
        leq_coh c1 c2 -> leq_coh c2 c1 -> c1 = c2.
Proof.
case=>c ->{c2} [c']; rewrite -starA => E. 
apply/cohE; split=>/=. 
- rewrite E !starfV; apply: fext=>s /=. 
  by rewrite orbA orbAC -!orbA orbb. 
- case V : (valid (sig c1)); last by move/finvalidE: (negbT V)=>->. 
  have : sig c1 = sig c1 \+ sig c \+ sig c'.
  - by rewrite [in X in X = _]E /= joinA.
  rewrite -{1 2}[sig c1]unitR -joinA in V *.
  by case/(funhKl V)/esym/fun0E=>-> _; rewrite unitR.
have {E} E s : s \In c1 <-> s \In star_coh c1 (star_coh c c').
- by rewrite -E.
apply: fext=>s; apply: pext; split.
- move=>H; case/E: H (H) =>_ [s1][s2][->{s}] T1 T2 H.
  move/(cohK T1): H T2=>->{s2}.
  case/star0E=>U1 U2; rewrite /= unitR.  
 split; first by apply: cohH T1.
 by exists s1, Unit; rewrite unitR.  
move=>C; case: C (C)=>_ [s1][s2][->{s}] C1 C2 C. 
case/E: C1 (C1) C=>_ [t1][t2][->{s1}] T1 T2 C1 C.
move/(cohK T1): C1 C T2=>->{t2} C; case/star0E=>U U'.
apply/E; split; first by [move/cohH: C].
exists t1, s2; rewrite unitR; split=>//.
split; first by [move/cohH: C2].
by exists s2, Unit; rewrite unitR. 
Qed.

(* star is monotone wrt. leq_coh *)

Lemma star_mono c1 c2 d1 d2 : 
        leq_coh c1 d1 -> leq_coh c2 d2 ->
        leq_coh (star_coh c1 c2) (star_coh d1 d2).
Proof.
case=>t1 ->{d1} [t2] ->{d2}; rewrite -!starA -(starCA c2) starA.
by exists (star_coh t1 t2). 
Qed.

Lemma star_emp c : leq_coh emp_coh c.
Proof. by exists c; rewrite star0C. Qed.

Implicit Arguments star_emp [c].
Hint Resolve star_emp.

Lemma star_id c d : leq_coh c (star_coh c d).
Proof. by rewrite -{1}[c]starC0; apply: star_mono. Qed.

Implicit Arguments star_id [c d].
Hint Resolve star_id.

(******************************************)
(* Order theory of framing on transitions *)
(******************************************)

Definition leq_guar coh coh' (g : guar coh) (g' : guar coh') := 
  leq_coh coh coh' /\
  forall s1 s s2, 
    s1 \+ s \In coh' -> s2 \+ s \In coh' -> 
    g s1 s2 -> g' (s1 \+ s) (s2 \+ s).

(* 
In the definition above, we require s2 \+ s \In coh' so that the
definition applies to acquire transitions. The check ensures that s
doesn't overlap with the state that s2 may acquire wrt s1.

When g is internal, we can get rid of the extra check. We can infer
that s1 and s2 have the same V and H footprints, and therefore s1 \+ s
\In coh' implies s2 \+ s \In coh'.  
*)

Lemma leq_guar_internal c c' (g : int_guar c) (g' : guar c') s1 s2 s :
        leq_guar g g' -> 
        s1 \+ s \In c' -> 
        g s1 s2 -> g' (s1 \+ s) (s2 \+ s).
Proof. 
case=>L H C G; apply: H=>//; case: L C=>c2 ->.
by move/(int_guar_extend _ _ G).
Qed.

(* now the order theory *)

Lemma leq_guar_refl coh (g : guar coh) : leq_guar g g.
Proof. 
split=>// s1 s s2 C _ H; case/guar_coh: (H)=>C1 _.
by move/(cohK C1): C=>->; rewrite !unitR. 
Qed.

Hint Resolve leq_guar_refl.

Lemma leq_guar_prop_refl c1 c2 (g1 : guar c1) (g2 : guar c2) : 
        c1 = c2 -> g1 = g2 :> (state -> state -> Prop)-> leq_guar g1 g2. 
Proof.
move=>E G; rewrite -{c2}E in g2 G *; split=>//.
move=>s1 s s2 C1 C2 H; rewrite -{}G.
by case: (leq_guar_refl g1)=>_; apply.
Qed.

Lemma leq_guar_trans c2 (g2 : guar c2) c1 (g1 : guar c1) c3 (g3 : guar c3) : 
        leq_guar g1 g2 -> leq_guar g2 g3 -> leq_guar g1 g3.
Proof. 
case=>F1 G1 [F2 G2]; split; first by apply: leq_coh_trans F1 F2.
have L s1 s2 : s1 \+ s2 \In c3 -> s1 \In c1 -> 
  exists r2 t2, s2 = r2 \+ t2 /\ s1 \+ r2 \In c2. 
- case: F1 F2=>c' E2; setoid_rewrite E2; case=>c'' ->. 
  rewrite -starA => H1 H2.
  case/(star_coh_prec H1): (H2)=>_ [t1][t2][E T1 T2]. 
  rewrite {s2}E joinA in H1 *; exists t1, t2. 
  split=>//; split; last by exists s1, t1. 
  by move/cohH/validHL: H1.
move=>s1 s s2 C1 C2 H; case/guar_coh: (H)=>D1 D2.
case: (L _ _ C1 D1)=>r1 [t1][X R1]. 
case: (L _ _ C2 D2)=>r2 [t2][E R2].
rewrite {s}X in E C1 C2 *.
have {t2 E} E : r1 = r2.
- apply: footeqV_prec (validR (cohV C1)) E _.
  apply: footeqVK (coh_footV R1 R2) (coh_footV D1 D2).
  by apply: cohV R1.
rewrite -{r2}E in R2; rewrite !joinA in C1 C2 *.
by apply: G2 C1 C2 _; apply: G1 R1 R2 H.
Qed.

Lemma leq_guar_asym c1 c2 (g1 : guar c1) (g2 : guar c2) : 
        leq_guar g1 g2 -> leq_guar g2 g1 -> 
        g1 = g2 :> (state -> state -> Prop).
Proof.
case=>F1 G1 [F2 G2]; move: (leq_coh_asym F1 F2)=>E.
apply: fext=>s; apply: fext=>s'; apply: pext.
split=>H; rewrite -[s]unitR -[s']unitR; case/guar_coh: (H).
- by rewrite E=>C1 C2; apply: G1 H; rewrite unitR.
by rewrite -E=>C1 C2; apply: G2 H; rewrite unitR.
Qed.

(* frame_guar is monotone in both arguments *)

Lemma frame_mono c1 c2 d1 d2 (g1 : guar c1) (g2 : guar c2) : 
        leq_coh d1 d2 -> leq_guar g1 g2 -> 
        leq_guar (frame_guar d1 g1) (frame_guar d2 g2).
Proof.
move=>D [C F]; split; first by apply: star_mono C D.
case: C D=>c Ec2 [d Ed2] s1 s s2 /= C1 C2 H.
case: H C1 C2; case=>t1 [t][t2][->->{s1 s2} G R _] C1 C2.
case/guar_coh: (G)=>T1 T2.
have T : t \In d1 by move/(star_coh_prec R): T1.
have {R} S: s \In star_coh c d.
- move: C1; rewrite Ec2 Ed2 -starA (starCA c) starA.
  by move=>C1; move/(star_coh_prec C1): R. 
case: S C1 C2=>_ [s1][s2][->{s} S1 S2] C1 C2; split=>//.
exists (t1 \+ s1), (t \+ s2), (t2 \+ s1).
rewrite -!joinA (joinCA t); split=>//.
rewrite -!joinA -(joinCA s1) !(joinA _ s1) -!(joinC (t \+ _)) -!joinA in C1 C2.
apply: F G; rewrite Ec2.
- split; last by exists t1, s1.
  by move/cohH/validHR/validHR: C1. 
split; last by exists t2, s1.
by move/cohH/validHR/validHR: C2. 
Qed.

Lemma emp_guar_bot c (g : refl_guar c) : leq_guar (id_guar emp_coh) g.
Proof.
split=>// s1 s s2 C _ G; case: G C=>->->. 
by rewrite unitL; apply/refl_guar_refl.
Qed.

Lemma frame_id c d (g : guar c) : leq_guar g (frame_guar d g).
Proof.
split=>[|s1 s s2 C1 C2 H]; first by apply: star_id.
by split=>//; exists s1, s, s2. 
Qed.

Lemma frame_emp c (g : guar c) : leq_guar (frame_guar emp_coh g) g.
Proof.
split=>[|s1 s s2 C _ [G] _ _]; first by rewrite starC0.
case: G C=>t1 [t][t2][->->{s1 s2} H]; rewrite -!joinA => C. 
by move/(cohK (proj1 (guar_coh H))): C=>->; rewrite !unitR. 
Qed.

Hint Resolve emp_guar_bot frame_id frame_emp.

(* union is a coproduct *)

Lemma union_inl c (g1 g2 : guar c) : leq_guar g1 (union_guar g1 g2).
Proof.
split=>// s1 s s2 C _ H; case/guar_coh: (H)=>S _.
by left; move/(cohK S): C=>->; rewrite !unitR. 
Qed.

Lemma union_inr c (g1 g2 : guar c) : leq_guar g2 (union_guar g1 g2).
Proof. by rewrite union_guarC; apply: union_inl. Qed.

Hint Resolve union_inl union_inr. 

(* universal property *)

Lemma union_case c d (g1 g2 : guar c) (g : guar d) : 
        leq_guar g1 g -> leq_guar g2 g -> leq_guar (union_guar g1 g2) g.
Proof. 
case=>F G1 [_ G2]; split=>// s1 s s2 C1 C2. 
by case; [apply: G1 | apply: G2].
Qed.

(* union is monotone *)

Lemma union_mono c1 c2 (g1 r1 : guar c1) (g2 r2 : guar c2) : 
        leq_guar g1 g2 -> leq_guar r1 r2 ->
        leq_guar (union_guar g1 r1) (union_guar g2 r2).
Proof.
case=>F G1 [_ G2]; split=>// s1 s s2 C1 C2.
by case; [left; apply: G1 | right; apply: G2].
Qed.

(* acqrel is monotone *)

Lemma acqrel_mono ca1 cr1 (g1 : acq_guar ca1) (r1 : rel_guar cr1) 
                  ca2 cr2 (g2 : acq_guar ca2) (r2 : rel_guar cr2) :
        (forall h, leq_guar (g1 h) (g2 h)) ->
        (forall h, leq_guar (r1 h) (r2 h)) ->
        leq_guar (acqrel_guar g1 r1) (acqrel_guar g2 r2).
Proof.
move=>G R; split=>[|s1 s s2 C1 C2 H].
- by apply: star_mono; [case: (G Unit) | case: (R Unit)].
case: H C1 C2; case=>t1 [k1][t2][k2][->->{s1 s2}][h][G1 R1] C C1 C2.
split=>//; case: (guar_coh G1) (guar_coh R1)=>T1 T2 [/cohSO K1 /cohSO K2]. 
case: (G h); case=>ca Ea {G} G.
case: (R h); case=>cr Er {R} R.
have {C} C : s \In star_coh ca cr.
- move: C1; rewrite Ea Er -starA -(starCA cr1) starA=>C1.
  by move/(star_coh_prec C1): C.
case: C C1 C2=>_ [sa][sr][->{s} Sa /cohSO Sr] C1 C2.  
exists (t1 \+ sa), (k1 \+ sr), (t2 \+ sa), (k2 \+ sr).
rewrite -!joinA -!(joinCA sa); split=>//.
rewrite -joinA -(joinCA sa) joinA (joinC (_ \+ sa)) -joinA in C1.
rewrite -joinA -(joinCA sa) joinA (joinC (_ \+ sa)) -joinA in C2.
suff [Ca1 Ca2] : t1 \+ sa \In ca2 /\ t2 \+ sa \In ca2.
- suff [Cr1 Cr2] : k1 \+ sr \In cr2 /\ k2 \+ sr \In cr2.
  - by exists h; split; [apply: G G1 | apply: R R1].
  rewrite joinA joinC in C1; rewrite joinA joinC in C2.
  by split; [move/(star_coh_prec C1): Ca1 | move/(star_coh_prec C2): Ca2].
rewrite Ea; split; split.
- by move/cohH/validHR/validHR: C1.
- by exists t1, sa.  
- by move/cohH/validHR/validHR: C2.
by exists t2, sa.
Qed.

(* do is monotone *)

Lemma do_mono c1 c2 d1 d2 (g1 : guar c1) (g2 : guar c2) 
                          (E1 : c1 = d1) (E2 : c2 = d2) : 
        leq_guar g1 g2 ->
        leq_guar (do_guar E1 g1) (do_guar E2 g2).
Proof. 
case=>E H; split; first by rewrite -E1 -E2.
by move=>s1 s s2 D1 D2; apply: H; rewrite E2. 
Qed.

Lemma do_id c (g : guar c) (pf : c = c) : do_guar pf g = g. 
Proof. by apply/guarE. Qed.

(* often used lemmas about do *)

Lemma do_monoR c1 c2 d2 (g1 : guar c1) (g2 : guar c2) (E : c2 = d2) : 
        leq_guar g1 g2 -> leq_guar g1 (do_guar E g2).
Proof. 
case=>C H; split; first by rewrite -E.
by move=>s1 s s2 D1 D2; apply: H; rewrite E.
Qed.

Lemma do_monoL c1 d1 c2 (g1 : guar c1) (g2 : guar c2) (E : c1 = d1) : 
        leq_guar g1 g2 -> leq_guar (do_guar E g1) g2.
Proof. 
case=>C H; split; first by rewrite -E.
by move=>s1 s s2 D1 D2; apply: H.
Qed.

(* a few frequent lemmas *)

Lemma leq_guar_entangle_inl (W1 W2 : mod) : 
        leq_guar (Mod.internal W1) (entangle_intguar W1 W2).
Proof.
apply: leq_guar_trans (union_inl _ _).
apply: leq_guar_trans (union_inl _ _).
by apply: frame_id.
Qed.

Lemma leq_guar_entangle_inr (W1 W2 : mod) :
        leq_guar (Mod.internal W2) (entangle_intguar W1 W2).
Proof.
apply: leq_guar_trans (union_inl _ _).
apply: leq_guar_trans (union_inr _ _).
by apply: do_mono=>//; apply: frame_id.
Qed.

Lemma leq_guar_apart_inl (W1 W2 : mod) :
        leq_guar (Mod.internal W1) (apart_intguar W1 W2).
Proof.
apply: leq_guar_trans (union_inl _ _).
by apply: frame_id.
Qed.

Lemma leq_guar_apart_inr (W1 W2 : mod) :
        leq_guar (Mod.internal W2) (apart_intguar W1 W2).
Proof.
apply: leq_guar_trans (union_inr _ _).
by apply: do_mono=>//; apply: frame_id.
Qed.

Hint Resolve leq_guar_entangle_inl leq_guar_entangle_inr.
Hint Resolve leq_guar_apart_inl leq_guar_apart_inr.


(*************)
(*************)
(* Injection *)
(*************)
(*************)

(* A module "injects" into another is a form of framing. Essentially, 
the two modules are bisimilar in the sense of following each other's
transitions, but the larger module's coherent states always include
the smaller module's ones. *)

Module Injection.

(* could be stated in terms of guar's, instead of relys *)
(* but I haven't defined the guar corresponding to internal+transfer *)

Structure injects (U V : mod) := Inject {
  extension : mod;

  (* U has a smaller state than V *) 
  _ : Mod.coh V = star_coh (Mod.coh U) (Mod.coh extension);

  (* the internal guarantee of U is simulated by the internal *)
  (* guarantee of V; this is just the leq_guar property *)
  (* but without the condition for leq_coh, which is listed *)
  (* above separately, wrt. explicitly the extension module *)
  _ : forall s1 s2 s, 
        s1 \+ s \In Mod.coh V -> s2 \+ s \In Mod.coh V -> 
        Mod.internal U s1 s2 -> Mod.internal V (s1 \+ s) (s2 \+ s);

  (* V is simulated by U *)
  (* here we can't stick to ordinary rely, because then the property *)
  (* won't hold for any module as the internal transition of V may *) 
  (* involve communication between U and extension *)
  (* OTOH, if we changed the rely in the first simulation to mrely *)
  (* then again the asymetry of entanglement will make it that 
     we don't have leqmod V (entangle U V) *)
  (* so it seems we have to stick with this division that one simulation *)
  (* uses internal tranitions only, but the other simulation must look *)
  (* at communicating transitions as well *)
  _ : forall s1 s1' s2 s2', 
        s1 \In Mod.coh U -> s2 \In Mod.coh U ->
        mguar V (s1 \+ s1') (s2 \+ s2') ->
        mguar U s1 s2 /\ mguar extension s1' s2'}.

Module Exports.
Notation inj_ext := extension. 
Notation injects := injects. 
Notation Inject := Inject.

Lemma cohE U V (w : injects U V) :
         Mod.coh V = star_coh (Mod.coh U) (Mod.coh (inj_ext w)).
Proof. by case: w=>W cohE sL sR; apply: cohE. Qed.

(* we can get rid of one coherence check because of internality *)
Lemma internal_extend U V (w : injects U V) s1 s2 s : 
        s1 \+ s \In Mod.coh V -> 
        Mod.internal U s1 s2 -> Mod.internal V (s1 \+ s) (s2 \+ s).
Proof. 
case: w=>W cohE sL sR C G; apply: sL=>//.
by rewrite cohE (int_guar_extend _ _ G) in C *.
Qed.

Lemma mguar_split U V (w : injects U V) s1 s1' s2 s2' : 
        s1 \In Mod.coh U -> s2 \In Mod.coh U ->
        mguar V (s1 \+ s1') (s2 \+ s2') ->
        mguar U s1 s2 /\ mguar (inj_ext w) s1' s2'.
Proof. by case: w=>W cohE sl sR; apply: sR. Qed.

End Exports.

End Injection.

Export Injection.Exports.



(* additional definitions and lemmas *)
(* these frequently appeared so I codify them *)

(* s is a state which includes s1 *)

Definition extends U V (w : injects U V) s s1 := 
  exists s2, [/\ s = s1 \+ s2, s1 \In Mod.coh U & s \In Mod.coh V].

Lemma ext_coh U V (w : injects U V) s s1 :
        extends w s s1 -> s \In Mod.coh V /\ s1 \In Mod.coh U.
Proof. by case=>s2 []. Qed.

Lemma ext_zagzig U V (w : injects U V) s s' p : 
        extends w (s \> p) s' -> 
        exists s1 s2 p1 p2, 
          [/\ s = s1 \+ s2, p = p1 \+ p2, s' = s1 \> p1 &
              extends w (s \< p) (s1 \< p1)].
Proof.
case=>s1 [E C']; rewrite (cohE w)=>C. 
have C1 : s1 \In Mod.coh (inj_ext w) by rewrite E in C; apply/(star_coh_prec C).
case/cohV/and3P: (C)=>_ _ H.
case/(zag_plusf H): E=>s1' [s2'][p1][p2][????]; subst s p s' s1.
move/coh_zigzag: C'=>C'; move/coh_zigzag: C1=>C1.
exists s1', s2', p1, p2; split=>//.
exists (s2' \< p2); split=>//; last by rewrite coh_zigzag (cohE w).
by apply: zig_plus; [case/cohV/and3P: C' | case/cohV/and3P: C1].
Qed.

Lemma ext_zigzag U V (w : injects U V) s s' p : 
        extends w (s \< p) s' -> 
        exists s1 s2 p1 p2, 
          [/\ s = s1 \+ s2, p = p1 \+ p2, s' = s1 \< p1 &
              extends w (s \> p) (s1 \> p1)].
Proof.
case=>s1 [E C']; rewrite (cohE w)=>C.
have C1 : s1 \In Mod.coh (inj_ext w) by rewrite E in C; apply/(star_coh_prec C).
case/cohV/and3P: (C)=>H _ _.
case/(zig_plusf H): E=>s1' [s2'][p1][p2][????]; subst s p s' s1.
move/coh_zigzag: C'=>C'; move/coh_zigzag: C1=>C1.
exists s1', s2', p1, p2; split=>//.
exists (s2' \> p2); split=>//; last by rewrite -coh_zigzag (cohE w). 
by apply: zag_plus; [case/cohV/and3P: C' | case/cohV/and3P: C1].
Qed.

Lemma ext_local U V (w : injects U V) s1 s2 s : 
        Mod.internal U s1 s2 -> 
        s1 \+ s \In Mod.coh V -> s2 \+ s \In Mod.coh V.
Proof. by move=>G; rewrite (cohE w) (int_guar_extend _ _ G). Qed.

Lemma ext_precise U V (w : injects U V) s1 s2 s : 
        s1 \In Mod.coh U ->  
        extends w (s1 \+ s) s2 -> s1 = s2.
Proof. by move=>C1 [s'][E] C2 C; apply: coh_prec (cohV C) E C1 C2. Qed.

(***********************************************)
(* Injection for concrete constructors         *)
(*                                             *)
(* should be canonized and eventually used for *) 
(* automatic search for injection witnesses    *)
(***********************************************)

(* first a few helper lemmas about extending *)

Lemma apart_extend (W1 W2 : mod) s1 s2 s : 
        s1 \+ s \In star_coh (Mod.coh W1) (Mod.coh W2) ->
        s2 \+ s \In star_coh (Mod.coh W1) (Mod.coh W2) ->
        Mod.internal W1 s1 s2 -> 
        Mod.internal (apart W1 W2) (s1 \+ s) (s2 \+ s).
Proof. by move=>C1 C2 G; left; split=>//; exists s1, s, s2. Qed.

Lemma entangle_extend (W1 W2 : mod) s1 s2 s : 
        s1 \+ s \In star_coh (Mod.coh W1) (Mod.coh W2) ->
        s2 \+ s \In star_coh (Mod.coh W1) (Mod.coh W2) ->
        Mod.internal W1 s1 s2 -> 
        Mod.internal (entangle W1 W2) (s1 \+ s) (s2 \+ s).
Proof. by move=>C1 C2 G; left; apply: apart_extend. Qed.

Lemma lefts_extend b (W1 W2 : mod) h s1 s2 s ct : 
        s1 \+ s \In star_coh (Mod.coh W1) (Mod.coh W2) ->
        s2 \+ s \In star_coh (Mod.coh W1) (Mod.coh W2) -> 
        ct \In Mod.cores W1 ->
        get b ct h s1 s2 ->
        exists ct', ct' \In entangle_lefts W1 W2 /\
                    get b ct' h (s1 \+ s) (s2 \+ s).
Proof. 
move=>C1 C2 H G; exists (frame_coreL (Mod.coh W2) ct).
split; first by apply/Mem_map. 
by case: b G=>G; split=>//; exists s1, s, s2.  
Qed.

Lemma rights_extend b (W1 W2 : mod) h s1 s2 s ct : 
        s1 \+ s \In star_coh (Mod.coh W1) (Mod.coh W2) ->
        s2 \+ s \In star_coh (Mod.coh W1) (Mod.coh W2) -> 
        ct \In Mod.cores W2 ->
        get b ct h s1 s2 ->
        exists ct', ct' \In entangle_rights W1 W2 /\
                    get b ct' h (s1 \+ s) (s2 \+ s).
Proof. 
move=>C1 C2 H G; exists (frame_coreR (Mod.coh W1) ct).
rewrite starC in C1 C2; split; first by apply/Mem_map. 
by case: b G=>G; split=>//; exists s1, s, s2.  
Qed.

(* a few helper lemmas on splitting *)

Lemma apart_split (W1 W2 : mod) (i1 i2 j1 j2 : state) : 
        i1 \In Mod.coh W1 -> j1 \In Mod.coh W1 -> 
        Mod.internal (apart W1 W2) (i1 \+ i2) (j1 \+ j2) ->
        [\/ Mod.internal W1 i1 j1 /\ i2 = j2 | 
            i1 = j1 /\ Mod.internal W2 i2 j2 ].
Proof.
move=>Ci1 Cj1 G; case: (guar_coh G)=>Ci Cj.
case: G=>G; [left | right]. 
- case: G; case=>s1 [s2][s1'][Ei Ej] G _ _.  
  move: (coh_prec (cohV Ci) Ei Ci1 (proj1 (guar_coh G)))=>?; subst s1.
  move/(joinsK (cohV Ci)): Ei=>?; subst s2.
  move: (coh_prec (cohV Cj) Ej Cj1 (proj2 (guar_coh G)))=>?; subst s1'.
  by move/(joinsK (cohV Cj)): Ej.
case: G; case=>s2 [s1][s2'][Ei Ej] G _ _; rewrite /= starC in Ci Cj.
rewrite -(joinC i2) in Ei Ci; rewrite -(joinC j2) in Ej Cj.
move/(star_coh_prec Ci): (Ci1)=>Ti.
move: (coh_prec (cohV Ci) Ei Ti (proj1 (guar_coh G)))=>?; subst s2.
move/(star_coh_prec Cj): (Cj1)=>Tj.
move: (coh_prec (cohV Cj) Ej Tj (proj2 (guar_coh G)))=>?; subst s2'.
move/(joinsK (cohV Ci)): Ei=>?; subst s1.
by move/(joinsK (cohV Cj)): Ej. 
Qed.

Lemma guarsum_split (W1 W2 : mod) (i1 i2 j1 j2 : state) : 
        i1 \In Mod.coh W1 -> j1 \In Mod.coh W1 -> 
        guar_sum (Mod.chans W1) (Mod.chans W2) (i1 \+ i2) (j1 \+ j2) ->
        exists ch1 ch2 h, 
          [/\ ch1 \In Mod.chans W1, ch2 \In Mod.chans W2 &
              (Chan.acq ch1 h i1 j1 /\ Chan.rel ch2 h i2 j2) \/
              (Chan.rel ch1 h i1 j1 /\ Chan.acq ch2 h i2 j2)].
Proof.
move=>Ci1 Cj1 G; case: (guar_coh G)=>Ci Cj.
case: G=>x [y][/Mem_map_inv][ct1][-> H1] /Mem_map_inv [ct2][->] H2 G.
exists (Chan ct1), (Chan ct2); case: G Ci Cj.
- case; case=>s1 [s2][s1'][s2'][Ei Ej][h][G1 G2] _ Ci Cj.
  move: (coh_prec (cohV Ci) Ei Ci1 (proj1 (guar_coh G1)))=>?; subst s1.
  move/(joinsK (cohV Ci)): Ei=>?; subst s2.
  move: (coh_prec (cohV Cj) Ej Cj1 (proj2 (guar_coh G1)))=>?; subst s1'.
  move/(joinsK (cohV Cj)): Ej=>?; subst s2'.
  by exists h; split; try by [apply: Mem_map]; left. 
case; case=>s2 [s1][s2'][s1'][Ei Ej][h][G2 G1] _ Ci Cj.
rewrite (joinC s2) in Ei; rewrite (joinC s2') in Ej.
move: (coh_prec (cohV Ci) Ei Ci1 (proj1 (guar_coh G1)))=>?; subst s1.
move/(joinsK (cohV Ci)): Ei=>?; subst s2.
move: (coh_prec (cohV Cj) Ej Cj1 (proj2 (guar_coh G1)))=>?; subst s1'.
move/(joinsK (cohV Cj)): Ej=>?; subst s2'.
by exists h; split; try by [apply: Mem_map]; right.
Qed.

Lemma entangle_split (W1 W2 : mod) (i1 i2 j1 j2 : state) : 
        i1 \In Mod.coh W1 -> j1 \In Mod.coh W1 -> 
        Mod.internal (entangle W1 W2) (i1 \+ i2) (j1 \+ j2) ->
        [\/ Mod.internal W1 i1 j1 /\ i2 = j2, 
            i1 = j1 /\ Mod.internal W2 i2 j2 |
            exists ch1 ch2 h, 
              [/\ ch1 \In Mod.chans W1, ch2 \In Mod.chans W2 &
                 (Chan.acq ch1 h i1 j1 /\ Chan.rel ch2 h i2 j2) \/
                 (Chan.rel ch1 h i1 j1 /\ Chan.acq ch2 h i2 j2)]].
Proof.
move=>Ci1 Cj1; case. 
- by case/(apart_split Ci1 Cj1)=>H; [apply: Or31 | apply: Or32].
by move/(guarsum_split Ci1 Cj1)=>H; apply: Or33. 
Qed.

Lemma lefts_split b (W1 W2 : mod) h (i1 i2 j1 j2 : state) ct : 
        i1 \In Mod.coh W1 -> j1 \In Mod.coh W1 ->
        ct \In entangle_lefts W1 W2 ->
        get b ct h (i1 \+ i2) (j1 \+ j2) ->
        exists ch, [/\ ch \In Mod.chans W1, get b ch h i1 j1 & i2 = j2].
Proof.
move=>Ci Cj /Mem_map_inv [ct'][->{ct}] H.
case E: ct'=>[a r]; case: b;
case; case=>s1 [s2][s1'][Ei Ej] G Ci1 Cj1;
move: (coh_prec (cohV Ci1) Ei Ci (proj1 (guar_coh G)))=>?; subst s1;
move/(joinsK (cohV Ci1)): Ei=>?; subst s2;
move: (coh_prec (cohV Cj1) Ej Cj (proj2 (guar_coh G)))=>?; subst s1';
move/(joinsK (cohV Cj1)): Ej ->;
by exists (Chan ct'); rewrite {4}E; split=>//; apply: Mem_map. 
Qed.

Lemma rights_split b (W1 W2 : mod) h (i1 i2 j1 j2 : state) ct : 
        i1 \In Mod.coh W1 -> j1 \In Mod.coh W1 ->
        ct \In entangle_rights W1 W2 ->
        get b ct h (i1 \+ i2) (j1 \+ j2) ->
        exists ch, [/\ ch \In Mod.chans W2, i1 = j1 & get b ch h i2 j2].
Proof.
move=>Ci Cj /Mem_map_inv [ct'][->{ct}] H.
case E: ct'=>[a r]; case: b;
case; case=>s2 [s1][s2'][Ei Ej] G Ci1 Cj1;
rewrite starC in Ci1 Cj1; rewrite -!(joinC s1) in Ei Ej;
have C : s1 \In Mod.coh W1
  by [rewrite Ei in Ci1; move/(star_coh_prec Ci1): (proj1 (guar_coh G))];
move: (coh_prec (cohV Ci1) Ei Ci C)=>?; subst s1;
move/(joinsK (cohV Ci1)): Ei=>?; subst s2;
move: (coh_prec (cohV Cj1) Ej Cj C)=>?; subst j1;
move/(joinsK (cohV Cj1)): Ej=>?; subst s2';
by exists (Chan ct'); rewrite {4}E; split=>//; apply: Mem_map. 
Qed.

Implicit Arguments lefts_split [W1 W2 h i1 i2 j1 j2 ct].
Implicit Arguments rights_split [W1 W2 h i1 i2 j1 j2 ct].

(* entanglement is an injection of the left module *)

Module InjectEntangleLeft.
Section InjectEntangleLeft. 
Variables U V : mod.

Lemma inl_cohE : Mod.coh (entangle U V) = star_coh (Mod.coh U) (Mod.coh V).
Proof. by []. Qed.

Lemma inl_extend s1 s2 s : 
        s1 \+ s \In Mod.coh (entangle U V) -> 
        s2 \+ s \In Mod.coh (entangle U V) ->
        Mod.internal U s1 s2 ->
        Mod.internal (entangle U V) (s1 \+ s) (s2 \+ s).
Proof. by apply: entangle_extend. Qed.

Lemma inl_split s1 s1' s2 s2' : 
        s1 \In Mod.coh U -> s2 \In Mod.coh U ->
        mguar (entangle U V) (s1 \+ s1') (s2 \+ s2') ->
        mguar U s1 s2 /\ mguar V s1' s2'.
Proof.
move=>C1 C2 G; case: (mguar_coh G)=>C1' C2'.
move/(star_coh_prec C2'): (C2)=>D2'.
case: G; last first.
- case=>_ [h][/Mem_map_inv][ct][->] Ch; case=>H.
  - case/(lefts_split acq C1 C2 Ch): H=>ch [H G] ->.
    split; last by apply: mguar_refl.
    by right; exists ch, h; split=>//; left. 
  case/(lefts_split rel C1 C2 Ch): H=>ch [H G] ->.
  split; last by apply: mguar_refl.
  by right; exists ch, h; split=>//; right. 
case/(entangle_split C1 C2); case.
- by move=>H ->; split; [left | apply: mguar_refl].
- by move=>-> H; split; [apply: mguar_refl | left].
move=>ch1 [ch2][h][Ch1 Ch2]; case; case=>H1 H2; split; right.
- by exists ch1, h; split=>//; left. 
- by exists ch2, h; split=>//; right. 
- by exists ch1, h; split=>//; right. 
by exists ch2, h; split=>//; left. 
Qed.

End InjectEntangleLeft.

Module Exports.

(* i can't make this one canonical for entangle *)
(* as there'll also be one for extending on the right *)
Definition entangle_extL U V := 
  Inject (inl_cohE U V) (@inl_extend U V) (@inl_split U V).

End Exports.

End InjectEntangleLeft.

Export InjectEntangleLeft.Exports.

(* entanglement is an injection of the right module *)

Module InjectEntangleRight.
Section InjectEntangleRight. 
Variables U V : mod.

Lemma inr_cohE : Mod.coh (entangle V U) = star_coh (Mod.coh U) (Mod.coh V).
Proof. by rewrite starC. Qed.

Lemma inr_extend s1 s2 s : 
        s1 \+ s \In Mod.coh (entangle V U) -> 
        s2 \+ s \In Mod.coh (entangle V U) ->
        Mod.internal U s1 s2 ->
        Mod.internal (entangle V U) (s1 \+ s) (s2 \+ s).
Proof. by rewrite /= entangle_intC starC; apply: entangle_extend. Qed.

Lemma inr_split s1 s1' s2 s2' : 
        s1 \In Mod.coh U -> s2 \In Mod.coh U ->
        mguar (entangle V U) (s1 \+ s1') (s2 \+ s2') ->
        mguar U s1 s2 /\ mguar V s1' s2'.
Proof.
rewrite (joinC s1) (joinC s2)=>D1 D2 R; case: (mguar_coh R)=>C1' C2'.
move/(star_coh_prec C1'): (D1)=>C1; move/(star_coh_prec C2'): (D2)=>C2.
case: R; last first.  
- case=>_ [h][/Mem_map_inv][ct][->] Ch; case=>H.
  - case/(lefts_split acq C1 C2 Ch): H=>ch [H G] ->.
    split; first by apply: mguar_refl.
    by right; exists ch, h; split=>//; left. 
  case/(lefts_split rel C1 C2 Ch): H=>ch [H G] ->.
  split; first by apply: mguar_refl.
  by right; exists ch, h; split=>//; right. 
case/(entangle_split C1 C2); case.
- by move=>H ->; split; [apply: mguar_refl | left].
- by move=>-> H; split; [left | apply: mguar_refl].
move=>ch1 [ch2][h][Ch1 Ch2]; case; case=>H1 H2; split; right.
- by exists ch2, h; split=>//; right. 
- by exists ch1, h; split=>//; left. 
- by exists ch2, h; split=>//; left. 
by exists ch1, h; split=>//; right. 
Qed.

End InjectEntangleRight.

Module Exports.

(* i can't make this one canonical for entangle *)
(* as there'll also be one for extending on the right *)
Definition entangle_extR U V := 
  Inject (inr_cohE U V) (@inr_extend U V) (@inr_split U V).

End Exports.

End InjectEntangleRight.

Export InjectEntangleRight.Exports.


(* apartness composition is an injection of the left module *)

Module InjectApartLeft.
Section InjectApartLeft. 
Variables U V : mod.

Lemma inl_cohE : Mod.coh (apart U V) = star_coh (Mod.coh U) (Mod.coh V).
Proof. by []. Qed.

Lemma inl_extend s1 s2 s : 
        s1 \+ s \In Mod.coh (apart U V) -> 
        s2 \+ s \In Mod.coh (apart U V) ->
        Mod.internal U s1 s2 ->
        Mod.internal (apart U V) (s1 \+ s) (s2 \+ s).
Proof. by apply: apart_extend. Qed.

Lemma inl_split s1 s1' s2 s2' : 
        s1 \In Mod.coh U -> s2 \In Mod.coh U ->
        mguar (apart U V) (s1 \+ s1') (s2 \+ s2') ->
        mguar U s1 s2 /\ mguar V s1' s2'.
Proof.
move=>C1 C2 G; case: (mguar_coh G)=>C1' C2'.
move/(star_coh_prec C2'): (C2)=>D2'.
case: G.
- case/(apart_split C1 C2); case.
  - by move=>H ->; split; [left | apply: mguar_refl].
  by move=>-> H; split; [apply: mguar_refl | left].
case=>_ [h][/Mem_map_inv][ct][->] Ch; case=>H.
- case/Mem_cat: Ch=>Ch.
  - case/(lefts_split acq C1 C2 Ch): H=>ch [H G ->].
    split; last by apply: mguar_refl.
    by right; exists ch, h; split=>//; left. 
  case/(rights_split acq C1 C2 Ch): H=>ch [H -> G].  
  split; first by apply: mguar_refl.
  by right; exists ch, h; split=>//; left. 
case/Mem_cat: Ch=>Ch.
- case/(lefts_split rel C1 C2 Ch): H=>ch [H G ->].
  split; last by apply: mguar_refl.
  by right; exists ch, h; split=>//; right. 
case/(rights_split rel C1 C2 Ch): H=>ch [H -> G].  
split; first by apply: mguar_refl.
by right; exists ch, h; split=>//; right.
Qed.

End InjectApartLeft.

Module Exports.

(* i can't make this one canonical for entangle *)
(* as there'll also be one for extending on the right *)
Definition apart_extL U V := 
  Inject (inl_cohE U V) (@inl_extend U V) (@inl_split U V).

End Exports.

End InjectApartLeft.

Export InjectApartLeft.Exports.


(* apartness composition is an injection of the right module *)

Module InjectApartRight.
Section InjectApartRight. 
Variables U V : mod.

Lemma inr_cohE : Mod.coh (apart V U) = star_coh (Mod.coh U) (Mod.coh V).
Proof. by rewrite starC. Qed.

Lemma inr_extend s1 s2 s : 
        s1 \+ s \In Mod.coh (apart V U) -> 
        s2 \+ s \In Mod.coh (apart V U) ->
        Mod.internal U s1 s2 ->
        Mod.internal (apart V U) (s1 \+ s) (s2 \+ s).
Proof. by rewrite /= apart_intC starC; apply: apart_extend. Qed.

Lemma inr_split s1 s1' s2 s2' : 
        s1 \In Mod.coh U -> s2 \In Mod.coh U ->
        mguar (apart V U) (s1 \+ s1') (s2 \+ s2') ->
        mguar U s1 s2 /\ mguar V s1' s2'.
Proof.
rewrite (joinC s1) (joinC s2)=>D1 D2 R; case: (mguar_coh R)=>C1' C2'.
move/(star_coh_prec C1'): (D1)=>C1; move/(star_coh_prec C2'): (D2)=>C2. 
case: R.
- case/(apart_split C1 C2); case.
  - by move=>H ->; split; [apply: mguar_refl | left].
  by move=>-> H; split; [left | apply: mguar_refl].
case=>_ [h][/Mem_map_inv][ct][->] Ch; case=>H.
- case/Mem_cat: Ch=>Ch.
  - case/(lefts_split acq C1 C2 Ch): H=>ch [H G] ->.
    split; first by apply: mguar_refl.
    by right; exists ch, h; split=>//; left. 
  case/(rights_split acq C1 C2 Ch): H=>ch [H -> G].  
  split; last by apply: mguar_refl.
  by right; exists ch, h; split=>//; left. 
case/Mem_cat: Ch=>Ch.
- case/(lefts_split rel C1 C2 Ch): H=>ch [H G] ->.
  split; first by apply: mguar_refl.
  by right; exists ch, h; split=>//; right. 
case/(rights_split rel C1 C2 Ch): H=>ch [H -> G].  
split; last by apply: mguar_refl.
by right; exists ch, h; split=>//; right. 
Qed.

End InjectApartRight.

Module Exports.

(* i can't make this one canonical for entangle *)
(* as there'll also be one for extending on the right *)
Definition apart_extR U V := 
  Inject (inr_cohE U V) (@inr_extend U V) (@inr_split U V).

End Exports.

End InjectApartRight.

Export InjectApartRight.Exports.

