Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels perms ordtype.
Require Import Eqdep prelude domains idynamic pcm unionmap heap codes feaps state.
Require Import zigzag worlds actions rely inject retract schedule.
Require Import process always model privstate footprint.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(********************************************)
(* Derived logical rules for easier proving *)
(********************************************)

Lemma vrf_coh W A (e : ST W A) i r : 
        (i \In Mod.coh W -> verify i e r) -> verify i e r.
Proof. by move=>H C; apply: H. Qed.

Lemma vrf_mono W A (e : ST W A) i (r1 r2 : cont A) : 
        r1 <== r2 -> verify i e r1 -> verify i e r2. 
Proof. by move=>T H1 C t; move/(H1 C); apply: aft_imp=>v m _; apply: T. Qed.

(* I can use a Hoare lemma for sequential composition *)
(* to avoid duplicating proofs in the val* and bnd* cases *)

Lemma prog_unfin W A (s : spec A) (e : STsep W s) : 
        Unfinished \In STsep.prog_of e.
Proof. by case: e; case. Qed.

Lemma seq_do W A B (e1 : ST W A) (e2 : A -> ST W B) i 
             (q : cont A) (r : cont B) : 
        verify i e1 q -> 
        (forall y j, q y j -> j \In Mod.coh W  -> verify j (e2 y) r) ->
        verify i (bind e1 e2) r.
Proof.
move=>H1 H2 Ci t [->|[t'][H3 H4]]. 
- by apply: alw_unfin=>//; move/alw_coh: (H1 Unfinished (prog_unfin e1)). 
by apply: aft_bnd H3 _; move/(H1 Ci): H4; apply: aft_imp=>y j Cj H; apply: H2.
Qed.

Implicit Arguments seq_do [W A B e1 e2 i r].

(* a better lemma for stepping *)
(* as it doesn't require to provide a q *)

Lemma step W A B (e1 : ST W A) (e2 : A -> ST W B) i (r : cont B) : 
        verify i e1 (fun y m => verify m (e2 y) r) ->
        verify i (bind e1 e2) r.
Proof. by move=>H; apply: (seq_do (fun y m => verify m (e2 y) r)). Qed.

(* Does the opposite direction hold? *)

Lemma step' W A B (e1 : ST W A) (e2 : A -> ST W B) i (r : cont B) : 
        verify i (bind e1 e2) r ->
        verify i e1 (fun y m => verify m (e2 y) r).
Proof.
(* probably, but I don't want to bother now *)
Abort.



(* generic do lemmas *)

Section DoLemmas.
Variables (W : mod) (A B : Type)(T : encoded_type).

Lemma val_do (e : ST W A) i (k : cont A) : 
         pre_of e i -> 
         (forall x m, post_of e i x m -> m \In Mod.coh W -> k x m) ->
         verify i e k.
Proof.
case: e=>s [e] /= H H1 H2 Ci t H3.
apply: aft_imp (H i t H1 Ci H3). 
by move=>v m Cm H4; apply: H2.
Qed.

End DoLemmas.


Section ParLemma.
Variables (W : mod) (A1 A2 : Type). 

Lemma par_do lp1 lp2 sh ep (e1 : ST W A1) (e2 : ST W A2)             
             (r1 : cont A1)  (r2 : cont A2) (r : cont (A1 * A2)) : 
        verify ([lp1, sh, ep] \> lp2) e1 r1 ->            
        verify ([lp2, sh, ep] \> lp1) e2 r2 ->
        (forall v lp1' lp2' sh' ep',
             [lp1' \* lp2', sh', ep'] \In Mod.coh W ->
             r1 v.1 ([lp1', sh', ep'] \> lp2') -> 
             r2 v.2 ([lp2', sh', ep'] \> lp1') ->
             r v [lp1' \* lp2', sh', ep']) ->
        verify [lp1 \* lp2, sh, ep] (par e1 e2) r.
Proof.
move=>H1 H2 H Ci t [->|[t1][t2][->] T1 T2]; first by apply: alw_unfin.
apply: aft_imp (aft_par (H1 _ _ T1) (H2 _ _ T2)); 
  try by apply/coh_zigzag; rewrite zigE // zipC. 
move=>v m Cm [lp1'][lp2'][sh'][ep'][E].
by rewrite {m}E in Cm *; apply: H Cm.
Qed.

End ParLemma.


Section RetLemmas.
Variables (W : mod) (A : Type).

Lemma val_ret i (v : A) (r : cont A) : 
       (forall m, menv_steps W i m -> r v m) ->       
       verify i (ret (W:=W) v) r. 
Proof. by move=>H; apply: val_do=>// x m []; move/H=>X ->. Qed.

End RetLemmas.



Section HideLemmas.
Variables (W V : mod) (U A B : Type) (f : refinement U W V) (gL : U).

Lemma val_hide i i' (e : ST V A) (r : cont A):      
        f gL i i' -> 
        verify i' e (fun y m' => forall gL' m, 
          f gL' m m' -> r y m) ->
        verify i (hide f gL e) r. 
Proof.
move=>Ri; case/rf_coh: (Ri)=>Ci Ci'.
move=>H1 _ t [->|[p][H2 ->{t}]]; first by apply: alw_unfin. 
apply: aft_imp (aft_hide Ri (H1 Ci' _ H2))=>v m _ [m'][gL'][Rm] H. 
by apply: H Rm.
Qed.


End HideLemmas.


(* refinement/hiding in the common case when the decorated world *)
(*  is an extension of priv *)
Section PrivHideLemmas. 
Variables (hp : nat) (V W : mod) (A T : Type).
Variables (d : decorate W T).
Variables (gL : T).
Notation U := (entangle (Priv hp) V).

(* the lemma is still too complicated; consider how to simplify *)
(* things to look for:  *)
(* 1. can we change definition of decorate to remove the need for gE *)
(*    as it will always be instantiated with a unit feap for W *)
(* 2. can we remove the precondition on valid (hp@sf->hL \+ i \+ j) *)
(*
The latter is needed so that we can derive that the labels of W and U
are disjoint. Maybe d (gL, gE) h j and the validity condition should
be packaged together in some new version of decorate. Basically,
decorate should have as input not h, but a precise predicate which
selects uniquely h out of hL. Make things look good in SL-style specs.
*)

(*
Lemma val_privhide i j (hL h : heap) (e1 : ST (entangle U W) A) (r : cont A) : 
        d (gL, gE) h j ->
        valid (hp@sf-> hL \+ i \+ j) ->
        verify (hp@sf-> hL \+ i \+ j) e1 
           (fun y m' => forall i' j' hL' h' gL',
               m' = hp@sf->hL' \+ i' \+ j' ->
               d (gL', gE) h' j' ->
               r y (hp@sf-> (hL' \+ h') \+ i')) ->
        verify (hp@sf-> (hL \+ h) \+ i) (priv_hide d e gL gE e1) r.
Proof.
rewrite /priv_hide; set f := (PrivEntangleRefine hp V d e).
move=>D Vl Vf; apply: vrf_coh=>C.
case/privent_stE2: C Vl Vf (C)=>hE [s'][->{i} C'].
rewrite !joinA -!stE1=>Vl Vf C. 
suff F: f (gL, gE) (hp ->> [hL \+ h, tt, hE] \+ s')
                   (hp ->> [hL, tt, hE] \+ s' \+ j).
- apply: val_hide F _; apply: vrf_mono Vf=>y m' H gL' m F. 
  case: F H=>hL' [h'][hs'][j'][->{m}->{m'}] D' _ _.
  move/(_ (hp@jn->tt \+ hp@ot->hs'.1 \+ hs'.2) _ hL' _ _ _ D').
  by rewrite /ctx !stE1 !joinA; apply. 
exists hL, h, (hE, s'), j; split=>//.
have Cl : ctx hp hL (hE, s') \In Mod.coh U by apply: (ctxCL star_id (h:=h)). 
split; last by exists (ctx hp hL (hE, s')), j; split=>//; move/dec_coh: D.
move: (cohH C).
rewrite /validH (flatsUn (cohV C)) (flatsUn Vl) (flatsUn (cohV Cl)).
by rewrite !priv_flatsE /= (dec_flat D) !(joinAC _ h).
Qed.
*)

Lemma val_privhide' i j (hL hE : heap) (e1 : ST (entangle U W) A) (r : cont A) : 
        d gL j ->
        valid (hp->>[hL, tt, hE] \+ i \+ j) ->
        verify (hp->>[hL, tt, hE] \+ i \+ j) e1 
           (fun y m' => forall i' j' hL' hE' gL',
               m' = hp->>[hL', tt, hE'] \+ i' \+ j' ->
               d gL' j' ->
               r y (hp->>[hL' \+ flats j', tt, hE'] \+ i')) ->
        verify (hp->>[hL \+ flats j, tt, hE] \+ i) (priv_hide d gL e1) r.
Proof.
rewrite /priv_hide; set f := (PrivEntangleRefine hp V d).
set h := flats j.
move=>D Vl Vf; apply: vrf_coh=>C.
suff F: f gL (hp ->> [hL \+ h, tt, hE] \+ i)
             (hp ->> [hL, tt, hE] \+ i \+ j).
- apply: val_hide F _; apply: vrf_mono Vf=>y m' H gL' m F. 
  case: F H=>hL' [[hE' i']][j'][->{m}->{m'}] D' _ _; rewrite /ctx /=.
  by move/(_ i' j' hL' hE' gL' _ D')=>H; apply: H.
exists hL, (hE, i), j; split=>//.
have Cl : ctx hp hL (hE, i) \In Mod.coh U by apply: (ctxCL star_id (h:=h)). 
split; last by exists (ctx hp hL (hE, i)), j; split=>//; move/dec_coh: D.
move: (cohH C).
rewrite /validH (flatsUn (cohV C)) (flatsUn Vl) (flatsUn (cohV Cl)).
by rewrite !priv_flatsE /= !(joinAC _ h).
Qed.

End PrivHideLemmas.



Section ActionLemmas.
Variables (W : mod) (A B : Type).
Notation guar := (Mod.internal W).
Variable a : int_action guar A. 

Lemma val_act i (r : cont A) :
        (forall j, menv_steps W i j -> j \In asafe a /\
           forall y k m, astep j a k y -> menv_steps W k m -> r y m) ->
        verify i (act a) r. 
Proof.
move=>H1; apply: val_do=>//; first by move=>j; case/H1. 
by move=>x m [j][k][/H1][_] H2 H3 H4 _; apply: H2 H3 H4. 
Qed.

End ActionLemmas. 



Section ExtendLemmas.
Variables (W V : mod) (A : Type) (w : injects V W).
Notation W2 := (inj_ext w).

Variable (e1 : ST V A).

Lemma val_extend i j (r : cont A) : 
        i \In Mod.coh V -> 
        verify i e1 (fun x i' => forall j', 
          i' \+ j' \In Mod.coh W -> menv_steps W2 j j' -> r x (i' \+ j')) ->
        verify (i \+ j) (extend w e1) r.
Proof.
move=>Ci H C t [->|[t' [H' ->{t}]]]; first by apply: alw_unfin. 
move/aft_extend: {H H'} (H Ci _ H'); move/(_ _ w _ C). 
apply: aft_imp=>v s Cs [i'][j'][E] Ci' S'.
by rewrite {s}E in Cs *; apply.
Qed.

End ExtendLemmas.


(*
(* I want framing lemmas here *)
(* as proving the specific actions lemmas is hell *)
(* I need some framing for programs in general, not just for *)
(* actions. But, let's start with actions *)

Section ActFrame.
Variables (W1 : MOD) (W2 : seq MOD). 
Notation W := (SEntangle W1 W2).
Variables (A : Type) (a : int_action (Mod.internal W1) A).
Variable qf : leq_guar (Mod.internal W1) (Mod.internal W).

Lemma act_frame i j (r : cont A) :
        i \In Mod.coh W1 ->
        (* stability of pre *)
        (forall i' j', i' \In Mod.coh W1 -> 
                       menv_steps W (i \+ j) (i' \+ j') -> i' \In asafe a) ->
        (* if we prove r is stable, we don't need senv_steps here *)
        verify i (act a) (fun y i' => 
          forall j', senv_steps W2 j j' -> r y (i' \+ j')) ->
        verify (i \+ j) (act (FrameAct a qf)) r. 
Proof.
move=>Ci Sf H C t; case=>->; first by apply: alw_unfin. 
apply: alw_act=>// s1 E1; split=>//.
- split; last by case/menvs_coh: E1.
  case/menvs_coh: E1 (E1)=>_; rewrite ssE.
  case=>_ [i'][j'][->{s1}] D _ E.
  by exists i', j'; split=>//; apply: Sf D E. 
have {Sf} : (act a) (Act a) by right.
move/H=>{H}; case/(_ Ci [:: ChoiceAct])=>_ H.
move=>s2 v s3 St; case: St E1.
case=>i1 [j'][i2][->{s1}->{s2}] St _ _ E1 E2 _ [<-].
case/menvs_coh: E2 (E2)=>_; rewrite /= ssE.
case=>_ [i3][j''][->{s3}] Ci3 Cj'' E2.
case/(senvsUn Ci (proj1 (ast_coh St))): E1=>Mi Sj.
case/(senvsUn (proj2 (ast_coh St)) Ci3): E2=>Si2 Sj'.
move/H: Mi=>{H} [X _]. 
have {X} X : tstep i1 (Act a) ChoiceAct i2 (Ret v). 
- by split=>//; exists v. 
by case/(_ _ _ X)=>_; apply=>//; apply: senvs_trans Sj Sj'. 
Qed.

Variables (B : Type) (s2 : A -> spec B) (e : forall x, STsep W (s2 x)).

Lemma act_frame_bnd i j (r : cont B) :
        i \In Mod.coh W1 ->
        (* stability of pre *)
        (forall i' j', i' \In Mod.coh W1 ->
                       menv_steps W (i \+ j) (i' \+ j') -> i' \In asafe a) ->
        (* if we prove r is stable, we don't need senv_steps here *)
        verify i (act a) (fun y i' => 
          forall j', senv_steps W2 j j' -> verify (i' \+ j') (e y) r) ->
        verify (i \+ j) (bind (act (FrameAct a qf)) e) r. 
Proof.
move=>Ci Sf H; apply: (seq_do (fun y i => verify i (e y) r))=>//.
by apply: act_frame.
Qed.

End ActFrame.
*)



(**************************)
(* SCSL Primitive Actions *)
(**************************)
       
Section ReadLemmas. 
Variables (A : Type) (hp : nat).
Notation V := (Priv hp).

(* it's annoying i have to switch from hp@sf notation to full notation *)
(* because of coherence conditions, which require reasoning about all components *)
(* makes on wonder if it's worth working with the abbreviated notation *)

(* notice that the continuation doesn't checks r on a specific state *)
(* which is specified concretely, rather than by menv_steps *)
(* thus, if r is not actually stable under W's menv_stepping *)
(* it won't be provable. But, we don't have to issue a separate *)
(* VC showing that r is actually stable *)

(* If we used val_read on a larger world, without framing the larger world *)
(* out, then we'd need stability. But on a small world, such as here *)
(* all the interference can be described, so no need to reason about stability *)
(* except when framing out worlds *)

Lemma val_read' (x : ptr) v hL s (r : cont A) :
        (forall s', hp@sf-> (x :-> v \+ hL) \+ s' \In Mod.coh V -> 
                    r v (hp@sf-> (x :-> v \+ hL) \+ s')) ->  
        verify (hp@sf-> (x :-> v \+ hL) \+ s) (act (Read A hp x)) r.
Proof.
move=>H; apply: val_act=>// i M; case/menvs_loc_eq1: M (M)=>s' ->{i} M.
split=>[|y j m {M}].
  (* this is a proof that precondition is stable under rely^* *)
  (* this should be a general property of actions *)
- case/menvs_coh: M=>_ C'; split=>//.
  case/priv_stE2: C'=>hE ->; exists hL, hE, v. 
  by rewrite joinA -stE1.
case; case=>hL1 [hE][F] ->{j} C2; rewrite stE1 -joinA in F.
case/(sfcancel (cohV C2)): F (C2)=>F _ ->{s' C2} C2.
rewrite joinA -stE1 in C2.
case/(hcancelV _): F=>[|<- _ _ {y hL1 C2} M].
- by move/cohH: C2; rewrite /validH priv_flatsE; move/validL.
by case/menvs_loc_eq1: M (M)=>m' -> M; apply: H; case/menvs_coh: M. 
Qed.

End ReadLemmas.



Section WriteLemmas.
Variables (A X : Type) (hp : nat).
Notation V := (Priv hp).

Lemma val_write' (x : ptr) (v : A) (w : X) hL s (r : cont unit) :
        (forall s', hp@sf-> (x :-> v \+ hL) \+ s' \In Mod.coh V ->
                    r tt (hp@sf-> (x :-> v \+ hL) \+ s')) ->
        verify (hp@sf-> (x :-> w \+ hL) \+ s) (act (Write hp x v)) r.
Proof.
move=>H; apply: val_act=>// i M; case/menvs_loc_eq1: M (M)=>s' ->{i} M. 
split=>[|[] j m {M}].
  (* this is a proof that precondition is stable under rely^* *)
  (* this should be a general property of actions *)
- case/menvs_coh: M=>_ C'; case/priv_stE2: C' (C')=>hE -> C'.
  rewrite joinA -stE1 in C' *.
  by exists hL, X, w, hE. 
case=>hL1 [D][z][hE][F] ->{j} C2; rewrite stE1 -joinA in F.
case/(sfcancel (cohV C2)): F (C2)=>F _ ->{s' C2} C2.
rewrite joinA -stE1 in C2.
have V : valid (x :-> w \+ hL).
- by move/cohH: C2; rewrite /validH priv_flatsE; move/validL.
move/(hcancelT V): (F)=>F'; rewrite -{D}F' in z F *.
case/(hcancelV V): F=>_ _ <-{z hL1} M; rewrite stE1 -joinA in M. 
by case/menvs_loc_eq1: M (M)=>m' -> M; apply: H; case/menvs_coh: M.
Qed.

End WriteLemmas.


(*********************************************************)
(* rules for pulling out the spec vars from binary posts *)
(*********************************************************)
      
Section Ghosts.
Variables (W : mod) (A : Type) (e : ST W A).

Lemma ghE B1 B2 (p : B1 -> B2 -> pre) (q : post A) :
        conseq e (fun i => exists x1 x2, p x1 x2 i) q <->
        conseq e (fun i => exists x, p x.1 x.2 i) q.
Proof.
split; move=>H1 i /=; case.
- by case=>x1 x2 /= H2; apply: H1; exists x1, x2. 
by move=>x1 [x2 H2]; apply: H1; exists (x1, x2). 
Qed.

Lemma ghA B1 B2 (p : pre) (q : B1 -> B2 -> post A) :
        conseq e p (fun i y m => forall x1 x2, q x1 x2 i y m) <->
        conseq e p (fun i y m => forall x, q x.1 x.2 i y m). 
Proof.
split; move=>H1 i; move/H1; apply: vrf_mono=>/= y m H.
- by case=>x1 x2; apply: H.
by move=>x1 x2; apply: (H (x1, x2)).
Qed.

Lemma ghAI B (p : pre) (q1 q2 : B -> pre) (r : B -> post A) :
        conseq e p (fun i y m => forall x, q1 x i -> q2 x i -> r x i y m) <->
        conseq e p (fun i y m => forall x, q1 x i /\ q2 x i -> r x i y m).
Proof.
split; move=>H1 i; move/H1; apply: vrf_mono=>/= y m H x.
- by case; apply: H. 
by move=>Q1 Q2; apply: H.
Qed.

Lemma ghC (B : Type) (p : B -> pre) (q : B -> post A) : 
        (forall i x, p x i -> verify i e (q x i)) ->
        conseq e (fun i => exists x, p x i) 
                 (fun i y m => forall x, p x i -> q x i y m). 
Proof.
move=>H i /= [x Hp] C t Ht.
have S : alwsafe i t by apply: alw_imp (H i x Hp C t Ht).  
by apply/aftA=>// y; apply/aftI=>//; move/H; apply. 
Qed.

End Ghosts.

Definition ghs := (ghE, ghA, ghAI).
