Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path rels ordtype finmap.
Require Import Eqdep prelude idynamic pcm codes unionmap heap feaps state.
Require Import zigzag worlds inject rely retract actions auto_split.
Require Import privstate footprint model lemmas.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.
(*
(* we introduce a bit stronger definition for precision *)
(* than usual; in that we demand that the ghosts are *)
(* determined by the heaps *)
(* strictly speaking, this is not necessary for much of the *)
(* development *)
(* however, it leads to a nice rule for hiding/retraction, whereby *)
(* we don't need the user to provide the ghosts by hand as *)
(* they can be uniquely determined from the heap *)
(* so, if this is not found good at some point, we can just *)
(* revert to use forall g, precise (inv_of g) *)
(* instead of aux_precise inv_of *)

Definition aux_precise (U : pcm) (P : U -> Pred heap) := 
  forall g1 g2 h1 h2 t1 t2, 
    valid (h1 \+ t1) ->
    h1 \+ t1 = h2 \+ t2 ->
    h1 \In P g1 -> h2 \In P g2 -> g1 = g2 /\ h1 = h2.
*)
Structure lockLike := Lock {
  label_of: nat;
  pcm_of : encoded_pcm; 
  inv_of : pcm_of -> Pred heap; 
  precise_of : forall g, precise (inv_of g)
}.

Section LockSpec.
Variables (hp : nat)(lkT : lockLike).
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation pf_prec := (@precise_of lkT).
Variables (f : U -> U) (pf : local f).

(* Locking procedure *)
Definition lock_spec is_lock locked : spec unit := 
  (fun i => exists (hl he : heap) (gl ge: U) i', 
     i = hp ->> [hl, tt, he] \+ i' /\ is_lock gl ge i', 
   fun i y m => forall hl he (gl ge: U) i', 
     i = hp ->> [hl, tt, he] \+ i' -> is_lock gl ge i' ->
     exists h ge' he' m',
       [/\ m = hp ->> [hl \+ h, tt, he'] \+ m', 
           locked gl ge' m',
           valid (gl \+ ge') & I (gl \+ ge') h]).

(* Unlocking procedure *)
Definition unlock_spec is_lock locked : spec unit := 
  (fun i => exists (gL gE: U)(hl he h: heap) i', 
     (fun i => 
       [/\ valid (f gL \+ gE), I (f gL \+ gE) h, 
           locked gL gE i' & i = hp ->> [hl \+ h, tt, he] \+ i']) i,
   fun i y m => forall gL gE hl he h i', 
     (fun i => 
       [/\ valid (f gL \+ gE), I (f gL \+ gE) h, 
           locked gL gE i' & i = hp ->> [hl \+ h, tt, he] \+ i']) i ->
     exists he' (gE': U) m',
       [/\ m = hp ->> [hl, tt, he'] \+ m' &
           is_lock (f gL) gE' m']).

End LockSpec.

Section LockStruct.
Variables (hp : nat).
Structure lock_struct (lock: lockLike) := Lockish {
  lock_mod : mod;
  is_lock : pcm_of lock -> pcm_of lock -> state -> Prop;
  _ : forall gl ge s, is_lock gl ge s -> s \In Mod.coh lock_mod;
  _ : forall gl ge s, is_lock gl ge s <-> is_lock ge gl (transp s);
  locked : pcm_of lock -> pcm_of lock -> state -> Prop;
  _ : forall s gl ge, locked gl ge s -> is_lock gl ge s; 
  _ : forall s gl ge, locked gl ge s -> locked ge gl (transp s)-> False;
  lock_proc : @STsep (entangle (Priv hp) lock_mod) unit (lock_spec hp is_lock locked);
  unlock_proc : forall (f : pcm_of lock -> pcm_of lock)(pf: local f), 
    @STsep (entangle (Priv hp) lock_mod) unit (unlock_spec hp f is_lock locked)
}.

End LockStruct.