Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path rels ordtype finmap.
Require Import Eqdep prelude idynamic pcm codes unionmap heap feaps state.
Require Import zigzag worlds auto_split inject rely actions rely.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.


(****************************)
(* Module for private state *)
(****************************)

Module PrivCoh.
Section PrivCoh.
Variable hp : nat. (* handle label for the module *)

Let validPriv s := exists hL hE : heap, s = hp ->> [hL, tt, hE].

Definition coh := [Pred s : state | validH s /\ validPriv s].

Lemma pcohSO s : s \In coh -> validSO s. 
Proof.
case=>Vh [hL][hE] E; move: E Vh ->.
rewrite /validH /validSO fsE !flatsE !pcm_flatPtE type_flatPt unitR. 
by rewrite zipPt allvalidPt. 
Qed.

Lemma pcohH s : s \In coh -> validH s. 
Proof. by case. Qed.

Lemma pcohfV s : s \In coh -> footV_eq s (mem_seq' [:: hp]).
Proof. by case=>_ [hL][hE] ->; rewrite fsE. Qed.

Lemma pcoh_weq s s' : s \In coh -> s' \In coh -> weq typeCode (shr s) (shr s').
Proof. by case=>_ [hL][hE] -> [_][hL'][hE'] ->; rewrite !fsE /=. Qed.

Lemma pcoh_sig s : s \In coh -> hp \-> (Unit : heap) = [zunit loc s].
Proof. by case=>_ [hL][hE] ->; rewrite fsE zunitPt. Qed.

Lemma pcoh_zigzag s p : (s \< p) \In coh <-> (s \> p) \In coh. 
Proof.
suff L : validPriv (s \< p) <-> validPriv (s \> p).
- split=>C; case: C (pcohSO C)=>Vh Vp Vso. 
  - by split; rewrite -?validH_zigzag // -?validSO_zigzag -?L.
  by split; rewrite ?validH_zigzag // ?validSO_zigzag ?L.
split; case=>hL [hE]; rewrite -[s]st_eta zigE zagE fsE.
- case; case/zipPtf=>h1 [h2][_ ->->->->].
  by exists h2, (hE \+ h1); rewrite zipPt -fsE. 
case=>->->; case/zipPtf=>h1 [h2][_ ->->].
by exists (h2 \+ hL), h1; rewrite zipPt -fsE. 
Qed.

End PrivCoh.

Module Exports.

Definition PrivCohMixin hp := 
  CohPredMixin (@pcohSO hp) (@pcohH hp) (@pcohfV hp) 
               (@pcoh_weq hp) (@pcoh_sig hp) (@pcoh_zigzag hp).
Definition priv_coh hp := Eval hnf in CohPred _ (PrivCohMixin hp).

Lemma priv_flatsE hp hL hE : flats (hp ->> [hL, tt, hE]) = hL \+ hE.
Proof. by rewrite fsE !flatsE /= !pcm_flatPtE type_flatPt /= unitR. Qed.

Lemma leq_cohUn coh1 coh2 s1 s2 :  
        leq_coh coh1 coh2 -> 
        s1 \+ s2 \In coh2 -> 
        (forall s, s \In coh1 -> footeqV s1 s) ->
        s1 \In coh1.
Proof.
case=>c' -> [Vh][s1'][s2'][E C1 C2].
by move/(_ _ C1)/(footeqV_prec (validHV Vh) E)=>->.
Qed.

Lemma priv_footV hp (hL hE : heap) s : 
        s \In priv_coh hp -> footeqV (hp ->> [hL, tt, hE]) s.
Proof. by case=>_ [hL'][hE'] ->. Qed.

Lemma get_priv coh s hp (hL hE : heap) :
        hp ->> [hL, tt, hE] \+ s \In star_coh (priv_coh hp) coh <->
        [/\ hp ->> [hL, tt, hE] \In priv_coh hp, s \In coh & 
            validH (hp ->> [hL, tt, hE] \+ s)].
Proof.
split; last by case=>C1 C2 Vh; split=>//; eexists _, _.
move=>C; move: (leq_cohUn star_id C (priv_footV _ _))=>H.
by split=>//; [apply/(star_coh_prec C) | move/cohH: C]. 
Qed.

Lemma priv_prec hp (hL hL' hE hE' : heap) : 
        hp ->> [hL, tt, hE] = hp ->> [hL', tt, hE'] ->
        hL = hL' /\ hE = hE'.
Proof.
rewrite !fsE; move/stE=>/= []. 
by move/(fcancelPt (C:=pcmCode))=><- _ /(fcancelPt (C:=pcmCode)) <-.
Qed.

End Exports.
End PrivCoh.

Export PrivCoh.Exports.


(* transitions *)

Module PrivGuar.
Section PrivGuar.
Variable hp : nat.
Notation coh := (priv_coh hp).

(*** internal transition ***)

Definition int_rel s s' := 
  (exists hL hL' hE : heap,
    [/\ s = hp ->> [hL, tt, hE], s' = hp ->> [hL', tt, hE], 
        valid hL' & dom hL =i dom hL']) /\ s \In coh.

Lemma intguar_coh s s' : int_rel s s' -> s \In coh /\ s' \In coh.
Proof.
case=>H C; split; first by apply: C. 
case: H C=>hL [hL'][hE][->-> V E][].
rewrite /validH priv_flatsE => Vh _.
have Vh' : valid (hL' \+ hE).
- suff: footeqH hL hL' by move/footeqH_validUnL=><-.
  by rewrite /footeqH V (validL Vh).
split; last by exists hL', hE.
by rewrite /validH !priv_flatsE.
Qed.

Lemma intguar_env s1 s2 : int_rel s1 s2 -> env s1 = env s2.
Proof. by case=>[][hL][hL'][hE][->->]; rewrite !fsE. Qed.

Lemma intguar_zagzig s s' p : 
        env s = env s' -> 
        int_rel (s \> p) (s' \> p) -> int_rel (s \< p) (s' \< p).
Proof.
move=>H; case; case=>hL [hL'][hE][H1 H2 V E] C.
split; last by apply/coh_zigzag.
move: H1 H2 H C; rewrite -[s]st_eta -[s']st_eta !zigE !zagE !fsE.
case=>->->; case/zipPtf=>h1 [h2][->->->].
case=>->->; case/zipPtf=>h3 [h4][_ ->] _; rewrite !zipPt. 
move/(fcancelPt (C:=pcmCode))=><-{h3} /cohH.
rewrite /validH -fsE priv_flatsE joinCA; move/validR=>VL.
have VL' : valid (hL' \+ h2).
- suff: footeqH hL hL' by move/footeqH_validUnL=><-.
  by rewrite /footeqH V (validL VL).
exists (hL \+ h2), (hL' \+ h2), h1; rewrite !(joinC h2) -!fsE; split=>//.
by move=>x; rewrite !domUn !inE /= E VL VL'. 
Qed.

Lemma intguar_footH s1 s2 : 
        int_rel s1 s2 -> footeqH (flats s1) (flats s2).
Proof.
move=>G; case/intguar_coh: (G); move/cohH=>V; move/cohH=>V'.
case: G V V'; case=>hL [hL'][hE][->-> _ E] _.
rewrite /footeqH /validH !priv_flatsE /= => V V'.
by rewrite V V'; split=>// x; rewrite !(@domUn heapUMC) !inE /= E V V'.
Qed.

Lemma intguar_refl s : s \In coh -> int_rel s s. 
Proof.
move=>C; split=>//; case: C=>Vh [hL][hE] E; rewrite {s}E in Vh *.
exists hL, hL, hE; split=>//; rewrite /validH priv_flatsE in Vh. 
by move/validL: Vh.
Qed.

(*** acquire transition ***)

Section Acquire.
Variable h : heap.

Definition acq_rel (s s' : state) := 
  (exists hL hE : heap, 
    [/\ s = hp ->> [hL, tt, hE], s' = hp ->> [hL \+ h, tt, hE] &
        valid (hL \+ h \+ hE)]) /\ s \In coh.

Lemma acqguar_coh s s' : acq_rel s s' -> s \In coh /\ s' \In coh.
Proof.
case=>H1 H2; split; first by apply: H2. 
case: H1 H2=>hL [hE][->-> V] [Vh Vp]. 
split.
- by rewrite /validH !priv_flatsE.
by exists (hL \+ h), hE.
Qed.

Lemma acqguar_env s s' : acq_rel s s' -> env s = env s'.
Proof. by case; case=>hL [hE][->-> _]; rewrite !fsE. Qed.

Lemma acqguar_zagzig s s' p : 
        env s = env s' -> 
        acq_rel (s \> p) (s' \> p) -> acq_rel (s \< p) (s' \< p).
Proof.
move=>H; case; case=>hL [hE][H1 H2 V] C.
split; last by apply/coh_zigzag.
move: H1 H2 H V; rewrite -[s]st_eta -[s']st_eta !zigE !zagE !fsE.
case=>->->; case/zipPtf=>h1 [h2][->->->].
case=>->->; case/zipPtf=>h3 [h4][_ ->] _; rewrite !zipPt.
move/(fcancelPt (C:=pcmCode))=><-{h3} V.
exists (h2 \+ hL), h1; rewrite -!joinA -!fsE; split=>//.
by rewrite (joinC h2) !joinA in V *.
Qed.

Lemma acqguar_footH s s' : acq_rel s s' -> footeqH (flats s \+ h) (flats s').
Proof. by case; case=>hL [hE][->-> V] _; rewrite /= !priv_flatsE joinAC. Qed.

End Acquire.

(*** release transition ***)

Section Release.
Variable h : heap.

Definition rel_rel s s' := 
  (exists hL hE : heap, s = hp ->> [hL \+ h, tt, hE] /\ 
     s' = hp ->> [hL, tt, hE]) /\ s \In coh.

Lemma relguar_coh s s' : rel_rel s s' -> s \In coh /\ s' \In coh.
Proof.
case=>H1 H2; split; first by apply: H2.
case: H1 H2=>hL [hE][->->][Vh Vp].
split; last by exists hL, hE. 
by move: Vh; rewrite /validH !priv_flatsE joinAC; move/validL.
Qed.

Lemma relguar_env s s' : rel_rel s s' -> env s = env s'.
Proof. by case; case=>hL [hE][->->]; rewrite !fsE. Qed.

Lemma relguar_zagzig s s' p : 
        env s = env s' -> 
        rel_rel (s \> p) (s' \> p) -> rel_rel (s \< p) (s' \< p).
Proof.
move=>H; case; case=>hL [hE][H1 H2] C.
split; last by apply/coh_zigzag.
move: H1 H2 H C; rewrite -[s]st_eta -[s']st_eta !zigE !zagE !fsE.
case=>->->; case/zipPtf=>h1 [h2][->->->].
case=>->->; case/zipPtf=>h3 [h4][_ ->] _; rewrite !zipPt.
move/(fcancelPt (C:=pcmCode))=><-{h3} _.
by exists (h2 \+ hL), h1; rewrite joinA -!fsE. 
Qed.

Lemma relguar_footH s s' : rel_rel s s' -> footeqH (flats s) (h \+ flats s').
Proof.
case; case=>hL [hE][->->]; move/cohH.
by rewrite /validH /= !priv_flatsE joinCA joinA.
Qed.

End Release.
End PrivGuar.

Module Exports.

Definition PrivGuarMixin hp := 
  GuarMixin (@intguar_coh hp) (@intguar_env hp) (@intguar_zagzig hp).
Canonical priv_int_guar hp := 
  Eval hnf in Guar (int_rel hp) (PrivGuarMixin hp).

Definition PrivIntGuarMixin hp := IntGuarMixin (@intguar_footH hp).
Canonical PrivIntGuar hp := 
  Eval hnf in IntGuar (priv_int_guar hp) (PrivIntGuarMixin hp).

Definition PrivReflGuarMixin hp := ReflGuarMixin (@intguar_refl hp).
Canonical PrivReflGuar hp := 
  Eval hnf in ReflGuar (priv_int_guar hp) (PrivReflGuarMixin hp).

Definition basePrivAcqGuarMixin hp h := 
  GuarMixin (@acqguar_coh hp h) (@acqguar_env hp h) (@acqguar_zagzig hp h).
Canonical priv_acq_guar hp h := 
  Eval hnf in Guar (acq_rel hp h) (basePrivAcqGuarMixin hp h).

Definition PrivAcqGuarMixin hp := AcqGuarMixin (@acqguar_footH hp).
Canonical PrivAcqGuar hp := 
  Eval hnf in AcqGuar (priv_acq_guar hp) (PrivAcqGuarMixin hp).

Definition basePrivRelGuarMixin hp h := 
  GuarMixin (@relguar_coh hp h) (@relguar_env hp h) (@relguar_zagzig hp h).
Canonical priv_rel_guar hp h := 
  Eval hnf in Guar (rel_rel hp h) (basePrivRelGuarMixin hp h).

Definition PrivRelGuarMixin hp := RelGuarMixin (@relguar_footH hp).
Canonical PrivRelGuar hp := 
  Eval hnf in RelGuar (priv_rel_guar hp) (PrivRelGuarMixin hp).

Definition PrivChanCore hp := ChanCore (PrivAcqGuar hp) (PrivRelGuar hp).
Canonical PrivCore hp : mod_core := @Mod.Core (priv_coh hp) (PrivReflGuar hp). 
Canonical Priv hp : mod :=  @Mod.Make (PrivCore hp) [:: PrivChanCore hp].

End Exports.
End PrivGuar.

Export PrivGuar.Exports.


(***********************)
(* actions for private *)
(***********************)

Module PrivReadAct.
Section PrivReadAct.
Variables (A : Type) (hp : nat) (x : ptr).
Notation coh := (Mod.coh (Priv hp)).
Notation guar := (Mod.internal (Priv hp)). 

Definition read_safe s := 
  (exists (hL hE : heap) (v : A),
     s = hp ->> [x :-> v \+ hL, tt, hE]) /\ s \In coh.

Definition read_step s s' (v : A) := 
  (exists hL hE : heap,
     [/\ s = hp ->> [x :-> v \+ hL, tt, hE] & s' = s ]) /\ s \In coh.
  
Lemma read_safe_coh : safe_coh coh read_safe.
Proof. by move=>s; case. Qed.

Lemma read_safe_zagzig : safe_zagzig read_safe.
Proof.
move=>s p [[hL][hE][v]]; rewrite -[s]st_eta fsE. 
case=>->-> /zipPtf [v1][v2][_ ->->] /coh_zigzag C.
split=>//; exists (hL \+ v2), v1, v.
by rewrite zigE zipPt joinC joinA fsE. 
Qed.

Lemma read_step_safe : step_safe read_safe read_step.
Proof. by move=>_ _ v [[hL][hE][-> _] C]; split=>//; exists hL, hE, v. Qed.

Lemma read_step_guar : step_guar guar read_step.
Proof.
move=>_ _ v [[hL][hE][->->] C]; split=>//; eexists _, _, _; split=>//.
by move/cohH: C; rewrite /validH flatsE fsE pcm_flatPtE; move/validL/validL. 
Qed.

Lemma read_step_frame : step_frame read_safe read_step.
Proof.
move=>s s' p v []; rewrite -[s]st_eta -[s']st_eta /=.
case=>hL [hE][w]; rewrite fsE. 
case=>->-> /zipPtf [v1][v2][_ ->->] C [].
case=>hL' [hE'][E] -> C'; eexists _; do !split=>//.
eexists hL, (v1 \+ v2); split=>//; rewrite zagE zipPt.
rewrite zigE zipPt joinC -joinA -!fsE in E *.
suff D: valid (x :-> w \+ (hL \+ v2)).
- by case/scancelPt: E; case/(hcancelV D)=>->.
move/cohH: C'; rewrite zigE zipPt joinC joinA /validH flatsE pcm_flatPtE. 
by move/validL/validL.  
Qed.

Lemma read_step_total : step_total read_safe read_step.
Proof.
move=>s [[hL]][hE][v ->{s}] C.
exists (hp ->> [x :-> v \+ hL, tt, hE]), v.
by split=>//; exists hL, hE.
Qed.

Definition Read_pre := 
  Action read_safe_coh read_safe_zagzig read_step_safe read_step_guar 
         read_step_frame read_step_total.

End PrivReadAct.

(* erasure lemmas *)

Lemma read_step_erase A hp x : int_step_erase (@Read_pre A hp x).
Proof.
move=>s1 h1 t1 v1 s2 h2 t2 v2; rewrite -[s1]st_eta -[s2]st_eta /=.
move=>V E H1 H2; case: H1 H2 V E=>[[hL][hE][->->] C][[hL'][hE'][->->] C'].
rewrite /validH !flatsE !fsE !pcm_flatPtE /= -!joinA => V E.
by split=>//; case/(hcancelV V): E. 
Qed.

Module Exports.

Definition Read A hp x := IntAction (@read_step_erase A hp x). 

End Exports.

End PrivReadAct.

Export PrivReadAct.Exports.



(* write should be similar, so let's do it *)

Module PrivWriteAct.
Section PrivWriteAct.
Variables (A : Type) (hp : nat) (x : ptr) (v : A).
Notation W := (Priv hp). 
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W). 

Definition write_safe s := 
  exists j B (w : B) (hE : heap),
    s = hp ->> [x :-> w \+ j, tt, hE] /\ s \In coh.

Definition write_step s s' (r : unit) := 
  exists j B (w : B) (hE : heap),
    [/\ s = hp ->> [x :-> w \+ j, tt, hE], 
        s' = hp ->> [x :-> v \+ j, tt, hE] & s \In coh].

Lemma write_safe_coh : safe_coh coh write_safe.
Proof. by move=>s [j][B][w][hE][]. Qed.

Lemma write_safe_zagzig : safe_zagzig write_safe.
Proof.
move=>s p [j][B][w][hE][]; rewrite -[s]st_eta fsE. 
case/stE=>/= ->-> /zipPtf [h1][h2][E ->->] /coh_zigzag C.
by exists (h2 \+ j), B, w, h1; split=>//; rewrite zigE zipPt joinCA fsE. 
Qed.

Lemma write_step_safe : step_safe write_safe write_step.
Proof. 
case=>lp sh ep s2 [][j][B][w][hE][]; rewrite fsE.
by case=>->->-> _ C; exists j, B, w, hE; rewrite -{1}fsE.
Qed.

Lemma write_step_guar : step_guar guar write_step.
Proof.
move=>s1 s2 [][j][B][w][hE][]; rewrite -[s1]st_eta fsE.
case/stE=>/= ->->->-> C; split=>//.
exists (x :-> w \+ j), (x :-> v \+ j), hE.
rewrite -fsE; split=>//; last by move=>z; rewrite !hdomPtUn !hvalidPtUn. 
move/cohH: C; rewrite /validH flatsE /= pcm_flatPtE -joinA.
by move/validL; rewrite !hvalidPtUn.
Qed.

Lemma write_step_frame : step_frame write_safe write_step.
Proof.
move=>s1 s2 p []; rewrite -[s1]st_eta -[s2]st_eta /=.
case=>j [B][w][hE][] /stE; rewrite fsE /=.
case=>->-> /zipPtf [h1][h2][_ ->->] C.
case=>j' [B'][w'][hE'][] /stE /= []; rewrite fsE /=.
have D : valid (x :-> w \+ (j \+ h2)).
- move/coh_zigzag/cohH: C; rewrite /validH flatsE; move/validL/validL.
  by rewrite /= zipPt pcm_flatPtE joinC -joinA.
rewrite zigE zipPt joinC -joinA; move/(@fpts_inj pcmCode)=>E.
move/(hcancelT D): (E)=>F; rewrite -{B'}F in w' E *.
case/(hcancelV D): E=>_ _ <-{j'} _.
move/(@fpts_inj pcmCode)=><-{hE'} /stE /= [->->->] _.
exists (hp ->> [x :-> v \+ j, tt, h1]); rewrite !fsE.
split; first by rewrite zigE zipPt joinA joinC. 
exists j, B, w, (h1 \+ h2); rewrite !fsE.
by split=>//; rewrite !zagE zipPt. 
Qed. 

Lemma write_step_total : step_total write_safe write_step.
Proof.
move=>s [hL][B][w][hE][->{s} C].
exists (hp ->> [x :-> v \+ hL, tt, hE]), tt.
by exists hL, B, w, hE. 
Qed.

Definition Write_pre := 
  Action write_safe_coh write_safe_zagzig write_step_safe write_step_guar
         write_step_frame write_step_total.

End PrivWriteAct.

Lemma write_step_erase A hp x v : int_step_erase (@Write_pre A hp x v). 
Proof.
move=>s1 h1 t1 [] s2 h2 t2 []; rewrite -[s1]st_eta -[s2]st_eta /=.
move=>V E H1 H2; move: H1 H2 V E. 
case=>j1 [B1][w1][hE1][] /stE /=; rewrite fsE /=.
case=>->->-> -> C1. 
case=>j2 [B2][w2][hE2][] /stE /=; rewrite fsE /=.
case=>->->-> -> C2.
rewrite /validH -!fsE !priv_flatsE -!joinA => V E.
move/(hcancelT V): (E) (V) => F; rewrite -{B2}F in w2 C2 E *.
by case/(hcancelV V): E=>_ _ ->; rewrite !hvalidPtUn.  
Qed.

Module Exports.

Definition Write A hp x v := IntAction (@write_step_erase A hp x v).

End Exports. 

End PrivWriteAct.

Export PrivWriteAct.Exports.


(* PRIV transfer actions *)

Module PrivAcquireAct.
Section PrivAcquireAct.
Variables (hp : nat) (h : heap).
Notation W := (Priv hp). 
Notation coh := (priv_coh hp).
Notation guar := (priv_acq_guar hp h).

Definition acq_safe s := 
  exists hL hE, 
    [/\ s = hp ->> [hL, tt, hE], valid (hL \+ h \+ hE) & s \In coh].

Definition acq_step (s s' : state) (r : unit) := 
  exists hL hE : heap, 
    [/\ s = hp ->> [hL, tt, hE], s' = hp ->> [hL \+ h, tt, hE] &
        s' \In coh].

Lemma acq_safe_coh : safe_coh coh acq_safe.
Proof. by move=>s [hL][hE][]. Qed.

Lemma acq_safe_zagzig : safe_zagzig acq_safe.
Proof.
move=>s p; rewrite -[s]st_eta; case=>hL [hE][/stE /=]. 
rewrite fsE /=; case=>->-> /zipPtf [v1][v2][->->->] V /coh_zigzag C.
exists (v2 \+ hL), v1; split=>//.
- by rewrite fsE; apply/stE; rewrite /= zipPt. 
by rewrite -!joinA joinC joinA -joinA.
Qed.

Lemma acq_step_safe : step_safe acq_safe acq_step.
Proof. 
move=>s s' [[hL]][hE][->-> C]; exists hL, hE.
rewrite -(joinC h) in C *; split=>//.
- by move/cohH: C; rewrite /validH priv_flatsE.
split; last by exists hL, hE.
by move/cohH: C; rewrite /validH !priv_flatsE -!joinA; move/validR. 
Qed.

Lemma acq_step_guar : step_guar guar acq_step. 
Proof.
move=>s s' [][hL][hE][->-> C]. 
have Vh : valid (hL \+ hE \+ h).
- by move/cohH: C; rewrite /validH priv_flatsE joinAC.
split; first by exists hL, hE; rewrite joinAC.
split; last by exists hL, hE. 
by rewrite /validH !priv_flatsE; move/validL: Vh.
Qed.

Lemma acq_step_frame : step_frame acq_safe acq_step.
Proof.
move=>s s' p []; rewrite -[s]st_eta -[s']st_eta /=.
case=>hL [hE][/stE /=]; rewrite fsE. 
case=>->-> /zipPtf [v1][v2][->->->] _ _. 
rewrite zigE zagE !zipPt /= -fsE.
case=>L1 [E1][/scancelPt [<-<-{L1 E1}]] /stE /=.
rewrite fsE /=; case=>->->-> C. 
exists (hp ->> [hL \+ h, tt, v1]).
split=>//; first by rewrite fsE zigE zipPt joinA.
exists hL, (v1 \+ v2); rewrite !fsE.
split=>//; first by rewrite zagE zipPt. 
by rewrite -coh_zigzag zigE zipPt joinA. 
Qed.

Lemma acq_step_total : step_total acq_safe acq_step.
Proof.
move=>s [hL][hE][->{s} V C].
exists (hp ->> [hL \+ h, tt, hE]), tt.  
exists hL, hE; split=>//.
split; last by exists (hL \+ h), hE. 
by rewrite /validH priv_flatsE. 
Qed.

Definition PrivAcq_pre :=
  Action acq_safe_coh acq_safe_zagzig acq_step_safe acq_step_guar 
         acq_step_frame acq_step_total. 

End PrivAcquireAct.

Lemma acq_st_erase hp : acq_step_erase (PrivAcq_pre hp).
Proof.
move=>s h t [] k s' h' t' [] k' _ E St St'. 
case: St St' E=>hL1 [hE1][->-> C1][hL2][hE2][->-> C2].
by rewrite !priv_flatsE -!joinA -(joinCA k) -(joinCA k').
Qed.

Lemma acq_st_owns hp : acq_step_owns (PrivAcq_pre hp). 
Proof.
move=>s s' [] k [hL][hE][_ -> _]; exists (hL \+ hE).
by rewrite priv_flatsE -joinA joinCA.
Qed.

Module Exports.

Definition PrivAcq hp := AcqAction (@acq_st_erase hp) (@acq_st_owns hp).

End Exports.

End PrivAcquireAct.

Export PrivAcquireAct.Exports. 


(* release action *)

Module PrivReleaseAct.
Section PrivReleaseAct.
Variable hp : nat.

Section Internal.
Variable h : heap.
Notation W := (Priv hp). 
Notation coh := (priv_coh hp).
Notation guar := (priv_rel_guar hp h). 

Definition rel_safe s := 
  exists hL hE: heap, 
    [/\ s = hp ->> [hL \+ h, tt, hE] & s \In coh].

Definition rel_step (s s' : state) (r : unit) := 
  exists hL hE : heap, 
    [/\ s = hp ->> [hL \+ h, tt, hE], 
        s' = hp ->> [hL, tt, hE] & s \In coh].

Lemma rel_safe_coh : safe_coh coh rel_safe.
Proof. by move=>s [hL][hE][]. Qed.

Lemma rel_safe_zagzig : safe_zagzig rel_safe.
Proof.
move=>s p; rewrite -[s]st_eta; case=>hL [hE][/stE /=]. 
rewrite fsE /=; case=>->-> /zipPtf [v1][v2][_ ->->] /coh_zigzag C.
exists (v2 \+ hL), v1; split=>//.
by rewrite fsE zigE zipPt joinA.  
Qed.

Lemma rel_step_safe : step_safe rel_safe rel_step.
Proof. by move=>s s' [][hL][hE][-> _ C]; exists hL, hE. Qed.

Lemma rel_step_guar : step_guar guar rel_step. 
Proof. by move=>s s' [][hL][hE][->-> C]; split=>//; exists hL, hE. Qed.

Lemma rel_step_frame : step_frame rel_safe rel_step.
Proof.
move=>s s' p []; rewrite -[s]st_eta -[s']st_eta /=.
case=>hL [hE][/stE /=]; rewrite fsE /=.
case=>->-> /zipPtf [v1][v2][_ ->->] C.
have D : valid (v2 \+ hL \+ h).
- move/cohH: C; rewrite /validH zagE zipPt -fsE priv_flatsE.
  by rewrite joinC -!joinA; move/validR.
rewrite zigE zagE zipPt joinA -fsE.
case=>_ [_][/scancelPt [/(joinKf D) <-<-]]. 
rewrite fsE; case/stE=>/= ->->->; exists (hp ->> [hL, tt, v1]).
rewrite fsE zigE zagE !zipPt; split=>//.
exists hL, (v1 \+ v2); rewrite !fsE; split=>//.
by rewrite zagE zipPt in C.
Qed.

Lemma rel_step_total : step_total rel_safe rel_step.
Proof.
move=>s [hL][hE][->{s} C].
exists (hp ->> [hL, tt, hE]), tt.
by exists hL, hE. 
Qed.

Definition PrivRel_pre :=
  Action rel_safe_coh rel_safe_zagzig rel_step_safe rel_step_guar 
         rel_step_frame rel_step_total.

End Internal.

Lemma rel_st_erase : rel_step_erase PrivRel_pre.
Proof.
move=>s1 h1 t1 [] k1 s2 h2 t2 [] k2 _ E St1 St2.
case: St1 St2 E=>hL1 [hE1][->-> C1][hL2][hE2][->-> C2]. 
by rewrite !priv_flatsE -!joinA (joinCA k1) (joinCA k2). 
Qed.

Lemma rel_sf_owns : rel_safe_owns PrivRel_pre.
Proof.
move=>s k [hL][hE][-> C]; exists (hL \+ hE).
by rewrite priv_flatsE -joinA joinCA.
Qed.

End PrivReleaseAct.

Module Exports. 

Definition PrivRel hp := 
  RelAction (@rel_st_erase hp) (@rel_sf_owns hp).

End Exports.
End PrivReleaseAct.

Export PrivReleaseAct.Exports.

(*************)
(* Stability *)
(*************)

Section PrivStableBase.
Variable hp : nat.
Notation V := (Priv hp).

Lemma mtx_priv_mrely (hL hE : heap) s : 
        mrely V (hp ->> [hL, tt, hE]) s ->
        exists hE', s = hp ->> [hL, tt, hE'].
Proof.
rewrite -[s]st_eta; case=>M.
- case: M; case=>hL' [hL1'][hE'][]; rewrite /transp !fsE /= -2!fsE.
  by case/priv_prec=>->->[->->->]; exists hL1'; rewrite fsE. 
case: M=>_ [h'][/Mem_map_inv][ct][->]; rewrite InE /= => ->{ct}.
case; case; case=>hL' [hE'][]; rewrite /transp !fsE /= -2!fsE;
case/priv_prec=>->->[->->->].
- by exists (hL' \+ h'); rewrite fsE.
by exists hL'; rewrite fsE. 
Qed.

Lemma mtx_priv (hL hE : heap) s : 
        menv_steps V (hp ->> [hL, tt, hE]) s ->
        exists hE', s = hp ->> [hL, tt, hE'].
Proof.
case=>n; elim: n hE=>[|n IH] /= hE; first by case=><-; exists hE. 
by case=>_ [] /mtx_priv_mrely [hE'] ->; apply: IH.
Qed.

End PrivStableBase.











