Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude heaps idynamic pcm codes feaps state.
Require Import zigzag worlds inject.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Section Rely.
Variable W : mod. 
Notation coherent := (Mod.coh W).
Notation guar := (Mod.internal W).

(* rely is a transpose of guar. *)

Definition rely s s' := guar (transp s) (transp s'). 

Lemma guar_rely lp1 lp2 lp3 sh sh' ep : 
        guar ([lp1, sh, ep] \> lp2) ([lp3, sh', ep] \> lp2) ->
        rely ([lp2, sh, ep] \> lp1) ([lp2, sh', ep] \> lp3).
Proof.
rewrite !zagE !(zipC ep) -!zagE; move/guar_zagzig; move/(_ (erefl _)).
by rewrite !zigE !(zipC ep).
Qed.

Lemma rely_coh s s' : rely s s' -> s \In coherent /\ s' \In coherent.
Proof. 
rewrite -[s]st_eta -[s']st_eta.
by case/guar_coh=>C C'; split; rewrite cohC.
Qed.

Lemma rely_loc s s' : rely s s' -> loc s = loc s'.
Proof. by move/guar_env. Qed.

Lemma rely_zigeq' s s' p :
        rely (s \< p) s' -> s' = [loc s, shr s', env s'] \< p.
Proof. by move/rely_loc; rewrite zigE !spP zipC -[s']st_eta /= =>->. Qed.

Lemma rely_zigeq s1 s2 p :
        rely (s1 \< p) s2 -> exists s, s2 = s \< p /\ loc s1 = loc s. 
Proof. 
move/rely_loc=>E; exists [loc s1, shr s2, env s2]. 
by rewrite -[s2]st_eta /= -E zigE spP. 
Qed.

Lemma rely_zigzag s s' p : 
        loc s = loc s' -> 
        rely (s \< p) (s' \< p) -> rely (s \> p) (s' \> p).
Proof.
rewrite -[s]st_eta -[s']st_eta /= => E.
rewrite /rely /transp /= -!(zipC (loc _)) !(zipC (env _)) -!zagE -!zigE.
by apply: guar_zagzig.  
Qed.

Lemma rely_footV s1 s2 : rely s1 s2 -> footeqV s1 s2. 
Proof. by move/guar_footV; rewrite -transp_footV. Qed.

Lemma rely_footH s1 s2 : rely s1 s2 -> footeqH (flats s1) (flats s2).
Proof. by move/int_guar_footH; rewrite -!transp_flatsE. Qed.

End Rely.



(* environment stepping which involves acqs and rels *)
(* of a module, rather than just the internal steps *)

Section MRely.
Variable W : mod. 
Notation coh := (Mod.coh W).

Definition mrely s s' := mguar W (transp s) (transp s').

Lemma mrely_footV s1 s2 : mrely s1 s2 -> footeqV s1 s2. 
Proof. by move/mguar_footV; rewrite -transp_footV. Qed.

Lemma mrely_refl s : s \In coh -> mrely s s.
Proof. by move/transp_coh; apply: mguar_refl. Qed.

Lemma mrely_coh s s' : mrely s s' -> s \In coh /\ s' \In coh.
Proof. by move/mguar_coh; rewrite !transp_coh. Qed.

Lemma mrely_loc s s' : mrely s s' -> loc s = loc s'.
Proof. by apply: mguar_env. Qed.

Lemma mrely_zigeq s1 s2 p :
        mrely (s1 \< p) s2 -> exists s, s2 = s \< p /\ loc s1 = loc s. 
Proof. 
rewrite /mrely transp_zig; case/mguar_zageq=>s.
by rewrite -transp_inj transpI transp_zag; exists (transp s). 
Qed.

Lemma mrely_zigzag s s' p : 
        loc s = loc s' -> 
        mrely (s \< p) (s' \< p) -> mrely (s \> p) (s' \> p).
Proof. by rewrite /mrely !transp_zig !transp_zag=>E; apply: mguar_zagzig. Qed.

(* transitive closure *)      

Fixpoint menv_steps' n (s s' : state) :=
  if n is n'.+1 then exists s'', mrely s s'' /\ menv_steps' n' s'' s'
  else s = s' /\ s' \In coh. 

Definition menv_steps s s' := exists n, menv_steps' n s s'.

Lemma menvs_refl s : s \In coh -> menv_steps s s.
Proof. by exists 0. Qed.

Lemma menvs_trans s1 s2 s3 : 
        menv_steps s1 s2 -> menv_steps s2 s3 -> menv_steps s1 s3.
Proof.
case=>n; elim: n s1 s2=>[|n IH] s1 s2 /=; first by case=>->. 
by case=>s [H1 H2]; case/(IH _ _ H2)=>m; exists m.+1, s. 
Qed.

Lemma menvs_coh s1 s2 : menv_steps s1 s2 -> s1 \In coh /\ s2 \In coh.
Proof.
case=>n; elim: n s1=>[|n IH] s1 /=; first by case=>->.
by case=>s []; case/mrely_coh=>C1 _; case/IH. 
Qed.

Lemma menvs_loc s1 s2 : menv_steps s1 s2 -> loc s1 = loc s2.
Proof.
case=>n; elim: n s1=>[|n IH] s1 /=; first by case=>->. 
by case=>s []; case=>[|[ch][h][H][]] /guar_env /= -> /IH. 
Qed.

Lemma menvs_zigeq s1 s2 p : 
        menv_steps (s1 \< p) s2 -> exists s, s2 = s \< p /\ loc s1 = loc s. 
Proof. 
move/menvs_loc => E; exists [loc s1, shr s2, env s2]. 
by rewrite -[s2]st_eta /= -E zigE spP. 
Qed.

Lemma menvs_zigzag s s' p : 
        loc s = loc s' -> 
        menv_steps (s \< p) (s' \< p) -> menv_steps (s \> p) (s' \> p).
Proof.
move=>E; case=>n; elim: n s s' E=>[|n IH] s s' E /= []. 
- by move/(zigzag_eq E)=>-> /coh_zigzag; apply: menvs_refl.
move=>s1 [H]; case/mrely_zigeq: H (H)=>s2 [->{s1}] E' H. 
move/IH; rewrite -E'; move/(_ E); apply: menvs_trans.
eexists 1, (s2 \> p); split; first by apply/mrely_zigzag. 
by split=>//; apply/coh_zigzag; case/mrely_coh: H. 
Qed.

End MRely.

(******************************************)
(* Dualization of extend and split lemmas *)
(******************************************)

(* rely *)

Lemma rely_extend U V (w : injects U V) s1 s2 s : 
        s1 \+ s \In Mod.coh V ->
        s2 \+ s \In Mod.coh V ->
        rely U s1 s2 -> 
        rely V (s1 \+ s) (s2 \+ s).
Proof.
move/transp_coh=>C1 /transp_coh C2; rewrite /rely !transpUn.
by apply: (internal_extend w).
Qed.

(* mrely *)

Lemma mrely_split U V (w : injects U V) s1 s1' s2 s2' : 
        s1 \In Mod.coh U -> s2 \In Mod.coh U ->
        mrely V (s1 \+ s1') (s2 \+ s2') ->
        mrely U s1 s2 /\ mrely (inj_ext w) s1' s2'.
Proof.
move/transp_coh=>C1 /transp_coh C2; rewrite /mrely !transpUn.
by apply: (mguar_split w) C1 C2.
Qed.

Lemma menvs_split U V (w : injects U V) s1 s1' s2 s2' :
        s1 \In Mod.coh U -> s2 \In Mod.coh U ->
        menv_steps V (s1 \+ s1') (s2 \+ s2') ->
        menv_steps U s1 s2 /\ menv_steps (inj_ext w) s1' s2'.
Proof.
move=>C1 C2 [n E]. 
elim: n s1 s1' E C1 C2=>[|n IH] /= s1 s1' E C1 C2. 
- case: E=>E C; rewrite -E (cohE w) in C.
  move: (coh_prec (cohV C) E C1 C2)=>?; subst s2.
  move/(joinsK (cohV C)): E=>?; subst s2'.
  split; exists 0=>//; split=>//.
  by apply/(star_coh_prec C). 
case: E=>s [H1 H2]; case: (mrely_coh H1)=>_ Cs.
rewrite (cohE w) in Cs.
have [s3 [s3' [E C3]]]: exists s3 s3', s = s3 \+ s3' /\ s3 \In Mod.coh U.
- by case: Cs=>_ [s3][s3'][F]; exists s3, s3'.  
rewrite {s}E in H1 H2 Cs.
case/(mrely_split w C1 C3): H1=>M1 M2.
case: (IH _ _ H2 C3 C2)=>N1 N2.
split; first by apply: menvs_trans N1; exists 1, s3.
apply: menvs_trans N2; exists 1, s3'; do !split=>//.
by move/(star_coh_prec Cs): C3.
Qed.

(* an extra lemma, holds only for entanglement, not general injections *)
(* needed when proving the meta theory of retraction *)

Lemma mrely_entangle_extend U V s1 s2 s : 
        s1 \+ s \In Mod.coh (entangle U V) ->
        s2 \+ s \In Mod.coh (entangle U V) ->
        mrely U s1 s2 -> 
        mrely (entangle U V) (s1 \+ s) (s2 \+ s).
Proof.
move/transp_coh=>C1 /transp_coh C2; rewrite !transpUn in C1 C2; case=>H.
- by move/(entangle_extend C1 C2): H; left. 
case: H=>_ [h][/Mem_map_inv][ct][->] Ch; case. 
- case/(lefts_extend (b:=acq) C1 C2 Ch)=>ct' [Ch'] H. 
  right; exists (Chan ct'), h; split=>//; first by apply/Mem_map. 
  by left.
case/(lefts_extend (b:=rel) C1 C2 Ch)=>ct' [Ch'] H. 
right; exists (Chan ct'), h; split=>//; first by apply/Mem_map. 
by right.
Qed.
