Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path rels perms.
Require Import ordtype Eqdep prelude idynamic pcm unionmap heap codes feaps state. 
Require Import zigzag. 
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(************************)
(************************)
(* Coherence predicates *)
(************************)
(************************)

Definition footV_eq (s : state) (fv : pred nat) := 
  [/\ fdom (loc s) =i fv, fdom (shr s) =i fv & fdom (env s) =i fv].

Module CohPred.

Structure mixin_of (coh : Pred state) := Mixin {
  fV_op : pred nat;
  sig_op : feap dpcm;
  _ : forall s, coh s -> validSO s;
  _ : forall s, coh s -> validH s;
  _ : forall s, coh s -> footV_eq s fV_op;
  _ : forall s s', s \In coh -> s' \In coh -> 
        weq typeCode (shr s) (shr s');
  _ : forall s, s \In coh -> sig_op = [zunit loc s];
  _ : forall s p, coh (s \< p) <-> coh (s \> p)}.

Section ClassDef.

Notation class_of := mixin_of (only parsing).

Structure cohpred : Type := Pack {sort : state -> Prop; _ : class_of sort}.
Local Coercion sort : cohpred >-> Funclass.

Variables (T : state -> Prop) (cT : cohpred).

Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.

Definition pack c := @Pack T c.
Definition clone := fun c & T = cT & phant_id (pack c) cT => pack c. 

Definition fV := fV_op class.
Definition sig := sig_op class.

End ClassDef.

Module Exports.
Coercion sort : cohpred >-> Funclass.
Notation cohpred := cohpred.
Notation CohPredMixin := Mixin.
Notation CohPred T m := (@pack T m).

Notation "[ 'cohPredMixin' 'of' T ]" := (class _ : mixin_of T)
  (at level 0, format "[ 'cohPredMixin'  'of'  T ]") : form_scope.
Notation "[ 'cohpred' 'of' T 'for' C ]" := (@clone T C _ (erefl _) id)
  (at level 0, format "[ 'cohpred'  'of'  T  'for'  C ]") : form_scope.
Notation "[ 'cohpred' 'of' T ]" := (@clone T _ _ (erefl _) id)
  (at level 0, format "[ 'cohpred'  'of'  T ]") : form_scope.

(* admit \In notation for elements of cohpred *)
Canonical cohpred_PredType := mkPredType sort. 

Notation fV c := (@fV_op _ (class c)).
Notation sig c := (@sig_op _ (class c)).

Lemma cohE (coh1 coh2 : cohpred) : 
        coh1 = coh2 <-> 
        [/\ fV coh1 = fV coh2, sig coh1 = sig coh2 & 
            coh1 = coh2 :> (state -> Prop)].
Proof.
move: coh1 coh2=>[c1 m1][c2 m2]; split=>/=.
- by case=>E; rewrite -{c2}E in m2 *; move/inj_pair2=>->. 
case: m1=>fV1 u1 ??????; case: m2=>fV2 u2 ??????.  
rewrite /=; case=>E1 X E2; subst fV2 u2 c2.
by congr Pack; congr CohPredMixin; apply: proof_irrelevance.
Qed.

Section Repack.
Variable coh : cohpred.

Lemma cohSO s : s \In coh -> validSO s.
Proof. by case: coh=>c [fV u Vs Vh Vf Vw Vu Vz]; apply: Vs. Qed.

Lemma cohH s : s \In coh -> validH s.
Proof. by case: coh=>c [fV u Vs Vh Vf Vw Vu Vz]; apply: Vh. Qed.

Lemma cohfV s : s \In coh -> [/\ fdom (loc s) =i fV coh, 
                                 fdom (shr s) =i fV coh & 
                                 fdom (env s) =i fV coh]. 
Proof. by case: coh=>c [fV u Vs Vh Vf Vw Vu Vz]; apply: Vf. Qed.

Lemma coh_weq s s' : 
        s \In coh -> s' \In coh -> weq typeCode (shr s) (shr s').
Proof. by case: coh=>c [fV u Vs Vh Vf Vw Vu Vz]; apply: Vw. Qed.

Lemma sig_loc s : s \In coh -> sig coh = [zunit loc s]. 
Proof. by case: coh=>c [fV u Vs Vh Vf Vw Vu Vz]; apply: Vu. Qed.

Lemma coh_zigzag s p : s \< p \In coh <-> s \> p \In coh.
Proof. by case: coh=>c [fV u Vs Vh Vf Vw Vu Vz]; apply: Vz. Qed.

End Repack.
End Exports.
End CohPred.

Export CohPred.Exports.


Section CohLemmas.
Variable coh : cohpred.

Lemma cohV s : s \In coh -> valid s.
Proof. by move/cohH/validHV. Qed.

Lemma coh_footV s s' : s \In coh -> s' \In coh -> footeqV s s'.
Proof. 
move=>C C'; case/and3P: (cohV C)=>V1 V2 V3; case/and3P: (cohV C')=>V1' V2' V3'.
case/cohfV: C=>H1 H2 H3; case/cohfV: C'=>H1' H2' H3'.
apply/and3P; split; apply/fdomeqP.
- by rewrite V1 V1'; split=>// x; rewrite H1 H1'. 
- by rewrite V2 V2'; split=>// x; rewrite H2 H2'. 
by rewrite V3 V3'; split=>// x; rewrite H3 H3'. 
Qed.

Lemma sig_env s : s \In coh -> sig coh = [zunit env s].
Proof.
move=>C; move/sig_loc: (C)=>->.
by case/cohSO/allvalidV/validZP: C. 
Qed.

Lemma cohC lp sh ep : [lp, sh, ep] \In coh <-> [ep, sh, lp] \In coh. 
Proof. 
suff L: forall lp ep, [lp, sh, ep] \In coh -> [ep, sh, lp] \In coh. 
- by split; apply: L.
move=>{lp ep} lp ep C; move/cohSO: C (C); move/allvalidV=>/=. 
case: validZP=>//; case=>_ _ E _.
by rewrite sp_eta /= -coh_zigzag spSS zipC -spSS coh_zigzag -E {1}E -sp_eta.
Qed.

(* coh determines a unique footprint of labels *)
Lemma coh_prec : precise coh.
Proof.
move=>t1 t2 s1 s2 V E C1 C2.
by apply: footeqV_prec V E (coh_footV C1 C2).
Qed.

Lemma transp_coh s : transp s \In coh <-> s \In coh.
Proof. by case: s=>lp sh ep; apply/cohC. Qed.

Lemma cohK s1 s2 : s1 \In coh -> s1 \+ s2 \In coh -> s2 = Unit.
Proof.
move=>H1 H2.
have E: s1 \+ s2 = s1 \+ s2 \+ Unit by rewrite !unitR.
move: (coh_prec (cohV H2) E H1 H2); rewrite -{1 2}[s1]unitR in H1 *.
by move/(joinsK (cohV H1)).
Qed.

Lemma zagcoh_eq s1 s2 p1 p2 s1' s2' p1' p2' : 
        valid (p1 \+ p2) ->
        valid (s1 \+ s2) ->
        p1 \+ p2 = p1' \+ p2' ->
        env (s1 \+ s2) = env (s1' \+ s2') ->
        s1 \> p1 \In coh -> s1' \> p1' \In coh ->
        [/\ p1 = p1', p2 = p2', env s1 = env s1' & env s2 = env s2'].
Proof.
move=>Vp Vs Ep Es C1 C2; case/and3P: Vs=>_ _ Vs.
case/and3P: (coh_footV C1 C2)=>_ _ /fdomeqP [_ F]. 
case/cohV/and3P: C1=>_ _ Vsp1; case/cohV/and3P: C2=>_ _ Vsp2. 
have {F Vsp1 Vsp2} [Ep1 Es1] : p1 = p1' /\ env s1 = env s1'.
by split; [apply: fdomE Vp Ep _ | apply: fdomE Vs Es _]; 
   move=>x; move: (F x); [rewrite !fdom_zagp| rewrite !fdom_zags].
rewrite /= -{p1'}Ep1 -{}Es1 in Ep Es *.
by rewrite (funhKl Vp Ep) (funhKl Vs Es).
Qed.

Lemma zigcoh_eq s1 s2 p1 p2 s1' s2' p1' p2' : 
        valid (p1 \+ p2) ->
        valid (s1 \+ s2) ->
        p1 \+ p2 = p1' \+ p2' ->
        loc (s1 \+ s2) = loc (s1' \+ s2') ->
        s1 \< p1 \In coh -> s1' \< p1' \In coh ->
        [/\ p1 = p1', p2 = p2', loc s1 = loc s1' & loc s2 = loc s2'].
Proof.
rewrite -!env_transp -valid_transp !transpUn.
rewrite -(transp_coh (s1 \< p1)) -(transp_coh (s1' \< p1')) !transp_zig.
by apply: zagcoh_eq.
Qed.

(* funny this proof doesn't use precision of coh *)
(* but footV of coh, so it's stronger *)
Lemma coh_prec_zigL p1 p2 p1' p2' s1 s2 s1' s2' : 
        valid (p1 \+ p1') -> 
        valid (s1 \+ s1') ->
        p1 \+ p1' = p2 \+ p2' -> 
        s1 \+ s1' = s2 \+ s2' ->
        s1 \< p1 \In coh ->
        s2 \< p2 \In coh ->
        [/\ s1 = s2, s1' = s2', p1 = p2 & p1' = p2'].
Proof.
move=>Vp Vs Ep Es C1 C2; case/and3P: (coh_footV C1 C2).
case/fdomeqP=>_ fdomPL /fdomeqP [_ fdomPS] /fdomeqP [_ fdomPE].
case/cohV/and3P: C1=>Vl1 _ _; case/cohV/and3P: C2=>Vl2 _ _. 
suff [Es' Ep'] : s1 = s2 /\ p1 = p2.
- rewrite -Es' -Ep' in Es Ep.
  by rewrite (joinsK Vs Es) (funhKl Vp Ep).
rewrite !spP in Vl1 Vl2 fdomPL fdomPS fdomPE.
have [fdomL fdomP] : fdom (loc s1) =i fdom (loc s2) /\ fdom p1 =i fdom p2.
- split=>x; move: (fdomPL x); first by rewrite !fdomZR Vl1 Vl2. 
  by rewrite !fdomZL Vl1 Vl2.
split; last by apply: fdomE Vp Ep fdomP.
case: Es=>El Es Ee; case/and3P: Vs=>Vl Vs Ve; apply/stE.
by rewrite (fdomE _ El _) // (fdomE _ Es _) // (fdomE _ Ee _).
Qed. 

End CohLemmas.

(*****************************************)
(* Constructors for coherence predicates *)
(*****************************************)

(* coherence for empty state *)

Module EmpCoh.
Section EmpCoh.

Definition emp_pred := [Pred s : state | s = Unit].

Lemma ecohSO s : s \In emp_pred -> validSO s. 
Proof. by move=>->. Qed.

Lemma ecohH s : s \In emp_pred -> validH s. 
Proof. by move=>->. Qed.

Lemma ecohfV s : s \In emp_pred -> footV_eq s pred0. 
Proof. by move=>->. Qed.

Lemma ecoh_weq s s' : 
        s \In emp_pred -> s' \In emp_pred -> weq typeCode (shr s) (shr s').
Proof. by move=>->->. Qed.

Lemma emodunit_loc s :  
        s \In emp_pred -> Unit = [zunit loc s].
Proof. by move=>->; rewrite zunit_unit. Qed.

Lemma ecoh_zigzag s p : (s \< p) \In emp_pred <-> (s \> p) \In emp_pred. 
Proof.
rewrite -[s]st_eta zigE zagE; split.
- case/stE=>/= /zip_unitE [->->] ->->.
  by rewrite -[Unit in Unit \* _]zunit_unit zunitL. 
case/stE=>/= ->-> /zip_unitE [->->].
by rewrite -[Unit in Unit \* _]zunit_unit zunitL. 
Qed.

End EmpCoh.

Module Exports.

Notation emp_pred := emp_pred.

Definition EmpCohMixin := 
  CohPredMixin ecohSO ecohH ecohfV ecoh_weq emodunit_loc ecoh_zigzag.
Definition emp_coh := Eval hnf in CohPred emp_pred EmpCohMixin.

End Exports.
End EmpCoh.

Export EmpCoh.Exports.

(* coherence for disjoint states *)

Module StarCoh.
Section StarCoh.
Variables coh1 coh2 : cohpred.

Let validEntangle s := 
  exists s1 s2, [/\ s = s1 \+ s2, s1 \In coh1 & s2 \In coh2].

Definition star_pred := 
  [Pred s : state | validH s /\ validEntangle s].

Let sigEntangle := sig coh1 \+ sig coh2.

Lemma scohSO s : s \In star_pred -> validSO s. 
Proof.
case=>/validHV V [s1][s2][E] /cohSO Vs1 /cohSO Vs2.
by rewrite {s}E in V *; apply: validSOUn=>//; apply: cohSO.
Qed.

Lemma scohH s : s \In star_pred -> validH s. 
Proof. by case. Qed.

Lemma scohfV s : s \In star_pred -> footV_eq s (predU (fV coh1) (fV coh2)).
Proof.
case=>/validHV/and3P [V1 V2 V3][s1][s2][E] /cohfV [F1 F2 F3] /cohfV [G1 G2 G3].
rewrite {s}E /= in V1 V2 V3 *. 
split=>/= x; rewrite fdomUn inE /= inE /=;
by rewrite ?V1 ?F1 ?G1 ?V2 ?F2 ?G2 ?V3 ?F3 ?G3.  
Qed.

Lemma scoh_weq s s' : 
        s \In star_pred -> s' \In star_pred -> weq typeCode (shr s) (shr s').
Proof.
case=>_ [s1][s2][->{s}] C1 C2 [_][s1'][s2'][->{s'}] C1' C2'.
by apply: weqUn (coh_weq C1 C1') (coh_weq C2 C2').
Qed.

Lemma scoh_sigloc s : s \In star_pred -> sigEntangle = [zunit loc s].
Proof.
case=>_ [s1][s2][->{s}] C1 C2; rewrite /sigEntangle.
by rewrite (sig_loc C1) (sig_loc C2) zunitUn.
Qed.

Lemma scoh_zigzag s p : (s \< p) \In star_pred <-> (s \> p) \In star_pred. 
Proof.
suff L: validSO (s \< p) -> valid (s \< p) ->
  (validEntangle (s \< p) <-> validEntangle (s \> p)).
- split=>C; case: C (C)=>Vh Vs /scohSO Vso.
  - by split; rewrite -?validH_zigzag -?L // -?validSO_zigzag; apply: validHV. 
  split; rewrite ?validH_zigzag ?L ?validSO_zigzag //. 
  by rewrite valid_zigzag ?validSO_zigzag //; apply: validHV.
move=>Vs V; move: (V); rewrite (valid_zigzag Vs)=>V' {Vs}. 
split; case=>s1 [s2][].
- case/and3P: V=>V _ _.
  case/(zig_plusf V)=>s1' [s2'][p1][p2][->->->->] {V' V}.
  move=>/coh_zigzag C1 /coh_zigzag C2.
  exists (s1' \> p1), (s2' \> p2); split=>//. 
  by apply: zag_plus; [case/cohV/and3P: C1 | case/cohV/and3P: C2].
case/and3P: V'=>_ _ V'.
case/(zag_plusf V')=>s1' [s2'][p1][p2][->->->->] {V' V}.
move=>/coh_zigzag C1 /coh_zigzag C2.
exists (s1' \< p1), (s2' \< p2); split=>//. 
by apply: zig_plus; [case/cohV/and3P: C1 | case/cohV/and3P: C2].
Qed.

End StarCoh.

Module Exports.
Notation star_pred := star_pred.

Definition StarCohMixin c1 c2 := 
  CohPredMixin (@scohSO c1 c2) (@scohH c1 c2) (@scohfV c1 c2)
               (@scoh_weq c1 c2) (@scoh_sigloc c1 c2) (@scoh_zigzag c1 c2).
Canonical star_coh c1 c2 := 
  Eval hnf in CohPred (star_pred c1 c2) (StarCohMixin c1 c2).

Lemma starfV c1 c2 : fV (star_coh c1 c2) = predU (fV c1) (fV c2).
Proof. by []. Qed.

Lemma starC c1 c2 : star_coh c1 c2 = star_coh c2 c1. 
Proof.
suff {c1 c2} L c1 c2 s : s \In star_coh c1 c2 -> s \In star_coh c2 c1. 
- apply/cohE; split; first by apply: fext=>x; rewrite starfV /= orbC. 
  - by rewrite /= joinC.
  by apply: fext=>s; apply: pext; split; move/L.
case=>Vh [s1][s2][E C1 C2]; rewrite {s}E in Vh *.
by split=>//; exists s2, s1; rewrite joinC. 
Qed.

Lemma starCA c1 c2 c3 : 
        star_coh c1 (star_coh c2 c3) = star_coh c2 (star_coh c1 c3).
Proof.
suff {c1 c2 c3} L c1 c2 c3 s : 
  s \In star_coh c1 (star_coh c2 c3) ->
  s \In star_coh c2 (star_coh c1 c3).
- apply/cohE; split; first by apply: fext=>s; rewrite /= orbCA.
  - by rewrite /= joinCA.
  by apply: fext=>s; apply: pext; split; move/L. 
case=>Vh [s1][s2][E1 C1][Vh2][t1][t2][E2 T1 T2]. 
rewrite {s}E1 {s2}E2 in Vh Vh2 *.
split=>//; exists t1, (s1 \+ t2); rewrite joinCA in Vh *; split=>//.
split; first by [move/validHR: Vh].
by exists s1, t2. 
Qed.

Lemma starA c1 c2 c3 : 
        star_coh c1 (star_coh c2 c3) = star_coh (star_coh c1 c2) c3.
Proof. by rewrite (starC c2) starCA starC. Qed.

Lemma starAC c1 c2 c3 : 
        star_coh (star_coh c1 c2) c3 = star_coh (star_coh c1 c3) c2.
Proof. by rewrite -!starA (starC c2). Qed.

Lemma star0C c : star_coh emp_coh c = c.
Proof.
apply/cohE; split=>//; first by rewrite /= unitL.
apply: fext=>s; apply: pext; split.
- by case=>_ [s1][s2][->->]; rewrite unitL. 
move=>H; split; first by [move/cohH: H].
by exists Unit, s; rewrite unitL. 
Qed.

Lemma starC0 c : star_coh c emp_coh = c.
Proof. by rewrite starC star0C. Qed.

Lemma star_coh_prec c1 c2 s1 s2 : 
        s1 \+ s2 \In star_coh c1 c2 -> (s1 \In c1 <-> s2 \In c2).
Proof.
case; move/validHV=>V [t1][t2][E H1 H2]; split=>H.
- by move/(coh_prec V E H): H1 (E)=><-; move/(joinsK V)=>->. 
rewrite !(joinC s1) (joinC t1) in V E.
by move/(coh_prec V E H): H2 (E)=><-; move/(joinsK V)=>->. 
Qed.

Lemma star0E c1 c2 : Unit \In star_coh c1 c2 -> Unit \In c1 /\ Unit \In c2.
Proof. by case=>_ [s1][s2][] /esym/st0E [->->]. Qed.

(* very important lemmas for framing wrt. worlds *)
(* it essentially relies on footeqV and footeqH *)
(* as the guarantee transitions of our worlds preserve these footprints *)
(* it will lead to framing wrt. worlds. This justifies the design *)
(* decision that guarantees preserve footprints, which doesn't hold *)
(* in other concurrency logics *)

Lemma star_extend c1 c2 s1 s2 s : 
        s1 \In c1 -> s2 \In c1 -> 
        footeqV s1 s2 -> footeqH (flats s1) (flats s2) ->
        s1 \+ s \In star_coh c1 c2 -> s2 \+ s \In star_coh c1 c2.
Proof.
move=>Cs1 Cs2 Fv Fh; case=>Vh [t1][t2][E Ct1 Ct2].
have {Ct1 E} E : s = t2. 
- move/(coh_prec (validHV Vh) E Cs1): Ct1=>?; subst t1.
  by move/(joinsK (validHV Vh)): E=>?; subst t2.
rewrite {s}E in Vh *; split; last by exists s2, t2. 
by rewrite !validHUn (footeqV_validUnL Fv) (footeqH_validUnL Fh) in Vh *.
Qed.

Lemma cohzag_plusf (c1 c2 : cohpred) s s1 s2 p : 
        s \> p \In star_coh c1 c2 -> s1 \In c1 ->
        s \> p = s1 \+ s2 ->  
        exists s1' s2' p1 p2, 
        [/\ s = s1' \+ s2', p = p1 \+ p2,
            s1 = s1' \> p1,  s2 = s2' \> p2 &
            s \< p = s1' \< p1 \+ s2' \< p2].
Proof.
move=>C C1 E.
have V : valid (env (s \> p)) by case/cohV/and3P: C. 
case/(zag_plusf V): (E)=>{V} s1' [s2'][p1][p2][????]; subst s p s1 s2.
exists s1', s2', p1, p2; split=>//; rewrite {}E in C.
have {C} C2 : s2' \> p2 \In c2 by apply/(star_coh_prec C).
apply: zig_plus.
- by case/coh_zigzag/cohV/and3P: C1. 
by case/coh_zigzag/cohV/and3P: C2.
Qed.

Lemma cohzig_plusf (c1 c2 : cohpred) s s1 s2 p : 
        s \< p \In star_coh c1 c2 -> s1 \In c1 ->
        s \< p = s1 \+ s2 ->  
        exists s1' s2' p1 p2, 
        [/\ s = s1' \+ s2', p = p1 \+ p2,
            s1 = s1' \< p1,  s2 = s2' \< p2 &
            s \> p = s1' \> p1 \+ s2' \> p2].
Proof.
rewrite -[s]transpI -[s1]transpI -[s2]transpI -transp_zag. 
rewrite 2!transp_coh -transpUn transp_inj.
move=>C C1; case/(cohzag_plusf C C1)=>s1' [s2'][p1][p2][->->->->] E.
exists (transp s1'), (transp s2'), p1, p2.
by rewrite !transp_zag -!transp_zig -!transpUn E. 
Qed.

Lemma envzag_plusf (ct cr : cohpred) s1 s2 t1 t2 r1 r2 p : 
        s1 \> p \In star_coh ct cr -> s2 \> p \In star_coh ct cr -> 
        t1 \In ct -> t2 \In ct ->
        env s1 = env s2 ->
        s1 \> p = t1 \+ r1 -> s2 \> p = t2 \+ r2 ->
        exists t1' t2' r1' r2' pt pr, 
          [/\ [/\ s1 = t1' \+ r1', s2 = t2' \+ r2' & p = pt \+ pr], 
              [/\ t1 = t1' \> pt, r1 = r1' \> pr, 
                  t2 = t2' \> pt & r2 = r2' \> pr],
              [/\ env t1' = env t2' & env r1' = env r2'],
              s1 \< p = t1' \< pt \+ r1' \< pr & 
              s2 \< p = t2' \< pt \+ r2' \< pr].
Proof.
move=>C1 C2 C1' C2' H E1 E2.
case/(cohzag_plusf C1 C1'): E1=>s1' [s2'][p1][p2][????] E1; subst s1 p t1 r1. 
case/(cohzag_plusf C2 C2'): E2=>d1' [d2'][q1][q2][? X ??] E2; subst s2 t2 r2.
have [Vq V] : valid (p1 \+ p2) /\ valid (s1' \+ s2').
- by case/cohV/valid_zag: C1. 
case: (zagcoh_eq Vq V X H C1' C2')=>?? Hr Ht; subst q1 q2.
by exists s1', d1', s2', d2', p1, p2. 
Qed.

Lemma envzig_plusf (ct cr : cohpred) s1 s2 t1 t2 r1 r2 p : 
        s1 \< p \In star_coh ct cr -> s2 \< p \In star_coh ct cr -> 
        t1 \In ct -> t2 \In ct ->
        loc s1 = loc s2 ->
        s1 \< p = t1 \+ r1 -> s2 \< p = t2 \+ r2 ->
        exists t1' t2' r1' r2' pt pr, 
          [/\ [/\ s1 = t1' \+ r1', s2 = t2' \+ r2' & p = pt \+ pr], 
              [/\ t1 = t1' \< pt, r1 = r1' \< pr, 
                  t2 = t2' \< pt & r2 = r2' \< pr],
              [/\ loc t1' = loc t2' & loc r1' = loc r2'],
              s1 \> p = t1' \> pt \+ r1' \> pr & 
              s2 \> p = t2' \> pt \+ r2' \> pr].
Proof.
move=>C1 C2 C1' C2' H E1 E2.
case/(cohzig_plusf C1 C1'): E1=>s1' [s2'][p1][p2][????] E1; subst s1 p t1 r1. 
case/(cohzig_plusf C2 C2'): E2=>d1' [d2'][q1][q2][? X ??] E2; subst s2 t2 r2.
have [Vq V] : valid (p1 \+ p2) /\ valid (s1' \+ s2').
- by case/cohV/valid_zig: C1. 
case: (zigcoh_eq Vq V X H C1' C2')=>?? Hr Ht; subst q1 q2.
by exists s1', d1', s2', d2', p1, p2. 
Qed.

Lemma sfilt_star (c1 c2 : cohpred) s : 
        s \In star_coh c1 c2 -> sfilter (fV c1) s \In c1.
Proof.
case=>Vh; case=>s1 [s2][E] C1 C2; rewrite {s}E in Vh *.
case/validHV/and3P: Vh=>V1 V2 V3; case/cohfV: (C1)=>E1 E2 E3.
rewrite /sfilter -lock /=.
have F1 : {in fdom (loc s1 \+ loc s2), fdom (loc s1) =i fV c1}. 
- by move=>s; rewrite /= E1. 
have F2 : {in fdom (shr s1 \+ shr s2), fdom (shr s1) =i fV c1}. 
- by move=>s; rewrite /= E2. 
have F3 : {in fdom (env s1 \+ env s2), fdom (env s1) =i fV c1}. 
- by move=>s; rewrite /= E3. 
rewrite -(eq_in_ffilter F1) ffilt_fdom //.
rewrite -(eq_in_ffilter F2) ffilt_fdom //.
rewrite -(eq_in_ffilter F3) ffilt_fdom //.
by rewrite st_eta. 
Qed.

End Exports.
End StarCoh.

Export StarCoh.Exports.

(* iterated star *)

Module StarIterCoh.
Section StarIterCoh.

Fixpoint validEntangle (ls : seq cohpred) s := 
  if ls is coh :: ls' then 
    exists s1 s2, [/\ s = s1 \+ s2, s1 \In coh & 
                      validEntangle ls' s2]
  else s = Unit. 
  
Definition star_iter_rel ls := 
  [Pred s : state | validH s /\ validEntangle ls s].

Lemma sscohSO ls s : s \In star_iter_rel ls -> validSO s. 
Proof.
elim: ls s=>[|coh ls IH] s; first by case=>_ ->. 
case=>Vh /= [s1][s2][E] /cohSO Vs V; rewrite {s}E in Vh *.
apply: validSOUn (validHV Vh) Vs _.
by apply: IH; split=>//; apply: validHR Vh.
Qed.

Lemma sscohH ls s : s \In star_iter_rel ls -> validH s. 
Proof. by case. Qed.

Lemma foldl_predU A (c : pred A) (t : seq (pred A)) : 
       foldl predU c t = predU c (foldl predU pred0 t).
Proof.
elim: t c=>[|x t IH] c /=; first by apply: fext=>x /=; rewrite orbF.
by apply: fext=>z; rewrite IH /= -orbA (IH (predU pred0 x)).
Qed.

Lemma sscohfV ls s : 
        s \In star_iter_rel ls -> 
        footV_eq s (foldl predU pred0 [seq fV c | c <- ls]).
Proof.
elim: ls s=>[|coh ls IH] s /=; first by case=>_ ->. 
case=>Vh /= [s1][s2][E] C1 C2; rewrite {s}E in Vh *.
rewrite foldl_predU.
have: s2 \In star_iter_rel ls by split=>//; apply: validHR Vh.
case/IH=>E1 E2 E3.
case/cohfV: C1=>S1 S2 S3. 
case/and3P: (validHV Vh)=>V1 V2 V3.
split=>x /=.
- by rewrite fdomUn inE V1 /= inE /= S1 E1.
- by rewrite fdomUn inE V2 /= inE /= S2 E2.
by rewrite fdomUn inE V3 /= inE /= S3 E3.
Qed.

Lemma sscoh_weq ls s s' : 
        s \In star_iter_rel ls -> s' \In star_iter_rel ls ->
        weq typeCode (shr s) (shr s').
Proof.
elim: ls s s'=>[|coh ls IH] s s' /=; first by case=>_ -> [_ ->].
case=>Vh /= [s1][s2][?] C1 H1 [/= Vh'][s1'][s2'][?] C2 H2; subst s s'. 
apply: weqUn; first by apply: coh_weq C1 C2.
by apply: IH; split=>//; [apply: validHR Vh | apply: validHR Vh'].
Qed.

Lemma foldl_sig (c : feap dpcm) (t : seq (feap dpcm)) : 
       foldl PCM.join c t = c \+ (foldl PCM.join Unit t).
Proof.
elim: t c=>[|x t IH] c /=; first by rewrite unitR. 
by rewrite IH unitL (IH x) joinA.  
Qed.

Lemma sscoh_sig ls s : 
        s \In star_iter_rel ls -> 
        foldl PCM.join Unit [seq sig c | c <- ls] = [zunit loc s].
Proof.
elim: ls s=>[|coh ls IH] s /=; first by case=>_ ->; rewrite zunit_unit. 
case=>H /= [s1][s2][E] C1 V; rewrite {s}E in H *.
have C : s2 \In star_iter_rel ls by split=>//; apply: validHR H. 
by rewrite unitL foldl_sig (IH _ C) (sig_loc C1) zunitUn. 
Qed.

Lemma sscoh_zigzag ls s p : 
        (s \< p) \In star_iter_rel ls <-> 
        (s \> p) \In star_iter_rel ls. 
Proof.
elim: ls s p=>[|coh ls IH] s p /=.
- split; case=>_. 
  - by case/zig_unitf=>->->; rewrite zagE unit_zipE.  
  by case/zag_unitf=>->->; rewrite zigE unit_zipE. 
split; move=>S; move: (sscohSO S)=>Vso;
case: (S)=>/= Vh [s1][s2][E] C V; rewrite E in S.
- split; first by rewrite -validH_zigzag.
  case/zig_plusf: E; first by case/validHV/and3P: Vh.
  move=>s1' [s2'][p1][p2][????]; subst s p s1 s2.
  have: s2' \< p2 \In star_iter_rel ls.
  - split=>//; rewrite zig_plus in Vh.
    - by move/validHR: Vh. 
    - by case/cohV/and3P: C.
    by case: S; case/validHR/validHV/and3P. 
  case/IH=>Vh' V'; move/coh_zigzag: C=>C.
  exists (s1' \> p1), (s2' \> p2). 
  split=>//; rewrite zag_plus //.
   - by case/cohV/and3P: C. 
  by case/validHV/and3P: Vh'. 
split; first by rewrite validH_zigzag // validSO_zigzag.
case/zag_plusf: E; first by case/validHV/and3P: Vh. 
move=>s1' [s2'][p1][p2][????]; subst s p s1 s2.
have: s2' \> p2 \In star_iter_rel ls.
- split=>//; rewrite zag_plus in Vh.
  - by move/validHR: Vh. 
  - by case/cohV/and3P: C. 
  by case: S; case/validHR/validHV/and3P. 
case/IH=>Vh' V'; move/coh_zigzag: C=>C.
exists (s1' \< p1), (s2' \< p2). 
split=>//; rewrite zig_plus //.
- by case/cohV/and3P: C. 
by case/validHV/and3P: Vh'. 
Qed.

End StarIterCoh.

Module Exports.
Notation star_iter_rel := star_iter_rel.

Definition StarIterCohMixin ls := 
  CohPredMixin (@sscohSO ls) (@sscohH ls) (@sscohfV ls) 
               (@sscoh_weq ls) (@sscoh_sig ls) (@sscoh_zigzag ls).
Definition star_iter_coh ls := 
  Eval hnf in CohPred (star_iter_rel ls) (StarIterCohMixin ls).

Lemma star_iter_fV ls : 
        fV (star_iter_coh ls) = foldl predU pred0 [seq fV c | c <- ls]. 
Proof. by []. Qed.

Lemma star_iter_emp : star_iter_coh [::] = emp_coh.
Proof. 
apply/cohE; split=>//; apply: fext=>s; apply: pext.
by split; [case=>_ -> | move=>->]. 
Qed.
 
Lemma star_iter_cons coh ls : 
        star_iter_coh (coh :: ls) = star_coh coh (star_iter_coh ls).
Proof.
elim: ls coh=>[|coh' ls IH] coh /=.
- rewrite star_iter_emp starC0. 
  apply/cohE; split=>//; first by rewrite /= unitL.
  apply: fext=>s1; apply: pext; split=>/=. 
  - by case=>Vh1 [s2][s3][E] C1 ?; subst s1 s3; rewrite unitR.
  move=>C; split; first by apply: cohH C. 
  by exists s1, Unit; rewrite unitR.
apply/cohE; split.
- rewrite IH !starfV !star_iter_fV. 
  by apply: fext=>s /=; rewrite foldl_predU orbA. 
- by rewrite IH /= 2!foldl_sig !unitL joinA. 
apply: fext=>s1; apply: pext.
split=>/=.
- case=>Vh1 [s2][s3][?] C2 [s4][s5][?] C4 V; subst s1 s3.
  split=>//; exists s2, (s4 \+ s5); split=>//=.
  rewrite IH; split; first by move/validHR: Vh1. 
  by exists s4, s5; do !split=>//; move/validHR/validHR: Vh1. 
case=>Vh1 [s2][s3][?] C2 [Vh3] /= [s4][s5][?] C4 V; subst s1 s3.
by split=>//; exists s2, (s4 \+ s5); split=>//; exists s4, s5. 
Qed.

Lemma star_iter_cat ls1 ls2 : 
        star_iter_coh (ls1 ++ ls2) = 
        star_coh (star_iter_coh ls1) (star_iter_coh ls2).
Proof.
elim: ls1 ls2=>[|coh ls1 IH] ls2 /=.
- by rewrite star_iter_emp star0C.
by rewrite !star_iter_cons IH starA. 
Qed.

Definition ssE := (star_iter_emp, star_iter_cons, star_iter_cat).

Lemma star_iter_perm ls1 ls2 : 
        perm ls1 ls2 -> star_iter_coh ls1 = star_iter_coh ls2.
Proof.
set P := fun ls1 ls2 => star_iter_coh ls1 = star_iter_coh ls2. 
suff {ls1 ls2} L : forall ls1 ls2, perm ls1 ls2 -> P ls1 ls2.
- by apply: L.
apply: perm_ind2=>//.
- move=>coh ls1 ls2 H IH; rewrite /P in IH *. 
  by rewrite !ssE IH.
- move=>c1 c2 ls1 ls2 H IH; rewrite /P in IH *. 
  by rewrite !ssE starCA IH.  
move=>s2 s1 s3 H1 IH1 H2 IH2; rewrite /P in IH1 IH2 *. 
by rewrite IH1 IH2.
Qed.

End Exports.
End StarIterCoh.

Export StarIterCoh.Exports.


(***********************)
(***********************)
(* Guarantee relations *)
(***********************)
(***********************)

Notation reltp := (state -> state -> Prop).

Module Guar.
Section ClassDef.
Variable coh : cohpred.

Structure mixin_of (guar : reltp) := Mixin {
  _ : forall s s', guar s s' -> s \In coh /\ s' \In coh;
  _ : forall s s', guar s s' -> env s = env s';
  (* precondition env s = env s' added to avoid cancellativity *)
  _ : forall s s' p, env s = env s' -> 
        guar (s \> p) (s' \> p) -> guar (s \< p) (s' \< p)}.

Definition class_of := mixin_of.

Structure guar := Pack {sort : reltp; _ : class_of sort}.
Local Coercion sort : guar >-> Funclass.

Variables (T : reltp) (cT : guar).
Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.
Definition pack c := @Pack T c.
Definition clone := fun c & T = cT & phant_id (pack c) cT => pack c. 

End ClassDef.

Module Exports.
Coercion sort : guar >-> Funclass.
Notation guar := guar.
Notation GuarMixin := Mixin.
Notation Guar T m := (@pack _ T m).

Notation "[ 'guarMixin' 'of' T ]" := (@class _ _ : mixin_of _ T)
  (at level 0, format "[ 'guarMixin'  'of'  T ]") : form_scope.
Notation "[ 'guar' 'of' T 'for' C ]" := (@clone _ T C _ (erefl _) id)
  (at level 0, format "[ 'guar'  'of'  T  'for'  C ]") : form_scope.
Notation "[ 'guar' 'of' T ]" := (@clone _ T _ _ (erefl _) id)
  (at level 0, format "[ 'guar'  'of'  T ]") : form_scope.

Lemma guarE coh (g1 g2 : guar coh) : 
        g1 = g2 <-> g1 = g2 :> reltp.
Proof.
move: g1 g2=>[r1 m1][r2 m2] /=; split; first by case.
move=>E; rewrite -{r2}E in m1 m2 *; case: m1=>*; case: m2=>*.
by congr Pack; congr GuarMixin; apply: proof_irrelevance.
Qed.

Section Repack. 
Variables (coh : cohpred) (g : guar coh).

Lemma guar_coh s s' : g s s' -> s \In coh /\ s' \In coh.
Proof. by case: g=>G [Gc Ge Gz]; apply: Gc. Qed.

Lemma guar_env s s' : g s s' -> env s = env s'.
Proof. by case: g=>G [Gc Ge Gz]; apply: Ge. Qed.

Lemma guar_zagzig s s' p : 
        env s = env s' -> g (s \> p) (s' \> p) -> g (s \< p) (s' \< p).
Proof. by case: g=>G [Gc Ge Gz]; apply: Gz. Qed.

End Repack.
End Exports.
End Guar.

Export Guar.Exports.

(* Guarantee refinements *)

Module InternalGuar.
Section ClassDef.
Variable coh : cohpred.

Structure mixin_of (g : guar coh) := 
  Mixin of (forall s s', g s s' -> footeqH (flats s) (flats s')).

Notation class_of := mixin_of (only parsing).

Structure int_guar := Pack {sort; _ : class_of sort}.
Local Coercion sort : int_guar >-> Guar.guar.

Variables (T : guar coh) (cT : int_guar).
Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.
Definition pack m := @Pack T m.
Definition clone := fun c & T = cT & phant_id (pack c) cT => pack c. 

End ClassDef.

Module Exports.
Coercion sort : int_guar >-> Guar.guar.
Notation int_guar := int_guar.
Notation IntGuarMixin := Mixin.
Notation IntGuar T m := (@pack _ T m).

Notation "[ 'int_guar' 'of' T 'for' C ]" := (@clone _ T C _ (erefl _) id)
  (at level 0, format "[ 'int_guar'  'of'  T  'for'  C ]") : form_scope.
Notation "[ 'int_guar' 'of' T ]" := (@clone _ T _ _ (erefl _) id)
  (at level 0, format "[ 'int_guar'  'of'  T ]") : form_scope.

Lemma int_guarE coh (g1 g2 : int_guar coh) : 
        g1 = g2 <-> g1 = g2 :> reltp.
Proof.
move: g1 g2=>[g1 m1][g2 m2] /=; split; first by case=>->. 
move/guarE=>E; rewrite -{g2}E in m2 *; case: m1; case: m2=>*. 
by congr (Pack (IntGuarMixin _)); apply: proof_irrelevance.
Qed.

Lemma int_guar_footH coh (g : int_guar coh) s s' : 
        g s s' -> footeqH (flats s) (flats s').
Proof. by case: g=>G []; apply. Qed.

End Exports.
End InternalGuar.

Export InternalGuar.Exports.

Module ReflexiveGuar.
Section ClassDef. 
Variable coh : cohpred. 

Structure mixin_of (g : guar coh) : Type := 
  Mixin of (forall s, s \In coh -> g s s).

Structure class_of g : Type := Class {
  base : InternalGuar.mixin_of g;
  mixin : mixin_of g}. 
Local Coercion base : class_of >-> InternalGuar.mixin_of.

Structure refl_guar := Pack {sort; _ : class_of sort}.
Local Coercion sort : refl_guar >-> guar.

Variables (T : guar coh) (cT : refl_guar).
Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.
Definition clone c of phant_id class c := @Pack T c. 
Definition pack m := 
  fun b bT & phant_id (@InternalGuar.class coh bT) b => Pack (@Class T b m).

(* inheritance *)
Definition to_internal := InternalGuar.Pack class.

End ClassDef.

Module Exports.
Coercion base : class_of >-> InternalGuar.mixin_of. 
Coercion sort : refl_guar >-> guar.
Coercion to_internal : refl_guar >-> int_guar.
Canonical to_internal.

Notation refl_guar := refl_guar.
Notation ReflGuarMixin := Mixin.
Notation ReflGuar T m := (@pack _ T m _ _ id).

Notation "[ 'refl_guar' 'of' T 'for' C ]" := (@clone _ T C _ id)
  (at level 0, format "[ 'refl_guar'  'of'  T  'for'  C ]") : form_scope.
Notation "[ 'refl_guar' 'of' T ]" := (@clone _ T _ _ id)
  (at level 0, format "[ 'refl_guar'  'of'  T ]") : form_scope.

Lemma refl_guarE coh (g1 g2 : refl_guar coh) : 
        g1 = g2 <-> g1 = g2 :> reltp.
Proof.
move: g1 g2=>[g1 m1][g2 m2] /=; split; first by case=>->. 
move/guarE=>E; rewrite -{g2}E in m2 *; case: m1=>*; case: m2=>*.
by congr Pack; congr Class; apply: proof_irrelevance.
Qed.

Lemma refl_guar_refl coh (g : refl_guar coh) s : s \In coh -> g s s.
Proof. by case: g=>G [b []]; apply. Qed.

End Exports.
End ReflexiveGuar.

Export ReflexiveGuar.Exports.


Module AcquireGuar.
Section ClassDef.
Variable coh : cohpred.

Structure mixin_of (g : heap -> guar coh) := 
  Mixin of forall h s s', g h s s' -> footeqH (flats s \+ h) (flats s').

Definition class_of := mixin_of.

Structure acq_guar : Type := Pack {sort : heap -> guar coh; _ : class_of sort}.
Local Coercion sort : acq_guar >-> Funclass.

Variables (T : heap -> guar coh) (cT : acq_guar).
Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.
Definition pack c := @Pack T c.
Definition clone := fun c & T = cT & phant_id (pack c) cT => pack c. 

End ClassDef.

Module Exports.
Coercion sort : acq_guar >-> Funclass.
Notation acq_guar := acq_guar.
Notation AcqGuarMixin := Mixin.
Notation AcqGuar T m := (@pack _ T m).

Notation "[ 'acq_guarMixin' 'of' T ]" := (@class _ _ : mixin_of _ T)
  (at level 0, format "[ 'acq_guarMixin'  'of'  T ]") : form_scope.
Notation "[ 'acq_guar' 'of' T 'for' C ]" := (@clone _ T C _ (erefl _) id)
  (at level 0, format "[ 'acq_guar'  'of'  T  'for'  C ]") : form_scope.
Notation "[ 'acq_guar' 'of' T ]" := (@clone _ T _ _ (erefl _) id)
  (at level 0, format "[ 'acq_guar'  'of'  T ]") : form_scope.

Lemma acq_guarE coh (g1 g2 : acq_guar coh) : 
        g1 = g2 <-> g1 = g2 :> (heap -> guar coh).
Proof.
move: g1 g2=>[r1 m1][r2 m2] /=; split; first by case.
move=>E; rewrite -{r2}E in m1 m2 *; case: m1=>*; case: m2=>*.
by congr Pack; congr AcqGuarMixin; apply: proof_irrelevance.
Qed.

Lemma acq_guar_footH coh (g : acq_guar coh) h s s' : 
        g h s s' -> footeqH (flats s \+ h) (flats s').
Proof. by case: g=>G []; apply. Qed.

(*
(* explicit coercion into heap -> state -> state -> Prop *)
Definition from_acq coh (g : acq_guar coh) h s1 s2 := g h s1 s2.
*)

End Exports.
End AcquireGuar.

Export AcquireGuar.Exports.


Module ReleaseGuar.
Section ClassDef.
Variable coh : cohpred.

Structure mixin_of (g : heap -> guar coh) := 
  Mixin of forall h s s', g h s s' -> footeqH (flats s) (h \+ flats s').

Definition class_of := mixin_of.

Structure rel_guar : Type := Pack {sort : heap -> guar coh; _ : class_of sort}.
Local Coercion sort : rel_guar >-> Funclass.

Variables (T : heap -> guar coh) (cT : rel_guar).
Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.
Definition pack c := @Pack T c.
Definition clone := fun c & T = cT & phant_id (pack c) cT => pack c. 

End ClassDef.

Module Exports.
Coercion sort : rel_guar >-> Funclass.
Notation rel_guar := rel_guar.
Notation RelGuarMixin := Mixin.
Notation RelGuar T m := (@pack _ T m).

Notation "[ 'rel_guarMixin' 'of' T ]" := (@class _ _ : mixin_of _ T)
  (at level 0, format "[ 'rel_guarMixin'  'of'  T ]") : form_scope.
Notation "[ 'rel_guar' 'of' T 'for' C ]" := (@clone _ T C _ (erefl _) id)
  (at level 0, format "[ 'rel_guar'  'of'  T  'for'  C ]") : form_scope.
Notation "[ 'rel_guar' 'of' T ]" := (@clone _ T _ _ (erefl _) id)
  (at level 0, format "[ 'rel_guar'  'of'  T ]") : form_scope.

Lemma rel_guarE coh (g1 g2 : rel_guar coh) : 
        g1 = g2 <-> g1 = g2 :> (heap -> guar coh).
Proof.
move: g1 g2=>[r1 m1][r2 m2] /=; split; first by case.
move=>E; rewrite -{r2}E in m1 m2 *; case: m1=>*; case: m2=>*.
by congr Pack; congr RelGuarMixin; apply: proof_irrelevance.
Qed.

Lemma rel_guar_footH coh (g : rel_guar coh) h s s' : 
        g h s s' -> footeqH (flats s) (h \+ flats s').
Proof. by case: g=>G []; apply. Qed.

(*
(* explicit coercion into heap -> state -> state -> Prop *)
Definition from_rel coh (g : rel_guar coh) h s1 s2 := g h s1 s2.
*)

End Exports.
End ReleaseGuar.

Export ReleaseGuar.Exports.

    
(**********)
(* Lemmas *)
(**********)

(* V lemmas *)

Lemma guar_footV coh (G : guar coh) s s' : G s s' -> footeqV s s'.
Proof. by case/guar_coh; apply: coh_footV. Qed.

Lemma guarV_validUnL {s} coh (G : guar coh) s1 s2 : 
        G s1 s2 -> valid (s1 \+ s) = valid (s2 \+ s).
Proof. by move/guar_footV; apply: footeqV_validUnL. Qed.

Lemma guarV_validUnR {s} coh (G : guar coh) s1' s2' : 
        G s1' s2' -> valid (s \+ s1') = valid (s \+ s2').
Proof. by rewrite !(joinC s); apply: guarV_validUnL. Qed.

Lemma guarV_validUn coh coh' (G : guar coh) (G' : guar coh') s1 s2 s1' s2' :
        G s1 s2 -> G' s1' s2' -> 
        valid (s1 \+ s1') = valid (s2 \+ s2').
Proof. by move/guarV_validUnL=>-> /guarV_validUnR=>->. Qed.

(* SO lemmas *)

Lemma guarSO_validUnL coh (G : guar coh) s1 s2 {s} :
        G s1 s2 -> validSO (s1 \+ s) = validSO (s2 \+ s).
Proof.
suff L t1 t2 t : validSO t1 -> validSO t2 -> 
  fdom_eq (loc t1) (loc t2) -> validSO (t1 \+ t) -> validSO (t2 \+ t).
- move=>H; case/guar_coh: H (H)=>/cohSO V1 /cohSO V2.
  case/guar_footV/and3P=>E _ _.
  by apply/idP/idP; apply: L=>//; rewrite fdomeq_sym.
rewrite /validSO => V1 V2 E Vs.
have V: valid (loc t \* env t).
- case/allvalidV/validZP: Vs => W1 W2 F.
  apply/validZP; rewrite (validR W1) (validR W2); split=>//.
  rewrite -valid_zunit !zunitUn /= in W1 F.  
  suff T : [zunit loc t1] = [zunit env t1]. 
  - by rewrite -T in F; apply: funhKl W1 F.
  by case/allvalidV/validZP: V1.
rewrite -!zipUn // ?allvalidV // !allvalidUn V1 V2 !andbT in Vs *.
case/andP: Vs=>Vs ->; rewrite andbT.
case: fvalidUn Vs=>// _ _ Vs _; case: fvalidUn=>//.
- by move/allvalidV: V2=>->.
- by rewrite V.
move=>x; move: {Vs} (Vs x)=>Vs.
rewrite !fdomZL V !allvalidV // in Vs *. 
by case/fdomeqP: E=>_ <-; move/Vs/negbTE=>->.
Qed.

Lemma guarSO_validUnR coh (G : guar coh) s1' s2' {s} : 
        G s1' s2' -> validSO (s \+ s1') = validSO (s \+ s2').
Proof. by rewrite !(joinC s); apply: guarSO_validUnL. Qed.

Lemma guarSO_validUn coh coh' (G : guar coh) (G' : guar coh') s1 s2 s1' s2' :
        G s1 s2 -> G' s1' s2' -> validSO (s1 \+ s1') = validSO (s2 \+ s2').
Proof.
by move=>H1 H2; rewrite -(guarSO_validUnL H1) -(guarSO_validUnR H2).
Qed.

(* H lemmas *)

Lemma guarH_validUnL {s} coh (G : int_guar coh) s1 s2 : 
        G s1 s2 -> validH (s1 \+ s) = validH (s2 \+ s).
Proof.
move=>H; rewrite !validHUn (guarV_validUnL H).
by rewrite (footeqH_validUnL (int_guar_footH H)).
Qed.

Lemma guarH_validUnR {s} coh (G : int_guar coh) s1' s2' : 
        G s1' s2' -> validH (s \+ s1') = validH (s \+ s2').
Proof. by rewrite !(joinC s); apply: guarH_validUnL. Qed.

Lemma guarH_validUnLR coh1 (G1 : acq_guar coh1) 
                      coh2 (G2 : rel_guar coh2) s1 s1' s2 s2' h : 
        G1 h s1 s2 -> G2 h s1' s2' -> 
        validH (s1 \+ s1') = validH (s2 \+ s2').
Proof.
move=>H H'; rewrite !validHUn (guarV_validUnL H) (guarV_validUnR H').
rewrite -(footeqH_validUnL (acq_guar_footH H)).
by rewrite (footeqH_validUnR (rel_guar_footH H')) joinA.
Qed.

Lemma guarH_validUnRL coh1 (G1 : rel_guar coh1) 
                      coh2 (G2 : acq_guar coh2) s1 s1' s2 s2' h : 
        G1 h s1 s2 -> G2 h s1' s2' ->
        validH (s1 \+ s1') = validH (s2 \+ s2').
Proof.
by rewrite (joinC s1) (joinC s2)=>H H'; apply: guarH_validUnLR H' H.
Qed.

(* extension lemma *)

Lemma int_guar_extend c1 c2 (g : int_guar c1) s1 s2 s : 
        g s1 s2 -> (s1 \+ s \In star_coh c1 c2 <-> s2 \+ s \In star_coh c1 c2).
Proof.
move=>G; case/guar_coh: G (guar_footV G) (int_guar_footH G)=>C1 C2 Fv Fh.
split; apply: star_extend=>//.
- by rewrite footeqV_sym.
by apply: footeqH_sym.
Qed.

(********************************)
(* Constructors for transitions *)
(********************************)

(* Zero transition *)

Module ZeroGuar.
Section ZeroGuar.
Variable coh : cohpred.

Definition zero_rel (s s' : state) := False. 

Lemma zeroguar_coh s s' : zero_rel s s' -> s \In coh /\ s' \In coh.
Proof. by []. Qed.

Lemma zeroguar_env s1 s2 : zero_rel s1 s2 -> env s1 = env s2.
Proof. by []. Qed.

Lemma zeroguar_zagzig s s' p : 
        env s = env s' -> 
        zero_rel (s \> p) (s' \> p) -> zero_rel (s \< p) (s' \< p).
Proof. by []. Qed.

Lemma zeroguar_footH s s' : zero_rel s s' -> footeqH (flats s) (flats s').
Proof. by []. Qed.

End ZeroGuar.

Module Exports.
Notation zero_rel := zero_rel.

Definition ZeroGuarMixin coh := 
  GuarMixin (@zeroguar_coh coh) zeroguar_env zeroguar_zagzig.
Canonical zero_guar coh := Eval hnf in Guar zero_rel (ZeroGuarMixin coh).
Definition IntZeroGuarMixin coh := 
  @IntGuarMixin coh (zero_guar coh) zeroguar_footH.
Canonical IntZeroGuar coh := 
  Eval hnf in IntGuar (zero_guar coh) (IntZeroGuarMixin coh).

End Exports.
End ZeroGuar.

Export ZeroGuar.Exports.


(* Identity transition *)

Module IdGuar.
Section IdGuar.
Variable coh : cohpred.

Definition id_rel s s' := s = s' /\ s \In coh.

Lemma idguar_coh s s' : id_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>->. Qed.

Lemma idguar_env s1 s2 : id_rel s1 s2 -> env s1 = env s2.
Proof. by case=>->. Qed.

Lemma idguar_zagzig s s' p : 
        env s = env s' -> 
        id_rel (s \> p) (s' \> p) -> id_rel (s \< p) (s' \< p).
Proof. by move=>E [/(zagzig_eq E)] -> /coh_zigzag. Qed.

Lemma idguar_footH s s' : id_rel s s' -> footeqH (flats s) (flats s').
Proof. by case=>-> /cohH. Qed. 

Lemma idguar_refl s : s \In coh -> id_rel s s.
Proof. by []. Qed.

End IdGuar.

Module Exports.
Notation id_rel := id_rel.

Definition IdGuarMixin coh := 
  GuarMixin (@idguar_coh coh) (@idguar_env coh) (@idguar_zagzig coh).
Canonical id_guar coh := Eval hnf in Guar (id_rel coh) (IdGuarMixin coh).
Definition IntIdGuarMixin coh := IntGuarMixin (@idguar_footH coh).
Canonical IntIdGuar coh := 
  Eval hnf in IntGuar (id_guar coh) (IntIdGuarMixin coh).
Definition ReflIdGuarMixin coh := ReflGuarMixin (@idguar_refl coh).
Canonical ReflIdGuar coh := 
  Eval hnf in ReflGuar (id_guar coh) (ReflIdGuarMixin coh).

End Exports.
End IdGuar.

Export IdGuar.Exports.

(*********************)
(* Framed transition *)
(*********************)

Module FrameGuar.
Section FrameGuar.
Variables coh1 coh2 : cohpred. (* coh1 is base, coh2 is the addition *)

Notation coh := (star_coh coh1 coh2).

(* we insist s' \In coh to ensure that in transfering guars *)
(* the transfered heap doesn't interfere with the framing *)
Definition frame_rel (g : reltp) s s' :=
  [/\ exists s1 s2 s1', 
        [/\ s = s1 \+ s2, s' = s1' \+ s2 & g s1 s1'], 
      s \In coh & s' \In coh].

Section Base.
Variable g : guar coh1.

Lemma fguar_coh s s' : frame_rel g s s' -> s \In coh /\ s' \In coh.
Proof. by case. Qed.

Lemma fguar_env s s' : frame_rel g s s' -> env s = env s'.
Proof. by case; case=>s1 [s2][s1'][->->] /=; move/guar_env=>->. Qed.

Lemma fguar_zagzig s s' p : 
        env s = env s' -> 
        frame_rel g (s \> p) (s' \> p) -> frame_rel g (s \< p) (s' \< p).
Proof.
move=>H G; split; try by apply/coh_zigzag; case/fguar_coh: G.
case: G; case=>t [t2][t'][E E'] G C C'; case/guar_coh: (G)=>T T'.
case: (envzag_plusf C C' T T' H E E')=>t1' [t2'][r1][r2][q'][q][].
case=>??? [??? X][Et Er] F F'; subst s s' p t t2 t'.
move/(zagzig_eq Er): X=>?; subst r2. 
exists (t1' \< q'), (r1 \< q), (t2' \< q'); split=>//.
by apply: (guar_zagzig Et).
Qed.

End Base.

(* framing for refinements *)

Lemma fintguar_footH (g : int_guar coh1) s s' : 
        frame_rel g s s' -> footeqH (flats s) (flats s').
Proof.
case; case=>s1 [s2][s1'][->->{s s'}] H /cohV V /cohV V'.
by rewrite !flatsUn //; apply: footeqHUn (int_guar_footH H) _.
Qed.

Lemma freflguar_refl (g : refl_guar coh1) s : 
        s \In star_coh coh1 coh2 -> frame_rel g s s.
Proof.
move=>C; split=>//; case: C=>_ [s1][s2][-> C1 _].
by exists s1, s2, s1; split=>//; apply: refl_guar_refl.
Qed.

Lemma facqguar_footH (g : acq_guar coh1) h s s' :
        frame_rel (g h) s s' -> footeqH (flats s \+ h) (flats s').
Proof.
case; case=>s1 [s2][s1'][->->{s s'}] H /cohV V /cohV V'.
by rewrite !flatsUn // joinAC; apply: footeqHUn (acq_guar_footH H) _.
Qed.

Lemma frelguar_footH (g : rel_guar coh1) h s s' : 
        frame_rel (g h) s s' -> footeqH (flats s) (h \+ flats s').
Proof.
case; case=>s1 [s2][s1'][->->{s s'}] H /cohV V /cohV V'. 
by rewrite !flatsUn // joinA; apply: footeqHUn (rel_guar_footH H) _.
Qed.

End FrameGuar.

Module Exports.

Notation frame_rel := frame_rel.

Section Exports.
Variables coh1 coh2 : cohpred.

Definition FrameGuarMixin g :=
  GuarMixin (@fguar_coh coh1 coh2 g) (@fguar_env coh1 coh2 g)
            (@fguar_zagzig coh1 coh2 g).
Canonical frame_guar (g : guar coh1) := 
  Eval hnf in Guar (frame_rel coh1 coh2 g) (FrameGuarMixin g).

Definition FrameIntGuarMixin (g : int_guar coh1) :
  InternalGuar.mixin_of (frame_guar g) := 
  IntGuarMixin (@fintguar_footH coh1 coh2 g).
Canonical FrameIntGuar (g : int_guar coh1) := 
  Eval hnf in IntGuar (frame_guar g) (FrameIntGuarMixin g).

Definition FrameReflGuarMixin (g : refl_guar coh1) : 
  ReflexiveGuar.mixin_of (frame_guar g) := 
  ReflGuarMixin (@freflguar_refl coh1 coh2 g).
Canonical FrameReflGuar (g : refl_guar coh1) := 
  Eval hnf in ReflGuar (frame_guar g) (FrameReflGuarMixin g).

(*
Definition baseFrameAcqGuarMixin (g : acq_guar coh1) h := FrameGuarMixin (g h).
Definition acq_frame_guar (g : acq_guar coh1) h :=
  Eval hnf in Guar (frame_rel coh2 (g h)) (baseFrameAcqGuarMixin g h).
*)

Definition acq_frame_guar (g : acq_guar coh1) := fun h => frame_guar (g h). 
Definition FrameAcqGuarMixin (g : acq_guar coh1) : 
  AcquireGuar.mixin_of (acq_frame_guar g) := 
  AcqGuarMixin (@facqguar_footH coh1 coh2 g).
Canonical FrameAcqGuar g := 
  Eval hnf in AcqGuar (acq_frame_guar g) (FrameAcqGuarMixin g).

(*
Definition baseFrameRelGuarMixin (g : rel_guar coh1) h := FrameGuarMixin (g h).
Definition rel_frame_guar (g : rel_guar coh1) h :=
  Eval hnf in Guar (frame_rel coh2 (g h)) (baseFrameRelGuarMixin g h).
*)

Definition rel_frame_guar (g : rel_guar coh1) := fun h => frame_guar (g h).
Definition FrameRelGuarMixin (g : rel_guar coh1) :
  ReleaseGuar.mixin_of (rel_frame_guar g) :=
  RelGuarMixin (@frelguar_footH coh1 coh2 g).
Canonical FrameRelGuar g := 
  Eval hnf in RelGuar (rel_frame_guar g) (FrameRelGuarMixin g).

End Exports.
End Exports.

End FrameGuar.

Export FrameGuar.Exports.

Lemma frame_relE (c1 c2 c1' c2' : cohpred) (g1 : reltp) (g2 : reltp) :
        c1 = c2 -> c1' = c2' -> g1 = g2 ->
        frame_rel c1 c1' g1 = frame_rel c2 c2' g2.
Proof.
move=>Ec Ec' Eg.  
apply: fext=>s; apply: fext=>s'; apply: pext; split.
- case; case=>s1 [s2][s1'][->->]; rewrite Eg Ec Ec' => G C C'. 
  by split=>//; exists s1, s2, s1'. 
case; case=>s1 [s2][s1'][->->]; rewrite -Eg -Ec -Ec' => G C C'. 
by split=>//; exists s1, s2, s1'. 
Qed.

Lemma frame_relA (c c1 c2 : cohpred) (g : guar c) : 
        frame_rel (star_coh c c1) c2 (frame_rel c c1 g) = 
        frame_rel c (star_coh c1 c2) g.
Proof.
apply: fext=>s; apply: fext=>s'; apply: pext; split.
- case; case=>s01 [s2][s01'][->->{s s'}] G C C'.
  have C2 : s2 \In c2 by rewrite -(star_coh_prec C); case/guar_coh: G.
  split; try by rewrite starA.
  case: G C C'; case=>s0 [s1][s0'][->->{s01 s01'}] G _ _ C C'.
  by exists s0, (s1 \+ s2), s0'; rewrite !joinA. 
case; case=>s0 [s12][s0'][->->{s s'}] G C C'.
have C12 : s12 \In star_coh c1 c2.
- by rewrite -(star_coh_prec C); case/guar_coh: G.
split; try by rewrite -starA.
case: C12 C C'=>_ [s1][s2][->{s12}] D D' C C'.
exists (s0 \+ s1), s2, (s0' \+ s1).
rewrite !joinA in C C' *; split=>//.
split; first by exists s0, s1, s0'. 
- by split; [move/cohH/validHL: C | exists s0, s1; case/guar_coh: G].
by split; [move/cohH/validHL: C' | exists s0', s1; case/guar_coh: G].
Qed.

(********************)
(* Transition Union *)
(********************)

Module UnionGuar.
Section UnionGuar.
Variable coh : cohpred.

Definition union_rel (g1 g2 : reltp) s s' := g1 s s' \/ g2 s s'.

Section Base.
Variables g1 g2 : guar coh.

Lemma uguar_coh s s' : union_rel g1 g2 s s' -> s \In coh /\ s' \In coh.
Proof. by case; move/guar_coh. Qed.

Lemma uguar_env s s' : union_rel g1 g2 s s' -> env s = env s'.
Proof. by case; move/guar_env. Qed.

Lemma uguar_zagzig s s' p : 
        env s = env s' -> 
        union_rel g1 g2 (s \> p) (s' \> p) -> union_rel g1 g2 (s \< p) (s' \< p).
Proof. by move=>H; case; move/(guar_zagzig H); [left | right]. Qed.

End Base.

Lemma uintguar_footH (g1 g2 : int_guar coh) s s' :
        union_rel g1 g2 s s' -> footeqH (flats s) (flats s').
Proof. by case; move/int_guar_footH. Qed.

(* No need for now for unioning other classes of guars, so I omit them *)

End UnionGuar.

Module Exports.

Notation union_rel := union_rel.

Section Exports.
Variable coh : cohpred. 

Definition UnionGuarMixin g1 g2 := 
  GuarMixin (@uguar_coh coh g1 g2) (@uguar_env coh g1 g2)
            (@uguar_zagzig coh g1 g2).
Canonical union_guar (g1 g2 : guar coh) := 
  Eval hnf in Guar (union_rel g1 g2) (@UnionGuarMixin g1 g2).

Definition UnionIntGuarMixin g1 g2 := 
  IntGuarMixin (@uintguar_footH coh g1 g2).
Canonical UnionIntGuar (g1 g2 : int_guar coh) := 
  Eval hnf in IntGuar (union_guar g1 g2) (@UnionIntGuarMixin g1 g2).

(* A union is reflexive if all components are internal and at least
one is reflexive. Here's a little program to search for the reflexive
component automatically in a union.

We first search left, then right. The problem is solved in a manner
similar to noalias in the "Less Ad Hoc" paper. We match against
union_rel g1 g2 when first resolving the canonical structures. Thus,
we have to start the work by projecting either into g2 or into
g2. Thus, we need to modify the type of g2 so that it picks g1 up,
similar to x and y in noalias. Then we double-tag the constructor out
of that projection, to first try left, then right.  *)

(* the structures for searching for reflexivity proofs *)

Structure tag (g1 : int_guar coh) := Tag {untag :> int_guar coh}.
Definition right_tag g1 := Tag g1 g1.
Canonical left_tag := right_tag.

Structure search_mixin (g1 g2 : int_guar coh) := SearchMixin {
   g2_of : tag g2;
   _ : ReflexiveGuar.mixin_of (union_guar g1 g2_of)}.

Definition refl_mixin g1 g2 (m : search_mixin g1 g2) : 
  ReflexiveGuar.mixin_of (union_guar g1 (g2_of m)) :=
  let: SearchMixin _ mx := m in mx. 

Lemma left_pf (g1 : refl_guar coh) g2 s : s \In coh -> union_rel g1 g2 s s. 
Proof. by move=>C; left; apply: refl_guar_refl C. Qed.

Lemma right_pf g1 (g2 : refl_guar coh) s : s \In coh -> union_rel g1 g2 s s.
Proof. by move=>C; right; apply: refl_guar_refl C. Qed.

Canonical UnionReflGuar g1 g2 (m : search_mixin g1 g2) :=
  ReflGuar (union_guar g1 (untag (g2_of m))) (refl_mixin m).

Canonical UnionReflGuarMixinLeft (g1 : refl_guar coh) g2 := 
  @SearchMixin g1 g2 (left_tag g2) (ReflGuarMixin (left_pf g1 g2)).

Canonical UnionReflGuarMixinRight g1 (g2 : refl_guar coh) := 
  @SearchMixin g1 g2 (right_tag g2) (ReflGuarMixin (right_pf g1 g2)). 

(* extra lemmas *)

Lemma union_relC (g1 g2 : reltp) : union_rel g1 g2 = union_rel g2 g1. 
Proof.
apply: fext=>s; apply: fext=>s'; apply: pext.
by split; (case; [right | left]).
Qed.

Lemma union_guarC (g1 g2 : guar coh) : union_guar g1 g2 = union_guar g2 g1. 
Proof. by apply/guarE; apply: union_relC. Qed.

Lemma union_relCA g1 g2 g3 : 
        union_rel g1 (union_rel g2 g3) = 
        union_rel g2 (union_rel g1 g3). 
Proof.
apply: fext=>s; apply: fext=>s'; apply: pext; split; 
by case; [right; left | case; [left | right; right]].
Qed.

Lemma union_guarCA g1 g2 g3 : 
        union_guar g1 (union_guar g2 g3) = 
        union_guar g2 (union_guar g1 g3).
Proof. by apply/guarE; apply: union_relCA. Qed.
  
End Exports.
End Exports.
End UnionGuar.

Export UnionGuar.Exports.

(* this not needed anymore *)
Lemma union_relE (g1 g1' : reltp) (g2 g2' : reltp) :
        g1 = g2 -> g1' = g2' ->
        union_rel g1 g1' = union_rel g2 g2'.
Proof. by move=>->->. Qed.

(********************************************)
(* Converting the coherence of a transition *)
(********************************************)

Module DoGuar.
Section DoGuar.
Variables (coh1 coh2 : cohpred) (pf : coh1 = coh2).

Definition do_rel (g : reltp) s s' := g s s'.

Section Base.
Variable g : guar coh1.

Lemma doguar_coh s s' : do_rel g s s' -> s \In coh2 /\ s' \In coh2.
Proof. by case/guar_coh; rewrite pf. Qed.

End Base.

(* conversion for refinements *)

Lemma dointguar_footH (g : int_guar coh1) s s' : 
        do_rel g s s' -> footeqH (flats s) (flats s').
Proof. by apply: int_guar_footH. Qed.

Lemma doreflguar_refl (g : refl_guar coh1) s : 
        s \In coh2 -> do_rel g s s.
Proof. by rewrite -pf; apply: refl_guar_refl. Qed.

Lemma doacqguar_footH (g : acq_guar coh1) h s s' :
        do_rel (g h) s s' -> footeqH (flats s \+ h) (flats s').
Proof. by apply: acq_guar_footH. Qed.

Lemma dorelguar_footH (g : rel_guar coh1) h s s' : 
        do_rel (g h) s s' -> footeqH (flats s) (h \+ flats s').
Proof. by apply: rel_guar_footH. Qed.

End DoGuar.

Module Exports.
Notation do_rel := do_rel.

Section Exports.
Variables (coh1 coh2 : cohpred) (pf : coh1 = coh2).

Definition DoGuarMixin g :=
  GuarMixin (@doguar_coh coh1 coh2 pf g) (@guar_env coh1 g)
            (@guar_zagzig coh1 g).
Canonical do_guar (g : guar coh1) := Eval hnf in Guar (do_rel g) (DoGuarMixin g).

Definition DoIntGuarMixin (g : int_guar coh1) :
  InternalGuar.mixin_of (do_guar g) := 
  IntGuarMixin (@dointguar_footH coh1 g).
Canonical DoIntGuar (g : int_guar coh1) := 
  Eval hnf in IntGuar (do_guar g) (DoIntGuarMixin g).

Definition DoReflGuarMixin (g : refl_guar coh1) : 
  ReflexiveGuar.mixin_of (do_guar g) := 
  ReflGuarMixin (@doreflguar_refl coh1 coh2 pf g).
Canonical DoReflGuar (g : refl_guar coh1) := 
  Eval hnf in ReflGuar (do_guar g) (DoReflGuarMixin g).

Definition baseDoAcqGuarMixin (g : acq_guar coh1) h := DoGuarMixin (g h).
Definition acq_do_guar (g : acq_guar coh1) h :=
  Eval hnf in Guar (do_rel (g h)) (baseDoAcqGuarMixin g h).

Definition DoAcqGuarMixin (g : acq_guar coh1) :
  AcquireGuar.mixin_of (fun h => do_guar (g h)) :=
  AcqGuarMixin (@doacqguar_footH coh1 g).
Canonical DoAcqGuar g := 
  Eval hnf in AcqGuar (acq_do_guar g) (DoAcqGuarMixin g).

Definition baseDoRelGuarMixin (g : rel_guar coh1) h := DoGuarMixin (g h).
Definition rel_do_guar (g : rel_guar coh1) h :=
  Eval hnf in Guar (do_rel (g h)) (baseDoRelGuarMixin g h).

Definition DoRelGuarMixin (g : rel_guar coh1) :
  ReleaseGuar.mixin_of (fun h => do_guar (g h)) := 
  RelGuarMixin (@dorelguar_footH coh1 g).
Canonical DoRelGuar g := 
  Eval hnf in RelGuar (rel_do_guar g) (DoRelGuarMixin g).

End Exports.
End Exports.
End DoGuar.

Export DoGuar.Exports.

Lemma do_relE (g : reltp) : do_rel g = g.
Proof. by []. Qed.

(* entangling communicating transitions *)

Module AcqRelGuar.
Section AcqRelGuar.
Variables (coh1 coh2 : cohpred).

Notation coh := (star_coh coh1 coh2).

Definition acqrel_rel (g1 g2 : heap -> reltp) s s' :=
  (exists s1 s2 s1' s2', 
     [/\ s = s1 \+ s2, s' = s1' \+ s2' &
         exists h, g1 h s1 s1' /\ g2 h s2 s2']) /\
   s \In coh.

Section Base.
Variables (g1 : acq_guar coh1) (g2 : rel_guar coh2).

Lemma arguar_coh s s' : 
        acqrel_rel [eta g1] [eta g2] s s' -> s \In coh /\ s' \In coh.
Proof.
case; case=>s1 [s2][s1'][s2'][->->{s s'}][h][G1 G2] C; split=>//.
case: C=>Vh _; split. 
- by rewrite -(guarH_validUnLR G1 G2).
by exists s1', s2'; case/guar_coh: G1; case/guar_coh: G2. 
Qed.

Lemma arguar_env s s' : acqrel_rel [eta g1] [eta g2] s s' -> env s = env s'.
Proof.
case; case=>s1 [s2][s1'][s2'][->->{s s'}][h][] /=. 
by do 2![move/guar_env=>->]. 
Qed.

Lemma arguar_zagzig s s' p : 
        env s = env s' -> 
        acqrel_rel [eta g1] [eta g2] (s \> p) (s' \> p) -> 
        acqrel_rel [eta g1] [eta g2] (s \< p) (s' \< p).
Proof.
move=>H G; case: (arguar_coh G)=>C C'; split; last by apply/coh_zigzag.
case: G; case=>s1 [s2][s1'][s2'][E E'][h][G1 G2] _. 
case: (guar_coh G1)=>C1 C1'.
case: (envzag_plusf C C' C1 C1' H E E')=>t1' [t2'][r1'][r2'][pt][pr][]. 
case=>??? [????][Et Er] F F'; subst s s' p s1 s2 s1' s2'. 
exists (t1' \< pt), (r1' \< pr), (t2' \< pt), (r2' \< pr); split=>//.
by exists h; split; apply: guar_zagzig.
Qed.

Lemma arguar_footH s s' : 
        acqrel_rel [eta g1] [eta g2] s s' -> footeqH (flats s) (flats s').
Proof.
case; case=>s1 [s1'][s2][s2'][->->{s s'}][h][H H'] /cohV V.
have V' : valid (s2 \+ s2') by rewrite -(guarV_validUn H H').
rewrite !flatsUn //. 
apply: footeqH_trans (footeqHUnL (acq_guar_footH H)).
apply: footeqH_trans (footeqHUnR (rel_guar_footH H')) _.
by rewrite joinA. 
Qed.

End Base.
End AcqRelGuar.

Module Exports.

Notation acqrel_rel := acqrel_rel.

Definition AcqRelGuarMixin c1 c2 g1 g2 := 
  GuarMixin (@arguar_coh c1 c2 g1 g2) (@arguar_env c1 c2 g1 g2)
            (@arguar_zagzig c1 c2 g1 g2).
Canonical acqrel_guar c1 c2 (g1 : acq_guar c1) (g2 : rel_guar c2) := 
  Eval hnf in Guar (acqrel_rel c1 c2 [eta g1] [eta g2]) 
                   (@AcqRelGuarMixin c1 c2 g1 g2).

Definition AcqRelIntGuarMixin coh1 coh2 g1 g2 := 
  IntGuarMixin (@arguar_footH coh1 coh2 g1 g2).
Canonical AcqRelIntGuar coh1 coh2 g1 g2 :=
  Eval hnf in IntGuar (acqrel_guar g1 g2) (@AcqRelIntGuarMixin coh1 coh2 g1 g2).

End Exports.
End AcqRelGuar.

Export AcqRelGuar.Exports.

(* this one not needed *)
Lemma acqrel_relE c1 c1' c2 c2' (g1 g1' g2 g2' : heap -> reltp) :
        c1 = c2 -> c1' = c2' ->
        (forall h, g1 h = g2 h) ->
        (forall h, g1' h = g2' h) ->
        acqrel_rel c1 c1' g1 g1' = acqrel_rel c2 c2' g2 g2'.
Proof.
move=>Ec Ec' Eg Eg'; apply: fext=>s; apply: fext=>s'; apply: pext; split.
- case; case=>s1 [s2][s1'][s2'][->->][h][].
  rewrite Eg Eg' Ec Ec' => G G' C.
  by split=>//; exists s1, s2, s1', s2'; split=>//; exists h. 
case; case=>s1 [s2][s1'][s2'][->->][h][].
rewrite -Eg -Eg' -Ec -Ec' => G G' C.
by split=>//; exists s1, s2, s1', s2'; split=>//; exists h. 
Qed.


(***********)
(***********)
(* Modules *)
(***********)
(***********)

(* channel is a pair of an acquire and release transition *)
(* packaging the dependency on a coherence predicate *)

Module Chan.

Structure CORE (coh : cohpred) := Core {
  acq : acq_guar coh; 
  rel : rel_guar coh}.

Structure CHAN := Chan {
  coh : cohpred; 
  core : CORE coh}.

Module Exports.
Notation chan_core := Chan.CORE.
Notation chan := Chan.CHAN.
Notation Chan := Chan.Chan.
Notation ChanCore := Chan.Core.
Prenex Implicits Chan.Chan.
Coercion Chan.core : chan >-> chan_core.

Structure trans := acq | rel.

Definition get c t (ch : chan_core c) h := 
  if t is acq then Chan.acq ch h else Chan.rel ch h.

End Exports.

End Chan.

Export Chan.Exports.


Module ChanSeq.

Definition axiom coh (s : seq chan) := 
  forall ch, ch \In s -> Chan.coh ch = coh.

Structure CHAN_SEQ := ChanSeq {
  coh : cohpred;
  chan_seq : seq chan;
  _ : axiom coh chan_seq}.

(*
Lemma cores_hd coh ch chs : 
        ChanSeq.axiom coh (ch :: chs) -> Chan.coh ch = coh.
Proof. by apply; left. Qed.

Lemma cores_tail coh ch chs : 
        ChanSeq.axiom coh (ch :: chs) -> ChanSeq.axiom coh chs.
Proof. by move=>Ax x H; apply: Ax; right. Qed.

(* this is a somewhat useless definition, as the dependencies *)
(* mask the fact that it's a map *)
(* so if i want to use it, i need to redo the map theory for it *)
(* which is excessive *)
Fixpoint cores coh chs : ChanSeq.axiom coh chs -> seq (chan_core coh) :=
  if chs is ch :: chs' then 
    fun pf => icoerce _ (Chan.core ch) (cores_hd pf) :: 
                        (cores (cores_tail pf))
  else fun _ => [::].
*)

Module Exports.
Notation chan_seq := ChanSeq.CHAN_SEQ.
Notation ChanSeq := ChanSeq.ChanSeq.
Coercion ChanSeq.chan_seq : chan_seq >-> seq.

Definition chan_seq_in chans ch := ch \In ChanSeq.chan_seq chans.
Canonical chanseq_PredType := mkPredType chan_seq_in.

Lemma chan_seqP chs : axiom (coh chs) chs.
Proof. by case: chs. Qed.

Lemma chan_seqE ch (chs : chan_seq) : ch \In chs -> Chan.coh ch = coh chs.
Proof. by case: chs=>c chs; apply. Qed.

(*
(* just one lemma from the map theory, to see how things work *)
(* the problematic one will be the one for ++, so i think i should drop this *)
Lemma mem_core_inv coh (chs : chan_seq) (pf : ChanSeq.axiom coh chs) x :
        x \In cores pf -> exists y : chan, Chan x = y /\ y \In chs.
Proof.
case: chs pf x=>c'; elim=>[|ch chs IH] pf pf' c //=. 
rewrite InE /=.
case.
- move=>->; exists (Chan (icoerce chan_core ch (cores_hd pf'))).
  split=>//; left; move: (pf' ch (or_introl (erefl _)))=>E. 
  rewrite -{coh c IH}E in pf pf' *; rewrite {pf pf'}ieqc. 
  by case: ch.
move/IH. 
have pf'' : axiom c' chs by move=>ch' H; apply: pf ch' (or_intror H). 
by case/(_ pf'')=>y [<-] H; exists (Chan c); split=>//; right.  
Qed.
*)

Lemma core2ch_coh coh (chs : seq (chan_core coh)) : axiom coh (map Chan chs).
Proof. by elim: chs=>[|ch chs IH] // ct; rewrite InE; case=>[->|/IH]. Qed.

(* pacakage up cores into chans *)
Definition core2ch coh cores := ChanSeq (@core2ch_coh coh cores).

(* In one direction, this lemma is just Mem_map_inv, except the latter is *)
(* more general. In the other direction it's Mem_map, but Mem_map first   *)
(* requires the retyping of the goal, so it's less general  *)

Lemma mem_chan c1 (cores : seq (chan_core c1)) (ch : chan) : 
        ch \In core2ch cores <-> exists ct, ch = Chan ct /\ ct \In cores.
Proof.
split; first by apply: Mem_map_inv.
by case=>ct' [->]; apply: Mem_map.
Qed.

(* special case of mem_chan, when c1 = c2 *)

Lemma mem_chan_eq coh (cores : seq (chan_core coh)) ct : 
        Chan ct \In core2ch cores <-> ct \In cores.
Proof.
elim: cores ct=>[|ct' cores IH] //= ct.
rewrite InE /=; split; last by case=>[<-|/IH]; [left | right].
case/Mem_map_inv=>_ [[/inj_pair2]] <-. 
by rewrite InE; case=>[<-|]; [left | right].
Qed.

End Exports.
End ChanSeq.

Export ChanSeq.Exports.


Module Mod.

(*
We separate the core of a module, from its communication component.
The operation on the core are commutative. Once the communication is
added it's debatable how to best define entanglement.
*)

Structure CORE := Core {
  coh : cohpred; 
  internal : refl_guar coh}.

Structure MOD := Make {
  core : CORE;
  cores : seq (chan_core (coh core))}.

Definition chans W : chan_seq := core2ch (Mod.cores W).

Module Exports.
Notation mod_core := Mod.CORE.
Notation mod := Mod.MOD.
Coercion Mod.core : mod >-> mod_core.

(* useless definition because it uses cores
Definition mkMod core chs pf := Make (@ChanSeq.cores (Mod.coh core) chs pf).
*)

End Exports.
End Mod.

Export Mod.Exports.


Lemma coreE (m1 m2 : mod_core) :
        m1 = m2 <-> 
        [/\ Mod.coh m1 = Mod.coh m2 &
            Mod.internal m1 = Mod.internal m2 :> reltp].
Proof.
case: m1 m2=>c1 i1 [c2 i2] /=; split;
case=>E; rewrite -{c2}E in i2 *.
- by move/inj_pair2=>->.
by move=>Ei; congr Mod.Core; apply/refl_guarE.
Qed.

Lemma chan_coreE c (cc1 cc2 : chan_core c) : 
        cc1 = cc2 <->
        [/\ forall h, Chan.acq cc1 h = Chan.acq cc2 h :> reltp &
            forall h, Chan.rel cc1 h = Chan.rel cc2 h :> reltp].
Proof.
split=>[->//|]; case: cc1 cc2=>a1 r1 [a2 r2][/= H1 H2].
congr ChanCore. 
- by apply/acq_guarE; apply/fext=>h; apply/guarE.
by apply/rel_guarE; apply/fext=>h; apply/guarE.
Qed.

Lemma chanE (ch1 ch2 : chan) : 
        ch1 = ch2 <-> 
        [/\ Chan.coh ch1 = Chan.coh ch2,
            forall h, Chan.acq ch1 h = Chan.acq ch2 h :> reltp &
            forall h, Chan.rel ch1 h = Chan.rel ch2 h :> reltp].
Proof.
split=>[->//|]; case: ch1 ch2=>c1 [a1 r1][c2 [a2 r2]] /= [E H1 H2].
by rewrite -{c2}E in a2 r2 H1 H2 *; congr Chan.Chan; apply/chan_coreE. 
Qed.

Lemma modE (m1 m2 : mod) : 
        m1 = m2 <-> 
        Mod.core m1 = Mod.core m2 /\ Mod.chans m1 = Mod.chans m2 :> seq _.
Proof.
move: m1 m2=>[cor1 com1][cor2 com2] /=; split;
case=>E; rewrite -{cor2}E in com2 *.
- by move/inj_pair2=>->. 
elim: com1 com2=>[|ct1 com1 IH1]; case=>//= ct2 com2.
case; move/inj_pair2=>?; subst ct2. 
by case/IH1; move/inj_pair2=>->. 
Qed.


(*************************************************************************)
(* Composition of the internal and communicating transitions of a module *)
(*************************************************************************)

Section MGuar.
Variable W : mod. 
Notation coh := (Mod.coh W).

Definition mguar s s' := 
  Mod.internal W s s' \/ 
  exists ch h, ch \In Mod.chans W /\ 
    (Chan.acq ch h s s' \/ Chan.rel ch h s s').

Lemma mguar_footV s1 s2 : mguar s1 s2 -> footeqV s1 s2. 
Proof. by case=>[|[ch][h][H][]]; move/guar_footV. Qed.

Lemma mguar_refl s : s \In coh -> mguar s s.
Proof. by move=>C; left; apply: refl_guar_refl. Qed.

Lemma mguar_coh s s' : mguar s s' -> s \In coh /\ s' \In coh.
Proof. by case=>[|[_][h][/Mem_map_inv][ch][->] H []]; case/guar_coh. Qed.

Lemma mguar_env s s' : mguar s s' -> env s = env s'.
Proof. by case=>[|[_][h][/Mem_map_inv][ch][->] H []]; move/guar_env. Qed.

Lemma mguar_zageq s1 s2 p :
        mguar (s1 \> p) s2 -> exists s, s2 = s \> p /\ env s1 = env s. 
Proof. 
move/mguar_env=>E; exists [loc s2, shr s2, env s1]. 
by rewrite -[s2]st_eta /= -E zagE spP. 
Qed.

Lemma mguar_zagzig s s' p : 
        env s = env s' -> 
        mguar (s \> p) (s' \> p) -> mguar (s \< p) (s' \< p).
Proof.
rewrite -[s]st_eta -[s']st_eta /= => E.
case=>H; [left | right]; first by apply: guar_zagzig.
case: H=>ch [h][Ch]; case=>H; exists ch, h; split=>//; [left | right]; 
by apply: guar_zagzig.
Qed.

End MGuar.


(**************************************************)
(* Empty Module                                   *)
(* serves as a unit in iterated module operations *)
(**************************************************)

Definition empty_mod : mod := 
  @Mod.Make (Mod.Core [refl_guar of id_guar emp_coh]) [::].

(************************)
(* Entanglement modules *)
(************************)

(* entanglement of two channels *)

Definition ent_guar c1 c2 (ch1 : chan_core c1) (ch2 : chan_core c2) :=
   [int_guar of union_guar (acqrel_guar (Chan.acq ch1) (Chan.rel ch2))
     (do_guar (starC _ _) (acqrel_guar (Chan.acq ch2) (Chan.rel ch1)))].

Lemma ent_guarC c1 c2 (ch1 : chan_core c1) (ch2 : chan_core c2) :
        ent_guar ch1 ch2 = ent_guar ch2 ch1 :> reltp.
Proof.
apply: fext=>s; apply: fext=>s'; apply: pext; split; 
by case; [right | left].
Qed.

(* the union of all communicating entanglements *)

Module GuarSum.
Section GuarSum.
Variables (chs chs' : chan_seq).

Notation coh := (star_coh (ChanSeq.coh chs) (ChanSeq.coh chs')).

Definition guarsum_rel s s' := 
  exists ch ch', [/\ ch \In chs, ch' \In chs' & 
                     ent_guar ch ch' s s'].

Lemma guarsum_coh s s' : guarsum_rel s s' -> s \In coh /\ s' \In coh.
Proof. 
by case=>ct [ct'][H1 H2 /guar_coh] /=; rewrite (chan_seqE H1) (chan_seqE H2).
Qed.

Lemma guarsum_env s s' : guarsum_rel s s' -> env s = env s'.
Proof. by case=>ct [ct'][_ _ /guar_env]. Qed.

Lemma guarsum_zagzig s s' p : 
        env s = env s' -> 
        guarsum_rel (s \> p) (s' \> p) -> guarsum_rel (s \< p) (s' \< p).
Proof. by move=>E [ct][ct'][H1 H2 /(guar_zagzig E)]; exists ct, ct'. Qed.

Lemma guarsum_footH s s' : guarsum_rel s s' -> footeqH (flats s) (flats s').
Proof. by case=>ct [ct'][_ _ /int_guar_footH]. Qed.

End GuarSum.

Module Exports.
Notation ent_guar := ent_guar.
Notation guarsum_rel := guarsum_rel.

Section Base.
Variables (chs chs' : chan_seq). 

Definition GuarSumGuarMixin := 
  GuarMixin (@guarsum_coh chs chs') (@guarsum_env chs chs')
            (@guarsum_zagzig chs chs'). 
Canonical guar_sum := Eval hnf in Guar (guarsum_rel chs chs') GuarSumGuarMixin.

Definition IntGuarSumGuarMixin := IntGuarMixin (@guarsum_footH chs chs').
Canonical IntGuarSum := Eval hnf in IntGuar guar_sum IntGuarSumGuarMixin.

End Base.

Lemma guar_sumC chs chs' : guar_sum chs chs' = guar_sum chs' chs :> reltp. 
Proof.
apply: fext=>s; apply: fext=>s'; apply: pext; split;
by case=>ct [ct'][H1 H2 G]; exists ct', ct; split=>//; rewrite ent_guarC.
Qed.

Lemma guarsum_congr chs1 chs1' chs2 chs2' : 
        chs1 =p chs2 -> chs1' =p chs2' ->
        guar_sum chs1 chs1' = guar_sum chs2 chs2' :> reltp.
Proof.
move=>E E'; apply: fext=>s; apply: fext=>s'; apply: pext; split;
case=>ch [ch'][H H' G]; exists ch, ch';
by move/E: H=>H; move/E': H'=>H'. 
Qed.

End Exports.

End GuarSum.

Export GuarSum.Exports.



(* There're many ways to connect the communication transitions of two *)
(* modules. But two stick out as limit points: connecting all the     *)
(* transition of the two modules pairwise, and connecting none.  The  *)
(* first combination we call entangled, the second one apartness      *)
(* composition. *)

Section EntangleCore.
Variables W1 W2 : mod.

Notation coh := (star_coh (Mod.coh W1) (Mod.coh W2)).

Let int1 := frame_guar (Mod.coh W2) (Mod.internal W1).
Let int2 := frame_guar (Mod.coh W1) (Mod.internal W2).

(* internal transition *)

Definition entangle_intguar := 
  union_guar (union_guar int1 (do_guar (starC _ _) int2))
             [int_guar of guar_sum (Mod.chans W1) (Mod.chans W2)].

Definition apart_intguar := 
             union_guar int1 (do_guar (starC _ _) int2).

End EntangleCore.

Canonical entangle_core W1 W2 : mod_core := 
  Mod.Core [refl_guar of entangle_intguar W1 W2].

Canonical apart_core W1 W2 : mod_core := 
  Mod.Core [refl_guar of apart_intguar W1 W2].

Lemma entangle_intC W1 W2 : 
        entangle_intguar W1 W2 = 
        entangle_intguar W2 W1 :> reltp.
Proof. 
apply: union_relE; first by apply: union_relC.
by apply: guar_sumC.
Qed.

Lemma apart_intC W1 W2 : 
        apart_intguar W1 W2 = apart_intguar W2 W1 :> reltp.
Proof. by apply: union_relC. Qed.

Lemma entangle_coreC W1 W2 : entangle_core W1 W2 = entangle_core W2 W1.
Proof. by apply/coreE; split; [rewrite /= starC | apply: entangle_intC]. Qed.

Lemma apart_coreC W1 W2 : apart_core W1 W2 = apart_core W2 W1.
Proof. by apply/coreE; split; [rewrite /= starC | apply: apart_intC]. Qed.

Lemma entangle_int_congr (W1 W2 W1' W2' : mod) : 
        Mod.core W1 = Mod.core W2 -> 
        Mod.chans W1 =p Mod.chans W2 ->
        Mod.core W1' = Mod.core W2' -> 
        Mod.chans W1' =p Mod.chans W2' ->
        entangle_intguar W1 W1' = entangle_intguar W2 W2' :> reltp.
Proof. 
case/coreE=>Ecoh Ei Ech /= /coreE [Ecoh' Ei'] Ech'. 
by rewrite Ei Ecoh Ei' Ecoh'; apply: union_relE=>//; apply: guarsum_congr.
Qed.

Lemma apart_int_congr (W1 W2 W1' W2' : mod) : 
        Mod.core W1 = Mod.core W2 -> 
        Mod.chans W1 =p Mod.chans W2 ->
        Mod.core W1' = Mod.core W2' -> 
        Mod.chans W1' =p Mod.chans W2' ->
        apart_intguar W1 W1' = apart_intguar W2 W2' :> reltp.
Proof. 
case/coreE=>Ecoh Ei Ech /= /coreE [Ecoh' Ei'] Ech'. 
by rewrite Ei Ecoh Ei' Ecoh'; apply: union_relE. 
Qed.



(* Now that we've computed the core of the entangle module. Let's see    *)
(* which communication transitions to pick up. We can again go for many  *)
(* combinations, but again two stick out and they stick out in exactly   *)
(* oposite situations in terms of core entanglement. For entangled       *)
(* combination, the transitions of the left module are propagated to the *)
(* combination. For apart combination, all transitions are propagated.   *)

Definition frame_coreL c1 c2 (ch : chan_core c1) := 
  ChanCore [acq_guar of acq_frame_guar c2 (Chan.acq ch)]
           [rel_guar of rel_frame_guar c2 (Chan.rel ch)].

(*
Definition frame_chanL c2 (ch : chan) := Chan (frame_coreL c2 (Chan.core ch)).
*)

Definition frame_coreR c2 c1 (ch : chan_core c2) :=
  ChanCore (DoAcqGuar (starC _ _) [acq_guar of acq_frame_guar c1 (Chan.acq ch)])
           (DoRelGuar (starC _ _) [rel_guar of rel_frame_guar c1 (Chan.rel ch)]). 

(*
Definition frame_chanR c1 (ch : chan) := Chan (frame_coreR c1 (Chan.core ch)). 
*)

Section EntangleComm.
Variables (W1 W2 : mod). 

Notation coh := (star_coh (Mod.coh W1) (Mod.coh W2)).

Definition entangle_lefts := map (frame_coreL (Mod.coh W2)) (Mod.cores W1).
Definition entangle_rights := map (frame_coreR (Mod.coh W1)) (Mod.cores W2).

(*
Lemma entangleAx : ChanSeq.axiom coh entangle_lefts.
Proof. by move=>ch /Mem_map_inv [c1][->] /Mem_map_inv [c2][->]. Qed.

Lemma apartAx : ChanSeq.axiom coh (entangle_lefts ++ entangle_rights).
Proof.
move=>ch /Mem_cat; case; first by apply: entangleAx.
by case/Mem_map_inv=>c1 [->] /Mem_map_inv [c2][->]; rewrite /= starC.
Qed.

Definition entangle : mod := mkMod entangleAx.
Definition apart : mod := mkMod apartAx.
*)

Definition entangle : mod := @Mod.Make (entangle_core W1 W2) entangle_lefts. 
Definition apart : mod := 
  @Mod.Make (apart_core W1 W2) (entangle_lefts ++ entangle_rights).

End EntangleComm.


Lemma entangle_chan_congr (W1 W2 W1' W2' : mod) :
        Mod.coh W1 = Mod.coh W2 ->
        Mod.chans W1 =p Mod.chans W2 ->
        Mod.coh W1' = Mod.coh W2' ->
        Mod.chans W1' =p Mod.chans W2' ->
        Mod.chans (entangle W1 W1') =p Mod.chans (entangle W2 W2').
Proof.
suff {W1 W2 W1' W2'} L (W1 W2 W1' W2' : mod) : 
  Mod.coh W1 = Mod.coh W2 -> Mod.chans W1 =p Mod.chans W2 ->
  Mod.coh W1' = Mod.coh W2' -> Mod.chans W1' =p Mod.chans W2' ->
  Mod.chans (entangle W1 W1') <=p Mod.chans (entangle W2 W2').
- by move=>Ec Ech Ec' Ech' x; split; apply: L=>//; rewrite ?Ech ?Ech'.
move=>Ec Ech Ec' Ech' _ /Mem_map_inv [_][->] /Mem_map_inv [x][->] 
/mem_chan_eq; rewrite ?Ech ?Ech'; case/Mem_map_inv=>y [E] T; apply/mem_chan.
exists (frame_coreL (Mod.coh W2') y).
split; last by apply/Mem_map. 
apply/chanE=>/=; split; first by [rewrite Ec Ec'];
by case/chanE: E T=>/= _ E1 E2 T h; rewrite ?E1 ?E2 Ec Ec'.  
Qed.

Lemma apart_chan_congr (W1 W2 W1' W2' : mod) :
        Mod.coh W1 = Mod.coh W2 ->
        Mod.chans W1 =p Mod.chans W2 ->
        Mod.coh W1' = Mod.coh W2' ->
        Mod.chans W1' =p Mod.chans W2' ->
        Mod.chans (apart W1 W1') =p Mod.chans (apart W2 W2').
Proof.
suff {W1 W2 W1' W2'} L (W1 W2 W1' W2' : mod) : 
  Mod.coh W1 = Mod.coh W2 -> Mod.chans W1 =p Mod.chans W2 ->
  Mod.coh W1' = Mod.coh W2' -> Mod.chans W1' =p Mod.chans W2' ->
  Mod.chans (apart W1 W1') <=p Mod.chans (apart W2 W2').
- by move=>Ec Ech Ec' Ech' x; split; apply: L=>//; rewrite ?Ech ?Ech'.
move=>Ec Ech Ec' Ech' _ /Mem_map_inv [y][->] /Mem_cat; case;
case/Mem_map_inv=>z [->{y}] /mem_chan_eq; rewrite ?Ech ?Ech';
case/Mem_map_inv=>w [E] T; apply/mem_chan.
- exists (frame_coreL (Mod.coh W2') w).
  split; last by apply/Mem_cat; left; apply/Mem_map. 
  apply/chanE=>/=; split; first by [rewrite Ec Ec'];
  by case/chanE: E T=>/= _ E1 E2 T h; rewrite ?E1 ?E2 Ec Ec'. 
exists (frame_coreR (Mod.coh W2) w).
split; last by apply/Mem_cat; right; apply/Mem_map. 
apply/chanE=>/=; split; first by [rewrite Ec Ec'];
by case/chanE: E T=>/= _ E1 E2 T h; rewrite ?E1 ?E2 Ec Ec'.
Qed.


Lemma apart_chanC W1 W2 : Mod.chans (apart W1 W2) =p Mod.chans (apart W2 W1).
Proof.
suff {W1 W2} L W1 W2 : Mod.chans (apart W1 W2) <=p Mod.chans (apart W2 W1).
- by move=>x; split; move/L.
move=>x; case/Mem_map_inv=>y [->{x}] /Mem_cat; case.
- case/Mem_map_inv=>z [->{y}] H; apply/mem_chan. 
  exists (frame_coreR (Mod.coh W2) z); split; last first.
  - by apply/Mem_cat; right; apply/Mem_map. 
  apply/chanE=>/=; split; first by [rewrite starC];
  by move=>h; case: z H.  
case/Mem_map_inv=>z [->{y}] H.
apply/mem_chan; exists (frame_coreL (Mod.coh W1) z).
split; last by apply/Mem_cat; left; apply/Mem_map. 
by apply/chanE=>/=; split=>//; rewrite starC.
Qed.

Lemma entguar_frameLE c c1 c2 (ch1 : chan_core c1) (ch2 : chan_core c2) : 
         ent_guar (frame_coreL c ch1) ch2 = 
         frame_guar c (ent_guar ch1 ch2) :> reltp.
Proof.
apply: fext=>s; apply: fext=>s'; apply: pext; split;
move=>G; case: (guar_coh G)=>/= C C'.
- case: G C C'. 
  - case; case=>s13 [s2][s13'][s2'][->->{s s'}][h][/= G1 G2] _ C C'.   
    case: ch1 G1=>acq rel /= G1. 
    case: G1 C C'; case=>s1 [s3][s1'][->->{s13 s13'}] G1 _ _ C C'. 
    split; try by rewrite starAC.
    exists (s1 \+ s2), s3, (s1' \+ s2').
    rewrite !(joinAC _ s3) in C *; split=>//.
    left; split; first by exists s1, s2, s1', s2'; split=>//; exists h. 
    split; first by move/cohH/validHL: C.
    by exists s1, s2; case/guar_coh: G1; case/guar_coh: G2. 
  case; case=>s2 [s13][s2'][s13'][->->{s s'}][h][G2 G1] _ C C'.
  case: ch1 G1=>acq rel /= G1.
  case: G1 C C'; case=>s1 [s3][s1'][->->{s13 s13'}] G1 _ _ C C'.
  split; try by rewrite starAC.
  exists (s2 \+ s1), s3, (s2' \+ s1').
  rewrite !joinA in C C' *; split=>//.
  right; split; first by exists s2, s1, s2', s1'; split=>//; exists h. 
  split; first by move/cohH/validHL: C.
  by exists s2, s1; case/guar_coh: G1; case/guar_coh: G2.
case: G C C'; case=>s12 [s3][s12'][->->{s s'}] G _ _ C C'.
have C3 : s3 \In c by rewrite -(star_coh_prec C); case/guar_coh: G. 
case: G=>G; [left | right].
- case: G C C'; case=>s1 [s2][s1'][s2'][??][h][G1 G2] _ C C'; subst s12 s12'.
  case: ch1 G1=>a1 _ /= G1. 
  split; last by rewrite starAC. 
  exists (s1 \+ s3), s2, (s1' \+ s3), s2'.
  rewrite -!(joinAC _ s3) in C C' *; split=>//; exists h; split=>//.
  split; first by exists s1, s3, s1'. 
  - split=>//; first by move/cohH/validHL: C. 
    by exists s1, s3; case/guar_coh: G1. 
  split; first by move/cohH/validHL: C'. 
  by exists s1', s3; case/guar_coh: G1. 
case: G C C'; case=>s2 [s1][s2'][s1'][??][h][G2 G1] _ C C'; subst s12 s12'.
case: ch1 G1=>_ rel /= G1.
split; last by rewrite starCA starA. 
exists s2, (s1 \+ s3), s2', (s1' \+ s3).
rewrite -!joinA in C C' *; split=>//; exists h; split=>//.
split; first by exists s1, s3, s1'. 
- split=>//; first by move/cohH/validHR: C.
  by exists s1, s3; case/guar_coh: G1. 
split=>//; first by move/cohH/validHR: C'. 
by exists s1', s3; case/guar_coh: G1. 
Qed.

Lemma entguar_frameE c c1 c2 (ch1 : chan_core c1) (ch2 : chan_core c2) : 
        ent_guar (frame_coreR c ch1) ch2 = 
        ent_guar (frame_coreL c ch1) ch2 :> reltp.
Proof. by case: ch1=>c' [acq rel]; rewrite /= starC. Qed. 

Lemma entguar_frameRE c c1 c2 (ch1 : chan_core c1) (ch2 : chan_core c2) : 
        ent_guar (frame_coreR c ch1) ch2 = 
        frame_guar c (ent_guar ch1 ch2) :> reltp.
Proof. by rewrite entguar_frameE entguar_frameLE. Qed.

Lemma entangle_intAC W1 W2 W3 : 
        entangle_intguar (entangle W1 W2) W3 = 
        entangle_intguar (entangle W1 W3) W2 :> reltp.
Proof.
suff {W2 W3} L : forall W2 W3 s s',
  entangle_intguar (entangle W1 W2) W3 s s' ->
  entangle_intguar (entangle W1 W3) W2 s s'.
- apply: fext=>s1; apply: fext=>s2.
  by apply: pext; split; move/L. 
move=>W2 W3 s s'; case; last first.  
- case=>_ [_][/Mem_map_inv][_][->] /Mem_map_inv [ct1][->] T1 
        /Mem_map_inv [ct3][->] T3 G.
  case: (guar_coh G)=>/= C C'; left; left; split; try by rewrite /= starAC.
  move: G C C'; rewrite entguar_frameLE.
  case; case=>s13 [s2][s13'][->->{s s'}] G _ _ C C'. 
  exists s13, s2, s13'; split=>//; right. 
  by exists (Chan ct1), (Chan ct3); split=>//; apply/Mem_map.
case; last first.
- case; case=>s3 [s12][s3'][->->{s s'}] G C C'.
  have C12 : s12 \In star_coh (Mod.coh W1) (Mod.coh W2). 
  - by rewrite -(star_coh_prec C); case/guar_coh: G. 
  case: C12 C C'=>H [s1][s2][?] /= C1 _ C C'; subst s12.
  left; left; split; try by rewrite -starA starCA. 
  rewrite !joinA in C C' *.
  exists (s3 \+ s1), s2, (s3' \+ s1); split=>//.
  left; right; split; first by exists s3, s1, s3'. 
  - by split; [move/cohH/validHL: C | exists s3, s1; case/guar_coh: G]. 
  by split; [move/cohH/validHL: C' | exists s3', s1; case/guar_coh: G]. 
case; case=>s12 [s3][s12'][->->{s s'}] G C C'.
have C3 : s3 \In Mod.coh W3.
- by rewrite -(star_coh_prec C); case/guar_coh: G. 
case: G C C'; last first.
- case=>_ [_][/Mem_map_inv][ct][->] H1 /Mem_map_inv [ct'][->] H2 G C C'.
  right; exists (Chan (frame_coreL (Mod.coh W3) ct)), (Chan ct').
  split=>//; try apply: Mem_map=>//; try by apply: Mem_map.
  by rewrite entguar_frameLE; split=>//; exists s12, s3, s12'. 
case; last first. 
- case; case=>s2 [s1][s2'][->->{s12 s12'}] G _ _ C C'.
  left; right; split; try by rewrite starCA starA. 
  by rewrite -!joinA in C C' *; exists s2, (s1 \+ s3), s2'.  
case; case=>s1 [s2][s1'][->->{s12 s12'}] G _ _ C C'.
left; left; split; try by rewrite starAC.
rewrite !(joinAC _ s2) in C C' *.
exists (s1 \+ s3), s2, (s1' \+ s3); split=>//. 
left; left; split; first by exists s1, s3, s1'.
- by split; [move/cohH/validHL: C | exists s1, s3; case/guar_coh: G].
by split; [move/cohH/validHL: C' | exists s1', s3; case/guar_coh: G].
Qed.

Lemma entangle_coreAC W1 W2 W3 : 
        entangle (entangle W1 W2) W3 = 
        entangle (entangle W1 W3) W2 :> mod_core.
Proof. by apply/coreE; split; [rewrite /= starAC | apply: entangle_intAC]. Qed.

Lemma entangle_chanAC W1 W2 W3 : 
        Mod.chans (entangle (entangle W1 W2) W3) =p
        Mod.chans (entangle (entangle W1 W3) W2).
Proof.
suff {W2 W3} L W2 W3 : Mod.chans (entangle (entangle W1 W2) W3) <=p 
                       Mod.chans (entangle (entangle W1 W3) W2).
- by move=>s; split; move/L.
move=>_ /Mem_map_inv [_][->] /Mem_map_inv [_][->] /Mem_map_inv [x][->] H.
apply/mem_chan; exists (frame_coreL (Mod.coh W2) (frame_coreL (Mod.coh W3) x)).
split; last by apply/Mem_map; apply/Mem_map. 
apply/chanE; split=>/=; first by [rewrite starAC];
by move=>h; rewrite !frame_relA starC. 
Qed.

Lemma apart_intAC W1 W2 W3 : 
        apart_intguar (apart W1 W2) W3 = 
        apart_intguar (apart W1 W3) W2 :> reltp.
Proof.
suff {W2 W3} L : forall W2 W3 s s',
  apart_intguar (apart W1 W2) W3 s s' -> apart_intguar (apart W1 W3) W2 s s'.
- by apply: fext=>s1; apply: fext=>s2; apply: pext; split; move/L. 
move=>W2 W3 s s'; case; last first.
- case; case=>s3 [s12][s3'][->->{s s'}] G C C'.
  have C12 : s12 \In star_coh (Mod.coh W1) (Mod.coh W2).
  - by rewrite -(star_coh_prec C); case/guar_coh: G. 
  case: C12 C C'=>_ [s1][s2][->{s12}] C1 C2 C C'.
  left; split; try by rewrite starAC starC. 
  exists (s3 \+ s1), s2, (s3' \+ s1). 
  rewrite !joinA in C C' *; split=>//.
  right; split; first by exists s3, s1, s3'. 
  - by split; [move/cohH/validHL: C | exists s3, s1; case/guar_coh: G].
  by split; [move/cohH/validHL: C' | exists s3', s1; case/guar_coh: G].
case; case=>s12 [s3][s12'][->->{s s'}] G C C'. 
have C3 : s3 \In Mod.coh W3 by rewrite -(star_coh_prec C); case/guar_coh: G. 
case: G C C'; last first.
  case; case=>s2 [s1][s2'][->->{s12 s12'}] G _ _ C C'.
  right; split; try by rewrite starC starAC. 
  by exists s2, (s1 \+ s3), s2'; rewrite -!joinA. 
case; case=>s1 [s2][s1'][->->{s12 s12'}] G _ _ C C'.
left; split; try by rewrite starAC.
exists (s1 \+ s3), s2, (s1' \+ s3). 
rewrite !(joinAC _ s2) in C C' *; split=>//.
left; split; first by exists s1, s3, s1'. 
- by split; [move/cohH/validHL: C | exists s1, s3; case/guar_coh: G]. 
by split; [move/cohH/validHL: C' | exists s1', s3; case/guar_coh: G].
Qed.

Lemma apart_coreAC W1 W2 W3 : 
        apart (apart W1 W2) W3 = 
        apart (apart W1 W3) W2 :> mod_core.
Proof. by apply/coreE; split; [rewrite /= starAC | apply: apart_intAC]. Qed.

Lemma apart_chanAC W1 W2 W3 : 
        Mod.chans (apart (apart W1 W2) W3) =p  
        Mod.chans (apart (apart W1 W3) W2).
Proof.
suff {W2 W3} L W2 W3 : Mod.chans (apart (apart W1 W2) W3) <=p
  Mod.chans (apart (apart W1 W3) W2).
- by move=>x; split; apply: L.
move=>_ /Mem_map_inv [x][->] /Mem_cat; case; last first.
- case/Mem_map_inv=>y [->{x}] T; apply/mem_chan. 
  exists (frame_coreL (Mod.coh W2) (frame_coreR (Mod.coh W1) y)).
  split; last first.
  - by apply/Mem_cat; left; apply/Mem_map; apply/Mem_cat; right; apply/Mem_map. 
  apply/chanE=>/=; split; first by [rewrite starAC];
  by move=>h; rewrite !do_relE -(starC (Mod.coh W3)) frame_relA.
case/Mem_map_inv=>y [->{x}] /Mem_cat; case; case/Mem_map_inv=>z [->{y}] T;
apply/mem_chan. 
- exists (frame_coreL (Mod.coh W2) (frame_coreL (Mod.coh W3) z)).
  split; last by do 2![apply/Mem_cat; left; apply/Mem_map].
  apply/chanE=>/=; split; first by [rewrite starAC];
  by case: z T=>a r /= T h; rewrite !frame_relA starC.
exists (frame_coreR (star_coh (Mod.coh W1) (Mod.coh W3)) z). 
split; last by apply/Mem_cat; right; apply/Mem_map. 
apply/chanE=>/=; split; first by [rewrite starAC];
by move=>h; rewrite !do_relE starC frame_relA. 
Qed.

Lemma apart_intA W1 W2 W3 : 
        apart_intguar (apart W1 W2) W3 = 
        apart_intguar W1 (apart W2 W3) :> reltp.
Proof.
rewrite apart_intAC. 
rewrite (@apart_int_congr (apart _ _) (apart _ _) 
                          W2 W2 (apart_coreC _ _) (apart_chanC _ _)) //. 
rewrite apart_intAC apart_intC.
by rewrite (@apart_int_congr W1 W1 (apart _ _) (apart _ _) 
                             _ _ (apart_coreC _ _) (apart_chanC _ _)).
Qed.

Lemma apart_coreA W1 W2 W3 : 
        apart (apart W1 W2) W3 = 
        apart W1 (apart W2 W3) :> mod_core.
Proof. by apply/coreE; split; [rewrite /= starA | apply: apart_intA]. Qed.

Lemma apart_chanA W1 W2 W3 : 
        Mod.chans (apart (apart W1 W2) W3) =p
        Mod.chans (apart W1 (apart W2 W3)).
Proof.
rewrite apart_chanAC.
rewrite (@apart_chan_congr (apart _ _) (apart _ _) W2 W2 (starC _ _) (apart_chanC _ _) 
         (erefl _) (@EqPredType_refl _ _ _)).
rewrite apart_chanAC apart_chanC.
by rewrite (@apart_chan_congr W1 W1 (apart _ _) (apart _ _) (erefl _) (@EqPredType_refl _ _ _)
           (starC _ _) (apart_chanC _ _)). 
Qed.

Lemma entangle_apartA W1 W2 W3 : 
        entangle_intguar W1 (apart W2 W3) =  
        entangle_intguar (entangle W1 W2) W3 :> reltp.
Proof.
apply: fext=>s; apply: fext=>s'; apply: pext; split.
- case; last first.
  - case=>_ [_][/Mem_map_inv][ct1][-> H1] /Mem_map_inv [ct'][->] /Mem_cat; case.
    - case/Mem_map_inv=>ct2 [->{ct'}] H2.
      rewrite ent_guarC entguar_frameLE.
      case; case=>s12 [s3][s12'][->->{s s'}] G C C'.
      left; left; split; try by rewrite -starA starCA starA. 
      exists s12, s3, s12'; split=>//.
      right; exists (Chan ct1), (Chan ct2).
      by rewrite ent_guarC; split=>//; apply: Mem_map.
    case/Mem_map_inv=>ct3 [->{ct'}] H3.
    rewrite ent_guarC entguar_frameRE.
    case; case=>s13 [s2][s13'][->->{s s'}] G C C'.
    right; exists (Chan (frame_coreL (Mod.coh W2) ct1)), (Chan ct3).
    split=>//; try apply: Mem_map=>//; try by apply: Mem_map.
    rewrite entguar_frameLE.
    split; try by rewrite -starA starCA starA. 
    by exists s13, s2, s13'; rewrite ent_guarC. 
  case.
  - case; case=>s1 [s23][s1'][->->{s s'}] G C C'.
    have C23 : s23 \In star_coh (Mod.coh W2) (Mod.coh W3).
    - by rewrite -(star_coh_prec C); case/guar_coh: G.
    case: C23 C C'=>_ [s2][s3][->{s23}] C2 C3 C C'.
    left; left; split; try by rewrite -starA. 
    exists (s1 \+ s2), s3, (s1' \+ s2).
    rewrite !joinA in C C' *; split=>//.
    left; left; split; first by exists s1, s2, s1'. 
    - by split; [move/cohH/validHL: C | exists s1, s2; case/guar_coh: G]. 
    by split; [move/cohH/validHL: C' | exists s1', s2; case/guar_coh: G].
  case; case=>s23 [s1][s23'][->->{s s'}] G C C'.
  have C1 : s1 \In Mod.coh W1.
  - by rewrite -(star_coh_prec C); case/guar_coh: G. 
  case: G C C'; last first.
  - case; case=>s3 [s2][s3'][->->{s23 s23'}] G _ _ C C'.
    left; right; split; try by rewrite starC -starA starC. 
    by exists s3, (s2 \+ s1), s3'; rewrite -!joinA.
  case; case=>s2 [s3][s2'][->->{s23 s23'}] G _ _ C C'.
  left; left; split; try by rewrite -starA starC. 
  exists (s2 \+ s1), s3, (s2' \+ s1).
  rewrite !(joinAC _ s3) in C C' *; split=>//.
  left; right; split; first by exists s2, s1, s2'. 
  - by split; [move/cohH/validHL: C | exists s2, s1; case/guar_coh: G].
  by split; [move/cohH/validHL: C' | exists s2', s1; case/guar_coh: G].
case; last first.
- case=>_ [_][/Mem_map_inv][_][->] /Mem_map_inv [ct1][->] H1
        /Mem_map_inv [ct3][->] H3; rewrite entguar_frameLE. 
  case; case=>s13 [s2][s13'][->->{s s'}] G C C'; right.
  exists (Chan ct1), (Chan (frame_coreR (Mod.coh W2) ct3)).
  split; first by apply/Mem_map.
  - by apply/Mem_map; apply/Mem_cat; right; apply/Mem_map.
  rewrite ent_guarC entguar_frameRE.
  split; try by rewrite -starA starCA starA.
  by exists s13, s2, s13'; rewrite ent_guarC.
case; last first.
- case; case=>s3 [s12][s3'][->->{s s'}] G C C'. 
  have C12 : s12 \In star_coh (Mod.coh W1) (Mod.coh W2).
  - by rewrite -(star_coh_prec C); case/guar_coh: G.
  case: C12 C C'=>_ [s1][s2][->{s12}] C1 C2 C C'.
  left; right; split; try by rewrite starC starA starC. 
  exists (s3 \+ s2), s1, (s3' \+ s2).
  rewrite !(joinC s1) !joinA in C C' *; split=>//.
  right; split; first by exists s3, s2, s3'.
  - by split; [move/cohH/validHL: C | exists s3, s2; case/guar_coh: G]. 
  by split; [move/cohH/validHL: C' | exists s3', s2; case/guar_coh: G]. 
case; case=>s12 [s3][s12'][->->{s s'}] G C C'.
have C3 : s3 \In Mod.coh W3. 
- by rewrite -(star_coh_prec C); case/guar_coh: G.
case: G C C'; last first.
- case=>_ [_][/Mem_map_inv][ct1][->] H1 /Mem_map_inv [ct2][->] H2 G C C'.
  right; exists (Chan ct1), (Chan (frame_coreL (Mod.coh W3) ct2)). 
  split; first by apply/Mem_map.
  - by apply/Mem_map; apply/Mem_cat; left; apply: Mem_map. 
  rewrite ent_guarC entguar_frameLE.
  split; try by rewrite -starA starCA starA. 
  by exists s12, s3, s12'; rewrite ent_guarC.
case.
- case; case=>s1 [s2][s1'][->->{s12 s12'}] G _ _ C C'.
  left; left; split; try by rewrite starA.
  by exists s1, (s2 \+ s3), s1'; rewrite !joinA.
case; case=>s2 [s1][s2'][->->{s12 s12'}] G _ _ C C'.
left; right; split; try by rewrite starC starA. 
exists (s2 \+ s3), s1, (s2' \+ s3).
rewrite -!(joinAC _ s3) in C C' *; split=>//.
left; split; first by exists s2, s3, s2'. 
- by split; [move/cohH/validHL: C | exists s2, s3; case/guar_coh: G].
by split; [move/cohH/validHL: C' | exists s2', s3; case/guar_coh: G].
Qed.

Lemma entangle_apart_coreA W1 W2 W3 : 
        entangle W1 (apart W2 W3) =  
        entangle (entangle W1 W2) W3 :> mod_core.
Proof. by apply/coreE; split; [rewrite /= starA | apply: entangle_apartA]. Qed.

Lemma entangle_apart_chanA W1 W2 W3 : 
        Mod.chans (entangle W1 (apart W2 W3)) =p 
        Mod.chans (entangle (entangle W1 W2) W3).
Proof.
move=>x; split; case/Mem_map_inv=>_ [->{x}] /Mem_map_inv [x][->] T.
- apply/mem_chan.
  exists (frame_coreL (Mod.coh W3) (frame_coreL (Mod.coh W2) x)).
  split; last by apply/Mem_map; apply/Mem_map. 
  apply/chanE=>/=; split; first by [rewrite starA];
  by case: x T=>/= a r T h; rewrite frame_relA.
case/Mem_map_inv: T=>y [->{x}] T; apply/mem_chan.
exists (frame_coreL (star_coh (Mod.coh W2) (Mod.coh W3)) y).
split; last by apply/Mem_map. 
apply/chanE=>/=; split; first by [rewrite starA];
by move=>h; rewrite frame_relA.
Qed.

Lemma internal_relE (W1 W1' W2 W2' : mod) :
        Mod.coh W1 = Mod.coh W2 ->
        Mod.coh W1' = Mod.coh W2' ->
        Mod.internal W1 = Mod.internal W2 :> reltp ->
        Mod.internal W1' = Mod.internal W2' :> reltp ->
        Mod.chans W1 =p Mod.chans W2 ->
        Mod.chans W1' =p Mod.chans W2' ->
        entangle_intguar W1 W1' = entangle_intguar W2 W2' :> reltp. 
Proof.
move=>Ec Ec' Eg Eg' Ejm Ejm'.
apply: union_relE; last by apply: guarsum_congr.
by apply: union_relE; apply: frame_relE.
Qed.

(***************************************************)
(***************************************************)
(* Lifting apartness composition to lists of worls *)
(***************************************************)
(***************************************************)

(* I should really use sets instead of lists, and build the module's *)
(* relation components by means of quantifiers; but that sounds like *)
(* too much right now, so let's stick with sequences *)

(* Computing an apartness composition of a sequence *)

Fixpoint apart_iter (ls : seq mod) := 
  if ls is W' :: ls' then 
    apart W' (apart_iter ls')
  else empty_mod.

(* AC lemmas *)

Section IterPermLemmas.

Lemma apartit_coh (ws : seq mod) : 
        Mod.coh (apart_iter ws) = 
        star_iter_coh (map (Mod.coh \o Mod.core) ws).
Proof.
elim: ws=>[|w ws IH] /=; first by rewrite star_iter_emp.
by rewrite ssE IH.
Qed.

Definition ssE := (ssE, apartit_coh).

Lemma apartit_coh_perm ls1 ls2 : 
        perm ls1 ls2 -> 
        Mod.coh (apart_iter ls1) = Mod.coh (apart_iter ls2).
Proof. by move=>H; rewrite !ssE (star_iter_perm (perm_map _ H)). Qed.

Lemma apartit_chan_perm ls1 ls2 : 
        perm ls1 ls2 -> 
        Mod.chans (apart_iter ls1) =p Mod.chans (apart_iter ls2).
Proof.
pose P := fun ls1 ls2 => 
  Mod.chans (apart_iter ls1) <=p Mod.chans (apart_iter ls2).
suff {ls1 ls2} L : forall ls1 ls2, perm ls1 ls2 -> P ls1 ls2.
- by move=>H x; split; apply: L=>//; apply/perm_sym.
apply: perm_ind2=>//=; last 1 first.
- by move=>ls1 ls2 ls3 H1 IH1 H2 IH2 x /IH1/IH2. 
- move=>W ls1 ls2 H IH x /=.
  case/Mem_map_inv=>y [->{x}] /Mem_cat T. 
  rewrite -toPredE /= /chan_seq_in /= map_cat. 
  apply/Mem_cat; case: T=>T; [left | right].
  - case/Mem_map_inv: T=>z [->{y}] T. 
    rewrite (apartit_coh_perm H).
    by apply/Mem_map; apply/Mem_map.
  case/Mem_map_inv: T=>z [->{y}] T. 
  case/mem_chan_eq/IH/Mem_map_inv: T=>w [[E]]. 
  rewrite E in z *; move/inj_pair2=>-> T.
  by apply/Mem_map; apply/Mem_map. 
move=>W1 W2 ls1 ls2 H IH x /=.
case/Mem_map_inv=>y [->{x}] /Mem_cat =>T.
rewrite -toPredE /= /chan_seq_in /= map_cat. 
apply/Mem_cat; case: T.
- case/Mem_map_inv=>z [->{y}] T; right; apply/mem_chan.
  exists (frame_coreR (Mod.coh W1) 
         (frame_coreL (Mod.coh (apart_iter ls2)) z)).
  split; last by apply/Mem_map; apply/Mem_cat; left; apply/Mem_map. 
  case: z T=>a r /= T; apply/chanE=>/=. 
  split; first by [rewrite starCA (apartit_coh_perm H)];
  by move=>h; rewrite frame_relA starC (apartit_coh_perm H). 
case/Mem_map_inv=>z [->{y}] /Mem_cat; case.
- case/Mem_map_inv=>w [->{z}] T; left; apply/mem_chan.
  exists (frame_coreL (star_coh (Mod.coh W2) (Mod.coh (apart_iter ls2))) w).
  split; last by apply/Mem_map. 
  case: w T=>a r /= T; apply/chanE=>/=.
  split; first by [rewrite starCA (apartit_coh_perm H)];
  by move=>h; rewrite frame_relA starC (apartit_coh_perm H).  
case/Mem_map_inv=>w [->{z}] T; right.
case/mem_chan_eq/IH/Mem_map_inv: T=>v [[E]] /=.
rewrite E in w *; move/inj_pair2=><-{v} T; apply/mem_chan. 
exists (frame_coreR (Mod.coh W1) (frame_coreR (Mod.coh W2) w)).
split; last by apply/Mem_map; apply/Mem_cat; right; apply/Mem_map. 
apply/chanE=>/=; split; first by [rewrite starCA];
by move=>h; rewrite -!(starC (Mod.coh (apart_iter _)));
            rewrite !do_relE !frame_relA starC.
Qed.

Lemma apartit_int_perm ls1 ls2 : 
        perm ls1 ls2 -> 
        Mod.internal (apart_iter ls1) = 
        Mod.internal (apart_iter ls2) :> reltp.
Proof.
move: ls1 ls2; apply/perm_ind2=>//; last 1 first.
- by move=>s2 s1 s3 H1 -> H2 ->.
- move=>W ls1 ls2 H IH /=. 
  apply: union_relE; first by rewrite (apartit_coh_perm H).
  by rewrite IH (apartit_coh_perm H).
move=>W1 W2 ls1 ls2 H IH; rewrite -!apart_intA. 
apply: apart_int_congr; last first.
- by apply: apartit_chan_perm H.
- by apply/coreE; split=>//; apply: apartit_coh_perm H.
- by apply: apart_chanC.
by apply: apart_coreC.
Qed.

Lemma apartit_core_perm ls1 ls2 : 
        perm ls1 ls2 -> apart_iter ls1 = apart_iter ls2 :> mod_core.
Proof.
move=>H; apply/coreE; split.
- by apply: apartit_coh_perm.
by apply: apartit_int_perm.
Qed.

End IterPermLemmas.


